#include "api/df_active_mode.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStartActiveConnection( const char *networkAddress, int listenPort, int deviceAddress, int timeout, int aliveTimeout, int maxConnections, int infoFlags, int unused_setTo0, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBeginPure( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( networkAddress == nullptr || inet_addr( networkAddress ) == INADDR_NONE
        || maxConnections < 1 || maxConnections > 140 ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t(%i, %i)", 
            (networkAddress != nullptr) ? networkAddress : "", listenPort, deviceAddress, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0,
            (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    uint16_t l_listen_port = static_cast<uint16_t>( listenPort );
    uint16_t l_timeout = static_cast<uint16_t>( timeout );
    uint16_t l_alive_timeout = static_cast<uint16_t>( aliveTimeout );
    uint16_t l_max_connections = static_cast<uint16_t>( maxConnections );
    uint32_t l_info_flags = static_cast<uint32_t>( infoFlags );
    int32_t l_error_number = IV_ERROR_SUCCESS;
    // 2012_07_09 (BO) V04.02.03: deviceID fix auf 0 gestellt.
    int iRes = theApp.startActiveMode( networkAddress, l_listen_port, 0, l_timeout, l_alive_timeout, l_max_connections, l_info_flags, unused_setTo0, &l_error_number );

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( 0, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( 0 ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t%i\t(%i, %i)", 
        networkAddress, listenPort, deviceAddress, timeout, aliveTimeout, maxConnections, infoFlags, unused_setTo0, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCStopActiveConnection( int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr ) 
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBeginPure( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( false )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    int iRes = theApp.stopActiveMode();

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFirstActiveConnectionId()
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return -1;
    }

    channelBridge->logBeginPure( __func__, "" );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = -1;
    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp.m_mutexMapChannelBridge );
        // Ersten Kanal im Active-Mode ermitteln.
        for ( auto iter = theApp.m_mapChannelBridge.begin(); iter != theApp.m_mapChannelBridge.end(); ++iter ) 
        {
            if ( iter->first < ACTIVE_CHANNEL_UPPER_BOUND )
            {
                if ( iter->second->isActive() )
                {
                    iRes = iter->first;
                    break;
                }
            }
            else
            {
                // Key liegt ausserhalb der Suchgrenze.
                break;
            }
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFirstActiveChannelID()
{
    return DFCGetFirstActiveConnectionId();
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetNextActiveConnectionId( int previousId )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return -1;
    }

    channelBridge->logBeginPure( __func__, "\t%i", previousId );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    // Naechste aktive Verbindung ermitteln
    int iRes = -1;
    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp.m_mutexMapChannelBridge );
        // Ersten Kanal im Active-Mode ermitteln.
        for ( auto iter = theApp.m_mapChannelBridge.begin(); iter != theApp.m_mapChannelBridge.end(); ++iter ) 
        {
            if ( iter->first < ACTIVE_CHANNEL_UPPER_BOUND )
            {
                if ( iter->first > previousId && iter->second->isActive() )
                {
                    iRes = iter->first;
                    break;
                }
            }
            else
            {
                // Key liegt ausserhalb der Suchgrenze.
                break;
            }
        }
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetNextActiveChannelID( int previousId )
{
    return DFCGetNextActiveConnectionId( previousId );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetInfoActiveChannel( int connectionNumber, char *value, int *valueLength )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    int iRes = 0;

    channelBridge->logBeginPure( __func__, "\t%i", connectionNumber );

    // zulaessige Bereiche pruefen
    if ( value == nullptr || valueLength == nullptr ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    // Besteht die Aktive Verbindung
    if ( theApp.isActiveMode() ) 
    {
        std::lock_guard<std::recursive_mutex> guard( theApp.m_mutexMapChannelBridge );
        if ( theApp.m_mapChannelBridge.count( connectionNumber ) && theApp.m_mapChannelBridge[connectionNumber]->isActive() ) 
        {
            Channel4 &channel = theApp.m_mapChannelBridge[connectionNumber]->m_channel4;
            // Aktuellen Infostring aus Kanal übernehmen.
            if ( channel.activeConnectionInfo( value, valueLength ) )
            {
                iRes = 1;
            }
        }
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", value, *valueLength, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetRecordAvailable( int enabled )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBeginPure( __func__, "\t%i", enabled );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = theApp.SetRecordAvailable( enabled );

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCRecordAvailable( int *connectionNumber, int *deviceAddress, char *infoText, int *infoTextLength, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBeginPure( __func__, "" );

    // zulaessige Bereiche pruefen
    if ( connectionNumber == nullptr || deviceAddress == nullptr 
        || infoTextLength == nullptr || (infoText == nullptr && *infoTextLength > 0) ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t\t\t\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    int iRes = 0;
    // Ist das Melden von Datenaetzen aktiviert?
    if ( theApp.IsNotifyRecord() ) 
    {
        do 
        {        
            if ( (iRes = theApp.RecordAvailable( connectionNumber, deviceAddress )) == IV_ERROR_SUCCESS ) 
            {
                std::lock_guard<std::recursive_mutex> guard( theApp.m_mutexMapChannelBridge );
                if ( theApp.m_mapChannelBridge.count( *connectionNumber ) )
                {
                    Channel4 &channel = theApp.m_mapChannelBridge[*connectionNumber]->m_channel4;
                    // Aktuellen Infostring aus Kanal übernehmen.
                    if ( *infoTextLength > 0 && channel.activeConnectionInfo( infoText, infoTextLength ) == false )
                    {
                        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
                        iRes = 0;
                    }
                    else
                    {
                        iRes = 1;
                    }
                }
                else
                {
                    // Wiederholung, da diese Kanalnummer nicht verfuegbar.
                    continue;
                }
            } 
            else if ( iRes > IV_ERROR_SUCCESS ) 
            {
                channelBridge->setLastErrorNumber( 0, IV_ERROR_PROCESSING, errorNumber );
                iRes = 0;
            }
            else 
            {
                channelBridge->setLastErrorNumber( 0, -iRes, errorNumber );
                iRes = 0;
            }
        } while ( false );
    } 
    else 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_AC_NO_NOTIFY_RECORD, errorNumber );
    }

    channelBridge->logEnd( "\t%i\t%i\t%s\t%i\t%i\t(%i, %i)", *connectionNumber, *deviceAddress, infoText, *infoTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCBindDeviceToChannel( int connectionNumber, int deviceType, int deviceSerial, const char *deviceIp, int *errorNumber )
{
    int id = LOG_CHANNEL_ID;
    ChannelBridge *channelBridge = theApp.channelBridge( id );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBeginPure( __func__, "\t%i\t%i\t%i", connectionNumber, deviceType, deviceSerial );

    // Liegen die Werte im gueltigen Bereich.
    if ( connectionNumber < 1 || connectionNumber >= ACTIVE_CHANNEL_LOWER_BOUND 
        || (deviceType != -1 && (deviceType < 0 || deviceType >= IV_VALUE_TYPE_DEVICE_COUNT)) 
        || (deviceIp != nullptr && strlen( deviceIp ) && inet_addr( deviceIp ) == INADDR_NONE) )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (deviceIp != nullptr) ? deviceIp : "", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    int iRes = 0;
    // Es gibt 4 moegliche mappings
    // 1. Loeschen eines bestehenden mappings
    // 2. map channelID => deviceType + deviceSerial
    // 3. map channelID => deviceType + deviceSerial + deviceIP
    // 4. map channelID => deviceIP
    ChannelBridge *channel = theApp.channelBridge( connectionNumber );
    if ( channel->isBind() && deviceType == -1 && (deviceIp == 0 || !strlen( deviceIp )) ) 
    {
        // Bindung wird entfernt.

        // Bindungsstruktur aufbauen und zuweisen
        Binding_info bindingInfo;
        bindingInfo.connection_id = connectionNumber;

        // Hinzufuegen zur Liste der zu bearbeitenden Eintraege.
        {
            std::lock_guard<std::mutex> guard( theApp.m_mutexListBindingInfo );
            theApp.m_listBindPending.push_back( bindingInfo );  
        }

        // Wenn der Active-Mode-Thread läuft, dann darauf warten, dass die BindingInfo verarbeitet wurde.
        if ( theApp.isActiveMode() )
        {
            std::unique_lock<std::mutex> lk( theApp.m_mutexConditionBindingInfo );
            theApp.m_conditionListBindingInfo.wait_for( lk, std::chrono::seconds( 10 ) );
        }

        iRes = 1;
    } 
    else if ( (deviceType > -1 && deviceSerial < 1) || (deviceType == -1 && (deviceIp == nullptr || !strlen( deviceIp ))) ) 
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
    } 
    else if ( channel->isBind() || channel->isOpen() )
    {
        // Der Kanal darf nicht bereits gebunden oder geoeffnet sein.
        channelBridge->setLastErrorNumber( 0, IV_ERROR_CHANNEL_USED, errorNumber );
    }
    else
    {
        // Bindungsstruktur aufbauen und zuweisen
        Binding_info bindingInfo;
        bindingInfo.connection_id = connectionNumber;
        bindingInfo.bi.state = 1;
        bindingInfo.bi.typeDevice = deviceType;
        bindingInfo.bi.serial = deviceSerial;
        if ( deviceIp != 0 )
        {
            bindingInfo.bi.ip = deviceIp;
        }

        // Hinzufuegen zur Liste der zu bearbeitenden Eintraege.
        {
            std::lock_guard<std::mutex> guard( theApp.m_mutexListBindingInfo );
            theApp.m_listBindPending.push_back( bindingInfo );
        }

        // Wenn der Active-Mode-Thread läuft, dann darauf warten, dass die BindingInfo verarbeitet wurde.
        if ( theApp.isActiveMode() )
        {
            std::unique_lock<std::mutex> lk( theApp.m_mutexConditionBindingInfo );
            theApp.m_conditionListBindingInfo.wait_for( lk, std::chrono::seconds( 10 ) );
        }

        iRes = 1;
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (deviceIp != nullptr) ? deviceIp : "", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}
