#pragma once
#ifndef CHANNEL_4_H
#define CHANNEL_4_H

// Basisklasse einbinden
#include "datafox/communication/module_table.h"

// Klasse für den Datenspeicher einbinden.
#include "channel_4_buffer.h"

// defines fuer Setup
#include "file/interface/studio_setup_structs.h"

// Geraetedateiarchiv einbinden
#include "file/file_device_zip.h"

// Studiodatei einbinden
#include "file/file_aes.h"

#include "file/file_hardware_info.h"

#include "common/device_info.h"
#include "common/tcpip_info.h"

// Klasse zur Behandlung des Firmwareupdates
#include "common/device_update.h"

#include <string>

#include <thread>
#include <mutex>

// Defines setzen, wenn nicht bereits geschehen
#if !defined(_MSC_VER) && !defined(__MINGW32__)

#ifndef WINAPI
#define WINAPI __stdcall
#endif

typedef void *LPVOID;
typedef uint32_t DWORD;

#endif

#if defined(_MSC_VER) || defined(__MINGW32__)
#ifndef S_ISREG
#define S_ISREG(m) ((m & _S_IFMT) == _S_IFREG)
#endif

#ifndef S_ISDIR
#define S_ISDIR(m) ((m & _S_IFMT) == _S_IFDIR)
#endif
#endif

#ifndef STRUCT_MEM_FILE_HEADER
#define STRUCT_MEM_FILE_HEADER
struct StructMemFileHeader
{
    int32_t structSize;          // Strukturgroesse
    int32_t type;                // Speichertyp
    int32_t baseAddress;         // Basisadresse
    int32_t memAddress;          // Anfangsadresse
    int32_t memSize;
    char tmp[76];
    //...
    int32_t crcHead;    // 32bit CRC fuer Head (Immer letzter Wert in der Struktur; 
    // Berechnungsmenge: (sizeof(SDFFileDeviceHeader) - sizeof(uint16_t)), also ohne sich selbst!)
};
#endif

// Struktur für die Datei der Fingertemplates
//############################################
// Innerer Aufbau der Datei ist wie folgt:
// StructFTFileHeader, (count * [(type == 0) ? structFingertemplateDIN : structFingertemplateIdencom, CRC16 ])
// - Zuerst liegt die Headerstruktur StructFTFileHeader in der Datei.
// - Der erste 16bit-Wert der Datei muss 0xDFDF sein.
// - Der crc der Headerstruktur berechnet sich über die Struktur ohne die zwei Byte des CRC selbst.
// - Die Daten beginnen ab der Position Seek(StructFTFileHeader.offs, Filebegin).
// - Der Datenteil besteht aus StructFTFileHeader.count Segmenten, die folgenden Aufbau haben.
//   + Zuerst kommt die Sturktur des Fingertemplates je nach StructFTFileHeader.type
//   + Darauf folgt ein 16bit CRC der sich über diese Struktur berechnet.
struct StructFTFileHeader
{
    uint16_t Id;    // Kennung der Datei: muss immer 0xDFDF sein.
    uint16_t offs;  // Offset auf den Datenbeginn, normalerweise Strukturgröße, falls Erweiterung nötig.
    uint16_t count; // Anzahl enthaltener Fingertemplates (zur Sicherheit mit aufgenommen)
    uint8_t type;   // Templatetyp. 0=>DIN V66400 (161Byte), 1=>Idencom-Compact (216Byte)
    uint16_t crc;   // 16bit CRC über die Headerstruktur (ohne diesen letzten 16bit Wert)
};

class Channel4 : public Datafox::Communication::ModuleTable, public Channel4Buffer
{
public:
    Channel4( int32_t connectionNumber = Datafox::Communication::ConnectionBase::defaultConnectionNumber(), uint8_t deviceAddress = Datafox::Communication::ProtocolBase::defaultDeviceAddress() );

    virtual ~Channel4();

    /// Schliessen der Schnittstelle.
    virtual int32_t closeConnection()
    {
        int32_t res = 0;
        // Ist die Verbindung im Active-Mode?
        if ( activeModeState() >= Datafox::Communication::ams_active )
        {
            // Beendigung sicherstellen, auch wenn er bereits beendet wurde.
            if ( m_communicationThread.joinable() )
            {
                // WorkerThread fuer Socket beenden.
                setCommunicationThreadStopSignal();
                m_communicationThread.join();
            }

            res = 1;
        }
        else if ( connection().isOpen() )
        {
            res = ModuleTable::closeConnection();
        }

        return res;
    }

    // Verbindungsannahme von uebergebenem Listen-Socket
    int32_t connect( SOCKET* listen_socket, int32_t connectionNumber, uint8_t deviceAddress = Datafox::Communication::ProtocolBase::defaultDeviceAddress(), int32_t timeout = 3000, int32_t aliveTimeout = 10, int32_t infoFlags = 0 );

    // Aktuelle Info über den Verbundenen Teilnehmen abrufen.
    bool activeConnectionInfo( char* info, int32_t* info_length );

    // Main - Funktion fuer Befehlsausfuehrung in Workerthread
public:
    tcpipInfo m_tcpipInfo;
    int32_t m_notify;
protected:
    volatile bool m_communicationThreadStartSignal;
    void setCommunicationThreadStartSignal( bool signal = false );
    bool communicationThreadStartSignal();

    volatile bool m_communicationThreadStopSignal;
    void setCommunicationThreadStopSignal( bool signal = true );
    bool communicationThreadStopSignal();

    volatile int32_t m_communicationThreadError;
    void setCommunicationThreadError( int32_t error );
    int32_t communicationThreadError();

    std::thread m_communicationThread;
    void communicationThread();

    // Socket fuer connect
    SOCKET *m_listenSocket;
    int32_t m_aliveTimeout;
    int32_t m_infoFlags;

public:
    // Ausfuehren eines Protokollbefehls
    virtual int32_t call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue = -1 );

public:
    // Beginn und Ende der Kommunikation ueber einen TimeboySlot
    int32_t setTimeboySlotState( uint8_t deviceAddress, int32_t state, int32_t commandResult = IV_ERROR_SUCCESS )
    {
        static clock_t stamp = 0;
        int32_t res;
        if ( (res = protocol().setTimeboySlotState( deviceAddress, state, commandResult )) == IV_ERROR_SUCCESS )
        {
            // TimeboyIV zur Kommunikationsinitialisiserung einen Ping senden
            if ( (state && ((clock() - stamp) / CLOCKS_PER_SEC > 3)) || protocol().packetNumber() == 0 )
            {
                if ( (res = ping( deviceAddress, nullptr, nullptr )) == IV_ERROR_SUCCESS )
                {
                    stamp = clock();
                }
            }
        }

        // Wenn es kein TimeboySlot ist, dann trotzdem OK melden.
        if ( res == -IV_ERROR_SLOT_NONE )
        {
            res = IV_ERROR_SUCCESS;
        }

        return res;
    }

    void initDeviceUpdate( const std::string &path, const uint32_t type )
    {
        if ( m_deviceUpdate )
        {
            std::string errorMsg;
            m_deviceUpdate->updatePathToDB( path, static_cast<DeviceUpdate::DBPathType>(type), errorMsg );
        }
    }

public:
    int32_t generateFingerprintIdList( uint8_t deviceAddress, int32_t flags, uint32_t &recordCount );
    int32_t fingerprintBackup( uint8_t deviceAddress, const std::string &filePath, int32_t flags );
    int32_t fingerprintRestore( uint8_t deviceAddress, const std::string &filePath, int32_t flags );

#define OFF_EXTERNAL_READER         0x00000001 // External Reader options 0=GIS, 1=PHG
#define OFF_IOMODULE                0x00000002 // IOModul
#define OFF_TBADAPTION              0x00000004 // TBAdaption
#define OFF_COMSERVER               0x00000008 // ComServer
#define OFF_TPWRITE                 0x00000010 // TransponderWrite
#define OFF_HTTP                    0x00000020 // Http-Protokoll
#define OFF_BIOKEY_UPDATE           0x00000040 // UpdateBiokey
#define OFF_SO_IOCONTROL            0x00000080 // SetupOption: Signalverarbeitung
#define OFF_KEEP_SETUP_AFTER_UPDATE 0x00000100 // Setup und Listen nach Update beibehalten
    int32_t setOptionFirmware( uint8_t deviceAddress, int32_t mask, int32_t options );
    int32_t getOptionFirmware( uint8_t deviceAddress, int32_t &mask, int32_t &options );

    int32_t m_loadAsDefaultLanguage; // Nachfolgende Uebertragung der Textdaten als Default.lng
    int32_t setTextDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t setTextData( uint8_t deviceAddress, const std::string &filePath );
    int32_t setTouchKeyDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t setTouchKeyData( uint8_t deviceAddress, const std::string &filePath );
    int32_t setDisplayColorsWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t setDisplayColors( uint8_t deviceAddress, const std::string &filePath );

    int32_t setWlanConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t setWlanConfigData( uint8_t deviceAddress, const std::string &filePath );
    int32_t getWlanConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t getWlanConfigData( uint8_t deviceAddress, const std::string &filePath );

    int32_t setHardwareConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t setHardwareConfigData( uint8_t deviceAddress, const std::string &filePath );
    int32_t getHardwareConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t getHardwareConfigData( uint8_t deviceAddress, const std::string &filePath );

    int32_t writeFileDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t writeFileData( uint8_t deviceAddress, const std::string &filePath );
    int32_t readFileDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t readFileData( uint8_t deviceAddress, const std::string &filePath );

private:
    // Dateiarchiv fuer Geraetedaten (Firmware) verwalten
    //FileDeviceZip m_fileDeviceZip;
    std::unique_ptr<DeviceUpdate> m_deviceUpdate;

public:
    int32_t readTablesWithTimeboySlot( uint8_t deviceAddress, const std::string &typeName );
    int32_t readTables( uint8_t deviceAddress, const std::string &type_name );

    int32_t writeSetupWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t writeSetupData( uint8_t deviceAddress, const std::string &filePath );
    int32_t writeSetupData4( uint8_t deviceAddress );
    int32_t writeSetupData4_2( uint8_t deviceAddress );

    int32_t readSetupWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath );
    int32_t readSetupData( uint8_t deviceAddress, const std::string &filePath );
    int32_t readSetupData4( uint8_t deviceAddress );
    int32_t readSetupData4_2( uint8_t deviceAddress );

    int32_t writeListWithTimeboySlot( uint8_t deviceAddress );
    int32_t writeListData( uint8_t deviceAddress );
    int32_t writeListData4( uint8_t deviceAddress );

    int32_t writeEntrance2ListWithTimeboySlot( uint8_t deviceAddress, uint8_t listNumber );
    int32_t writeEntrance2ListData4( uint8_t deviceAddress, uint8_t listNumber );
    int32_t writeEntrance2List4( uint8_t deviceAddress, uint8_t listNumber );

    int32_t writeTimeboyListWithTimeboySlot( uint8_t deviceAddress, uint8_t id, uint8_t listNumber, int32_t reserve );
    int32_t writeTimeboyList4( uint8_t deviceAddress, uint8_t id, uint8_t listNumber, int32_t reserve );
    int32_t writeTimeboyListData4( uint8_t deviceAddress, uint8_t id, uint8_t listNumber, int32_t reserve );

    std::string m_hardwareInfoAsText; // Puffer fuer die Hardwareinfo in Textform.
    int32_t writeDeviceWithTimeboySlot( uint8_t deviceAddress, std::string &filePath );
    int32_t writeDevice( uint8_t deviceAddress, std::string &filePath );
    int32_t writeDevice4( uint8_t deviceAddress );

    SoftwarebuildVersion m_minimalUpdateVersion;
    std::list<std::string> m_blacklistVersion{};
    // Erhalte die Älteste Firmware-Version und eine Liste von Blacklist-Versionen, welche für info zulässig ist anhand der Datenbank einer dfz-Datei. 
    std::tuple<std::string, const std::list<std::string>> minFwVersion(const std::string& dbPath, HardwareInfo& info);

#pragma pack(push, 1)
    struct UploadModuleSignatur 
    {
        uint8_t version;
        uint8_t type;
        uint16_t flags;
        uint32_t crc32;
    } *m_uploadModuleSignatur;
#pragma pack(pop)
    int32_t writeModuleWithTimeboySlot(uint8_t deviceAddress, int32_t type, const std::string &filePath);
    int32_t writeModule(uint8_t deviceAddress, int32_t type, const std::string &filePath);

    int32_t getVersionFirmwareFromFile( const std::string &filePath, std::string &version );

    int32_t writeMemToFile( uint8_t deviceAddress, int32_t type, int32_t address, int32_t size, const std::string &filePath );
    int32_t writeFileToMem( uint8_t deviceAddress, const std::string &filePath );

    int32_t checkData( uint8_t deviceAddress );

    // Überlagerungen aus Channel4ModuleApplication
    int32_t readRecord( uint8_t deviceAddress, uint8_t *buffer, uint16_t &count, uint8_t &status );
    int32_t quitRecord( uint8_t deviceAddress, uint8_t &status, std::string &param );
    int32_t clearRecords( uint8_t deviceAddress );
    int32_t restoreRecords( uint8_t deviceAddress );
    int32_t writeMemoToDisplay( uint8_t deviceAddress, uint8_t maschine, uint8_t time, uint8_t font, uint8_t audio, uint16_t flags, const std::string &text );
    int32_t writeMessageToDisplay( uint8_t deviceAddress, uint8_t time, uint8_t action, uint8_t audio, const std::string &text );
    int32_t writeSetupVariable( uint8_t deviceAddress, const std::string &name, const std::string &val );
    int32_t readSetupVariable( uint8_t deviceAddress, const std::string &name, std::string &val );
    int32_t writeFontType( uint8_t deviceAddress, uint8_t type );

    // Überlagerung aus Channel4ModuleSystem
    int32_t readSystemLogRecord( uint8_t deviceAddress, uint8_t *buffer, uint16_t &count, uint8_t &status );
    int32_t quitSystemLogRecord( uint8_t deviceAddress, uint8_t &status );
    int32_t restoreSystemLogRecords( uint8_t deviceAddress );
    int32_t readRtcDateTime( uint8_t deviceAddress, char* time );
    int32_t writeRtcDateTime( uint8_t deviceAddress, char* time );
    int32_t readFirmwareVersion( uint8_t deviceAddress, std::string &version );
    int32_t systemReset( uint8_t deviceAddress, uint16_t modus );
    int32_t readSerialNumber( uint8_t deviceAddress, uint32_t &serial );
    int32_t writeRelayState( uint8_t deviceAddress, const std::string &name, uint8_t busNumber, uint8_t relayNumber, uint8_t status, uint16_t time );
    int32_t readRelayState( uint8_t deviceAddress, const std::string &name, uint8_t busNumber, uint8_t relayNumber, uint8_t &status, uint16_t &time );
    int32_t writeSystemVariable( uint8_t deviceAddress, const std::string &name, const std::string &val );
    int32_t readSystemVariable( uint8_t deviceAddress, const std::string &name, std::string &val );
    int32_t writeVirtualKey( uint8_t deviceAddress, uint8_t keyCode, uint8_t flags );

    // Informationen über das Gerät sammeln.
    int32_t readDeviceInfo( uint8_t deviceAddress, DeviceInfo &info );
    // Informationen über die verbaute Hardware im Gerät sammeln.
    int32_t readHardwareInfo( uint8_t deviceAddress, HardwareInfo &info );
    // Informationen über die Enthaltene Firmware.
    int32_t readSoftwareInfo( uint8_t deviceAddress, SoftwareInfo &info );
    // Informationen über die Zusatzoptionen der Firmware.
    int32_t readFirmwareOptions( uint8_t deviceAddress, FirmwareOptions &options );
    // Setzen der Zusatzoptionen der Firmware.
    int32_t writeFirmwareOptions( uint8_t deviceAddress, const FirmwareOptions &options );

    // Daten getunnelt über Docking schicken.
    int32_t enterTransparencyMode( uint8_t deviceAddress, uint16_t id, uint16_t flags, uint32_t readBufferSize, uint32_t writeBufferSize, uint32_t timeout );
    int32_t exitTransparencyMode( uint8_t deviceAddress );
    bool isTransparencyMode();

    // Gelbe LED bei Dockingstation-V2 setzen/lesen
    int32_t writeLedState( uint8_t deviceAddress, uint8_t slot, uint8_t id, uint8_t state);
    int32_t readLedState( uint8_t deviceAddress, uint8_t slot, uint8_t id, uint8_t &state);

    int32_t readDataBlockWithTimeboySlot( uint8_t deviceAddress, const std::string &name, char *data, uint32_t *length, uint32_t flags );
    int32_t readDataBlock( uint8_t deviceAddress, const std::string &name, char *data, uint32_t *length, uint32_t flags );

    // FW-Update
    int32_t dfzList( std::list<std::string> *list );

private:
    void cleanUpUpdateDB();
    int32_t detectHWInfo( const uint8_t &deviceAddress, StructHardwareInfo &oldHWInfo, HardwareInfo &hwInfo );
    int32_t determineMinFWVersion( const StructHardwareInfo &oldHWInfo, HardwareInfo &hwInfo, const std::string &filePath,
                               std::list<std::pair<std::string, std::string>> &processedModules );
    int32_t determineFWFile( std::string zipFilePath, const StructHardwareInfo &oldHWInfo );
};

#endif
