/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/protocol_base.h"
#include <stdio.h>
#include <stdlib.h>
#include <boost/format.hpp>
#include <memory.h>

#include <thread>
#include <chrono>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connection_id Nummer der Kanalklasse fuer z.B. Trace.
* @param device_address Busnummer der Kanalklasse.
*/
Datafox::Communication::ProtocolBase::ProtocolBase( int connectionNumber, uint8_t deviceAddress )
    : m_connection( std::make_shared<ConnectionBase>( connectionNumber ) ), m_trace( connection().tracePointer() )
{
    // Umgebung zuruecksetzen
    reset();

    // DeviceID der Klasse zuweisen
    m_deviceNumberSender = deviceAddress;
    m_deviceNumberReceiver = defaultDeviceNumberStandAlone();

    // Puffer zur schnellen Uebertragung
    std::fill_n( m_temp, sizeof( m_temp ), 0 );
    m_tempPos = m_temp;

    // Die Umgebung des verwendeten CRC32 initialisieren
    initCrc32Table();

    // Membervariablen fuer Paketauswertung initialisieren
    m_packetStart = false;
    m_skipCount = 0;
    m_packetNumber = 1;
    m_lastFlags = 0;

    m_clocksWrite = m_clockWrite = 0;
    m_clocksRead = m_clockRead = 0;

    m_packetWriteCounter = 0;
    m_packetWritePayloadSize = 0;

    m_packetReadCounter = 0;
    m_packetLostCounter = 0;
    m_packetReadPayloadSize = 0;

    m_qs = 0;

    // Callbacks intialisieren
    m_callBackFunction = nullptr;
    remoteCall = nullptr;

    // Puffer fuer Senden, Empfang
    std::fill_n( m_packetSend, sizeof( m_packetSend ), 0 );
    std::fill_n( m_packetReceive, sizeof( m_packetReceive ), 0 );

    // Hilfsmember initialisieren
    m_previousModule   = 0;
    m_previousFunction = 0;

    // Zufallsgenerator starten
    srand( static_cast<uint32_t>( time( nullptr ) ) );

    // Verschluesselung aktivieren
    df4Init( true );

    // Vorgegebene Version der Dockingstation. _timeboyStation wird nur bei Version == 1 verwendet.
    m_timeboyStationEnabled = 1;

    // Kommunikationsbasis an TimeboyStation weitergeben
    m_timeboyStation.setConnection( connection() );
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Datafox::Communication::ProtocolBase::~ProtocolBase() 
{
}


//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Datafox::Communication::ProtocolBase::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    resetError();
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion.
*/
void Datafox::Communication::ProtocolBase::setCallBackFunction(  std::function<int32_t ()> func  )
{
    trace( "%s(%i):INFO(3): %s( --omitted-- )\n", __FILE__, __LINE__, __func__ );

    m_callBackFunction = func;
}

//#########################################################
//######################### CRC ###########################
//#########################################################

// ##################################################################
/**
* @brief Reflection is a requirement for the official CRC-32 standard.
* You can create CRCs without it, but they won't conform to the standard.
*/
uint32_t Datafox::Communication::ProtocolBase::reflect( uint32_t ref, int8_t ch )
{
    // Used only by initCrc32Table()
    trace( "%s(%i):INFO(3): %s( %u, %d )\n", __FILE__, __LINE__, __func__, ref, ch );

    uint32_t val = 0;

    // Swap bit 0 for bit 7
    // bit 1 for bit 6, etc.
    int32_t i;
    for( i = 1; i < (ch + 1); ++i )
    {
        if ( ref & 1 )
            val |= 1 << (ch - i);

        ref >>= 1;
    }

    return val;
}

// ##################################################################
/**
* @brief remoteCall this func only once to initialize the CRC table.
*/
void Datafox::Communication::ProtocolBase::initCrc32Table() 
{
    // Called by Init
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // This is the official polynomial used by CRC-32 
    // in PKZip, WinZip and Ethernet. 
    uint32_t ulPolynomial = 0x04c11db7;

    // 256 values representing ASCII character codes.
    uint16_t i;
    uint8_t j;
    for( i = 0; i <= 0xFF; ++i )
    {
        m_crc32Table[i] = reflect( i, 8 ) << 24;
        for ( j = 0; j < 8; ++j )
            m_crc32Table[i] = (m_crc32Table[i] << 1) ^ (m_crc32Table[i] & (1 << 31) ? ulPolynomial : 0);

        m_crc32Table[i] = reflect( m_crc32Table[i], 32 );
    }
}

// ##################################################################
/**
* @brief This func uses the crc32_table lookup table to generate a CRC for data.
*/
uint32_t Datafox::Communication::ProtocolBase::crc32( uint8_t *data, std::size_t size, uint32_t crc, bool closure )
{
    trace( "%s(%i):INFO(3): %s( %p, %u, 0x%08X, %u )\n", __FILE__, __LINE__, __func__, data, size, crc, closure );

    // Be sure to use unsigned variables,
    // because negative values introduce high bits
    // where zero bits are required.
    //unsigned int crc = 0xffffffff;
    std::size_t len;
    uint8_t *buffer;

    len = size;
    // Save the text in the buffer.
    buffer = data;
    // Perform the algorithm on each character
    // in the string, using the lookup table values.
    while ( len-- )
        crc = (crc >> 8) ^ m_crc32Table[(crc & 0xFF) ^ *buffer++];

    // Exclusive OR the result with the beginning val.
    return (closure) ? crc ^ 0xffffffff : crc;
}

// ##################################################################
/**
* @brief Erzeugt einen 16bit CRC Wert fuer die uebergebenen Daten.
*/
uint16_t Datafox::Communication::ProtocolBase::crc16( uint8_t *data, std::size_t size, uint16_t crc )
{
    trace( "%s(%i):INFO(3): %s( %p, %u, 0x%04X ]\n", __FILE__, __LINE__, __func__, data, size, crc );

    std::size_t i;
    uint8_t j;

    for ( i = 0; i < size; ++i )
    {
        crc = crc ^ data[i];

        for ( j = 0; j < 8; ++j )
        {
            if ( crc & 0x01 )
            {
                crc >>= 1; 
                crc = crc ^ 0xA001;
            }
            else 
            {
                crc >>= 1;
            }
        }
    }

    return crc;
}


//#########################################################
//################ Verschluesselung #######################
//#########################################################

void Datafox::Communication::ProtocolBase::df4LoadCryptKey()
{
    uint8_t cryptKey_DF4[32] = 
    {
        0xc5, 0xc2, 0xc3, 0xc0, 0xc1, 0xd9, 0xdb, 0xf0, 
        0xcd, 0x07, 0xa7, 0xaf, 0xaa, 0x0a, 0x50, 0xae,
        0xc5, 0x4e, 0xc3, 0xd9, 0xaf, 0xd9, 0x09, 0xc1,
        0xa7, 0xef, 0xac, 0xc2, 0x0a, 0xbf, 0xcd, 0xae
    };

    std::copy_n( cryptKey_DF4, sizeof( cryptKey_DF4 ), m_key );
}

// ##################################################################
/**
* @brief Initialisiert die sbox fuer die DF4-Verschlusselungsroutine.
*/
void Datafox::Communication::ProtocolBase::df4Init( bool binary_key )
{
    uint16_t i;
    uint8_t j;
    uint8_t *kptr = m_key;
    uint8_t trans;

    for( i = 0; i <= 255; ++i )
    {
        m_sbox[i] = static_cast<uint8_t>( i );

        m_k[i] = *kptr;
        kptr++;

        if ( binary_key )
        {
            if( kptr == (m_key + sizeof( m_key )) )
                kptr = m_key;
        }
        else
        {
            if ( *kptr == 0 )
                kptr = m_key;
        }
    }

    for( i = 0, j = 0; i <= 255; ++i )
    {
        j = (j + m_sbox[i] + m_k[i]) % 256;
        trans = m_sbox[i];
        m_sbox[i] = m_sbox[j];
        m_sbox[j] = trans;
    }
}

// ##################################################################
/**
* @brief Ver-/Entschluesselt die uebergebenen Daten.
*/
void Datafox::Communication::ProtocolBase::df4( uint8_t *data, std::size_t size )
{
    uint8_t i = 0;
    uint8_t j = 0;
    uint8_t t;
    std::size_t pos = 0;
    uint8_t trans;

    // Kopie von sbox, da sie unten manipuliert wuerde.
    std::copy_n( m_sbox, sizeof( m_k ), m_k );

    // Wenn die Angabe einer Laenge fehlt, wird davon ausgegangen,
    // dass mit einer Zeichenkette gearbeitet wird.
    if ( size == 0 )
    {
        size = strlen( (const char *)data );
    }

    for( ; (size - pos) > 0; ++pos )
    {
        i = (i + 1) % 256;
        j = (j + m_k[i]) % 256;
        trans = m_k[i];
        m_k[i] = m_k[j];
        m_k[j] = trans;
        t = (m_k[i] + m_k[j]) % 256;
        data[pos] ^= m_k[t];
    }
}

// ##################################################################
/**
* @brief Schreiben eines teilaufbereiteten Pakets in die Kommunikationsschnittstelle.
* Der Kopf und Paket-CRC wird vor Paketversand gebildet und zur Paketgroesse dazugerechnet
* und nach Versand wieder entfernt.
*/
int32_t Datafox::Communication::ProtocolBase::writeProtocol( uint8_t *buffer, uint8_t writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %u )\n", __FILE__, __LINE__, __func__, buffer, writeTrials );

    int32_t res = IV_ERROR_SUCCESS;
    uint32_t transmittedCount = 0;
    uint32_t transmitted;
    bool enableMasking = true;
    uint16_t masked = 0;

    // Laengenangabe pruefen
    if ( *(uint16_t *)&buffer[4] < IV_PACKET_LENGTH_FRAME && *(uint16_t *)&buffer[4] >= IV_PACKET_LENGTH_MAX )
    {
        trace( "%s(%i):ERROR(1): %s() [ packet length fail: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_PACKETLENGTH );
        return -IV_ERROR_PACKETLENGTH;
    }

    // Temppuffer setzen
    m_tempPos = m_temp;

    // Handelt es sich um eine Dockingstation?
    if ( isTimeboySlot( deviceNumberReceiver() ) )
    {
        // Keine Maskierung der Paketdaten bei versenden ueber eine Dockingstation.
        enableMasking = false;
    }

    // CRC - groesse daraufrechnen
    *(uint16_t *)&buffer[4] += sizeof( uint16_t );
    // CRC von Paketkopf einfuegen
    *(uint16_t *)&buffer[8] = crc16( buffer, IV_PACKET_LENGTH_HEAD - sizeof( uint16_t ) );
    // CRC von Paket anfuegen
    *(uint16_t *)&buffer[*(uint16_t *)&buffer[4] - sizeof( uint16_t )] = crc16( buffer, *(uint16_t *)&buffer[4] - sizeof( uint16_t ) );
    /*
    if (IV_PACKET_MODULE(buffer) != 0) 
    {
    // CRC von Paket fälschen
    if (!(rand() % 8)) 
    {
    *(unsigned short*)&buffer[*(unsigned short*)&buffer[4] - sizeof(unsigned short)] += 1;
    Trace("%s(%i):ERROR(1): %s [ verfälsche gesamt CRC ]\n", __FILE__, __LINE__, __func__);
    }
    }
    */
    
    // Ggf. Paketdaten raustracen.
    if ( connection().traceMode() >= TraceBase::TraceMode::packets )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }


    // Datenstrom zeichenweise bearbeiten und besondere Zeichen maskieren
    for ( ; transmittedCount < *(uint16_t *)&buffer[4]; ++transmittedCount )
    {
        // Paketzeichen im Datenstrom mit Escapesequenz umschreiben
        if ( enableMasking && (transmittedCount > 1) 
            && ((buffer[transmittedCount] == IV_PACKET_SIGN_START)
            || (buffer[transmittedCount] == IV_PACKET_SIGN_ESCAPE)) )
        {
            unsigned char replace[2];
            replace[0] = IV_PACKET_SIGN_ESCAPE;
            // Anzahl maskierte mitzaehlen
            masked++;

            switch ( buffer[transmittedCount] ) 
            {
            case IV_PACKET_SIGN_START:
                replace[1] = 0x01;
                break;

            case IV_PACKET_SIGN_ESCAPE:
                replace[1] = 0x00;
                break;
            }

            // Wuerde Pufferende erreicht?
            if ( ((m_tempPos + sizeof( replace )) - m_temp) > static_cast<int32_t>( sizeof( m_temp ) ) ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Puffer overflow: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_TRANSMIT );
                return -IV_ERROR_TRANSMIT;
            }

            *m_tempPos = reinterpret_cast<uint8_t *>( replace )[0];
            m_tempPos++; // Auf naechstes Zeichen

            *m_tempPos = reinterpret_cast<uint8_t *>( replace )[1];
            m_tempPos++; // Auf naechstes Zeichen

        } 
        else 
        {
            // Wuerde Pufferende erreicht?
            if ( ((m_tempPos + 1) - m_temp) > static_cast<int32_t>( sizeof( m_temp ) ) ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Puffer overflow: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_TRANSMIT );
                return -IV_ERROR_TRANSMIT;
            }

            *m_tempPos = buffer[transmittedCount];
            m_tempPos++; // Auf naechstes Zeichen
        }
    }

    // Absenden des Gesamtpaketes
    if ( transmittedCount == *(uint16_t *)&buffer[4] )
    {
        m_clockWrite = 0;
        // Anfang des Versendens halten
        clock_t clockStart = clock();
        if ( isTimeboySlot( deviceNumberReceiver() ) )
        {
            if ( (res = m_timeboyStation.writeBytes( deviceNumberReceiver(), m_timeboyBaseAddress, 0, m_temp, *(uint16_t *)&buffer[4] + masked )) != IV_ERROR_SUCCESS )
            {
                trace( "%s(%i):ERROR(1): %s() [ writeBytes fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                goto label_error;
            }
        }
        else
        {
            if ( (res = connection().writeData( m_temp, *(uint16_t *)&buffer[4] + masked, transmitted, writeTrials )) != IV_ERROR_SUCCESS )
            {
                trace( "%s(%i):ERROR(1): %s() [ TransmitData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                goto label_error;
            }
        }

        // Gesamtzeit des Versendevorgangs fuer Statistik halten
        m_clockWrite = clock() - clockStart;
        m_clocksWrite += m_clockWrite;
    }

label_error:
    // CRC - groesse wieder abziehen
    *(uint16_t *)&buffer[4] -= sizeof( uint16_t );
    // CRC von Paketkopf loeschen
    *(uint16_t *)&buffer[8] = 0;
    // CRC von Paket loeschen
    *(uint16_t *)&buffer[*(uint16_t *)&buffer[4]] = 0;

    return res;
}

// ##################################################################
/**
* @brief Schreiben eines Pakets fuer die angegebene Modulfunktion.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
// Aufbereiten und senden eines Anfrage-Pakets
//  deviceAddress -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int32_t Datafox::Communication::ProtocolBase::writePacket( uint8_t deviceAddress, uint8_t type, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t count, uint8_t writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u )\n", __FILE__, __LINE__, __func__, deviceAddress, module, func, buffer, count );

    int err = IV_ERROR_SUCCESS;

    // CallBack-Function vor Paketversand aufrufen.
    if ( m_callBackFunction != nullptr )
    {
        if ( !m_callBackFunction() ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ CallBack abbruch ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_BREAK;
        }
    }

    // Remoteadresse setzen
    setDeviceNumberReceiver( deviceAddress );

    /* 
    1.	(  8 Bit): Startzeichen ("~", 0x7E)
    2.	(  8 Bit): Empfängeradresse
    3.	(  8 Bit): Absenderadresse
    4.	(  8 Bit): Steuerinformationslänge
    5.	(16 Bit): Paketgesamtlänge
    6.	(  8 Bit): Flags (z. B. für aktive Verschlüsselung)
    7.	(  8 Bit): Reserve
    8.	(16 Bit): CRC über die ersten 7 Felder
    9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
    10.	(n*8 Bit): Zu transportierende Daten
    11.	(16 Bit): CRC über Felder 1 - 10.
    */
    // 1.
    m_packetSend[0] = IV_PACKET_SIGN_START;
    // 2.
    m_packetSend[1] = deviceNumberReceiver();
    // 3.
    m_packetSend[2] = deviceNumberSender();
    // 4.
    m_packetSend[3] = IV_PACKET_LENGTH_CONTROLBLOCK;
    // 5.
    *(uint16_t *)&m_packetSend[4] = uint16_t(IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t ) + sizeof( uint16_t ) * 2 + count);
    // 6.
    m_packetSend[6] = 0;
    // 7.
    m_packetSend[7] = 0;
    // 8.
    // Wird von Transmit - Funktion erledigt
    // 9.
    // Zu Punkt 9: Steuerinformation -> Packezaehler
    // Ist Paket eine Anfrage?
    if ( type & IV_PACKET_REQUEST ) 
    {
        // Aktuelle Paketnummer
        m_packetSend[IV_PACKET_LENGTH_HEAD] = packetNumber();
    } 
    else if ( type & IV_PACKET_ANSWER ) 
    {
        // Paket ist Antwort auf Anfrage!
        // Steuerinformation verfuegbar? 
        if ( m_packetReceive[3] > 0 ) 
        {
            // Dann in die Antwort uebernehmen.
            m_packetSend[IV_PACKET_LENGTH_HEAD] = m_packetReceive[IV_PACKET_LENGTH_HEAD];
        }
    }

    // 10.
    m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3]] = IV_PACKET_VERSION_4 | type;
    *(uint16_t *)&m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t )] = module;
    *(uint16_t *)&m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t ) + sizeof( uint16_t )] = func;

    // Daten in Sendebuffer kopieren
    memcpy( &m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3] + sizeof( uint8_t ) + sizeof( uint16_t ) * 2], buffer, count );	

    // 10a
    // Daten Verschluesselung wird durch CChannel4ProtocolCrypt::writeProtocolCrypt durchgefuehrt

    // Zu 10: Paketdaten ggf. verschluesseln
    if ( (err = writeProtocolCrypt()) != IV_ERROR_SUCCESS )
    {
        trace( "%s(%i):ERROR(1): %s() [ writeProtocolCrypt result: %i ]\n", __FILE__, __LINE__, __func__, err );
        // Ggf. Paketdaten raustracen.
        if ( traceMode() >= TraceBase::TraceMode::packets )
            tracePacket( m_packetSend );

        return err;
    }
    // 11.
    // Wird von Transmit-Funktion erledigt

    trace( "%s(%i):INFO(1): %s() (id: %i, %d -> %d, X: %02X, B: %02X, M: %d, F: %d, PaketNummer: %d)\n", __FILE__, __LINE__, __func__,
        connectionNumber(), deviceNumberSender(), deviceNumberReceiver(),
        m_packetSend[6], m_packetSend[IV_PACKET_LENGTH_HEAD + m_packetSend[3]], module, func, m_packetSend[IV_PACKET_LENGTH_HEAD] );

    // Protokoll-Paket senden
    if ( (err = writeProtocol( m_packetSend, writeTrials )) != IV_ERROR_SUCCESS )
    {
        trace( "%s(%i):ERROR(1): %s() [ writeProtocol result: %i ]\n", __FILE__, __LINE__, __func__, err );
        return err;
    }

    // Zeitpunkt des Paketversands halten
    m_timeWritePacket = time( nullptr );

    // Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
    m_packetWriteCounter++;
    m_packetWritePayloadSize += count;

    return err;
}

// ##################################################################
/**
* @brief Lesen eines Pakets inklusive Pruefung von Kopf und Gesamt-CRC.
*/
int32_t Datafox::Communication::ProtocolBase::readProtocol( uint8_t *buffer, uint8_t readTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %u )\n", __FILE__, __LINE__, __func__, buffer, readTrials );

    uint32_t countEscape = 0;
    uint32_t received;
    bool escape = false;

    m_clockRead = 0;
    clock_t clockRx = 0;

    int32_t res;
    int error = IV_ERROR_SUCCESS;

    int receiveCount = IV_PACKET_LENGTH_HEAD;
    int receivedCount = 0;

    m_skipCount = 0;

    // Liegt Paketstart breits vor?
    if ( m_packetStart )
    {
        buffer[0] = IV_PACKET_SIGN_START;
        receivedCount++;
        // bereits gelesenen Paketstart uebernommen
        m_packetStart = false;
    }

    // Datenstrom zeichenweise lesen und Paketstartzeichen herausfiltern
    for ( ; receivedCount < receiveCount; ++receivedCount ) 
    {
        // Einstieg zum lesen eines Zeichens
        received = 0;

        // Empfangszeit fuer Pakete ermitteln.
        clock_t clockStart = clock();
        if ( (res = connection().readData( &buffer[receivedCount], 1, received, readTrials )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
            return (error != IV_ERROR_SUCCESS) ? error : res;
        }
        clockRx += clock() - clockStart;

        // Empfangene Anzahl Zeichen erhoehen
        connection().addSizeRead( received );

        // Error abarbeiten
        if ( error != IV_ERROR_SUCCESS ) 
        {
            // Erstes Zeichen muss Startzeichen sein.
            if ( buffer[receivedCount] != IV_PACKET_SIGN_START ) 
            {
                // Startzeichen fehlt noch
                trace( "%s(%i):WARNING(1): %s() [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[receivedCount] );
                // Maximal zu lesende Zeichen ohne Paketstart
                if (++m_skipCount > IV_SKIPCOUNT) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, __func__, error );
                    return -IV_ERROR_SKIPCOUNT;
                }

                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            } 
            else 
            {
                // Paketstart gelesen
                m_packetStart = true;
            }

            trace( "%s(%i):ERROR(1): %s() [ error: %i ]\n", __FILE__, __LINE__, __func__, error );
            return error;
        }

        // Erstes Zeichen muss Startzeichen sein.
        if ( !receivedCount && buffer[receivedCount] != IV_PACKET_SIGN_START ) 
        {
            // Startzeichen fehlt noch
            trace( "%s(%i):WARNING(1): %s() [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[receivedCount] );

            // Paketzeit reset
            //ct1 = clock(); 

            // Maximal zu lesende Zeichen ohne Paketstart
            if ( ++m_skipCount > IV_SKIPCOUNT ) 
            {
                trace("%s(%i):ERROR(1): %s() [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_SKIPCOUNT);
                return -IV_ERROR_SKIPCOUNT;
            }

            // Empfang zuruecksetzen
            receivedCount = -1;
            continue;
        }
        else if ( receivedCount && buffer[receivedCount] == IV_PACKET_SIGN_START ) 
        {
            // Ein Startzeichen wurde im Paket empfangen.
            trace("%s(%i):ERROR(1): %s() [ Startbyte in packet: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_START_IN_PACKET);

            m_packetStart = true;
            return -IV_ERROR_START_IN_PACKET;
        } 
        else if ( receivedCount && (buffer[receivedCount] == IV_PACKET_SIGN_ESCAPE) ) 
        {
            // Escape bereits eingeschaltet?
            if ( escape ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Double escape: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_ESCAPE );

                // Escape ausschalten
                escape = false;
                // Fehler setzen
                error = -IV_ERROR_ESCAPE;
                m_skipCount = 0;
                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            }

            // Escape einschalten
            escape = true;
            // Emfpangszeiger zuruecksetzen
            receivedCount--;
            continue;
        } 
        else if ( escape ) 
        {
            // Escape ausschalten
            escape = false;

            // Zu escapierendes Zeichen
            switch ( buffer[receivedCount] ) 
            {
            case 0:
                buffer[receivedCount] = IV_PACKET_SIGN_ESCAPE;
                break;

            case 1:
                buffer[receivedCount] = IV_PACKET_SIGN_START;
                break;

            case 2:
                // Bleibt im Puffer erhalten
                break;

            default:
                trace( "%s(%i):ERROR(1): %s() [ Unknown escape-sequenz: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_ESCAPESEQUENZ );
                //escapeError =  -IV_ERROR_ESCAPESEQUENZ;
                m_skipCount = 0;
                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            }

            // Escapezeichen mitzaehlen
            countEscape++;
        }

        // Paketdaten auswerten
        // Es gibt 2 Auswertesituationen.
        // 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Paketdaten setzen.
        // 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen
        if ( (receivedCount + 1) == receiveCount ) 
        {
            // Paketkopf auswerten
            if ( (receivedCount + 1) == IV_PACKET_LENGTH_HEAD ) 
            {
                // CRC pruefen
                if ( crc16( buffer, receivedCount + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad head-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_HEAD );

                    int p = 0; // Fehlerhafte Kopfdaten ausgeben
                    for (; p < (receivedCount + 1); p++) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip( %3u ) head Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, p + 1, buffer[p] );
                    }

                    error = -IV_ERROR_CRC_HEAD;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }

                // Paketgroesenangabe pruefen
                if ( *(uint16_t *)&buffer[4] < IV_PACKET_LENGTH_FRAME || *(uint16_t *)&buffer[4] >= IV_PACKET_LENGTH_MAX )
                {
                    error = -IV_ERROR_PACKETLENGTH;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }

                // Neue Empfangsgroesse setzen
                receiveCount = *(uint16_t *)&buffer[4];
            }
            else // Gesamt-CRC pruefen
            {
                // CRC pruefen
                if ( crc16( buffer, receivedCount + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );

                    int p = 0; // Fehlerhafte Kopf- / Paketdaten ausgeben
                    for ( ; p < (receivedCount + 1); ++p ) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip( %3u ) %s Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, p + 1, (p < IV_PACKET_LENGTH_HEAD) ? "head" : "paket", buffer[p] );
                    }

                    error = -IV_ERROR_CRC_PACKET;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }
            }
        } 
    }

    // Gesamtzeit des Lesevorgangs fuer Statistik halten
    m_clockRead = clockRx;
    m_clocksRead += m_clockRead;

    // Ggf. Paketdaten raustracen.
    if ( traceMode() >= TraceBase::TraceMode::packets )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen eines Pakets inklusive Pruefung von Kopf und Gesamt-CRC.
*/
int32_t Datafox::Communication::ProtocolBase::readTimeboyStation( uint8_t *buffer, uint8_t readTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %u )\n", __FILE__, __LINE__, __func__, buffer, readTrials );

    static uint32_t crcLast = 0;
    static uint32_t crcRepeated = 0;
    uint32_t received;

    m_clockRead = 0;
    clock_t clockStart = clock();
    clock_t clockEnd = clockStart + connection().readTimeout();
    clock_t clockEndActive = clockStart + IV_PENDING_TIMEOUT;

    int res = IV_ERROR_SUCCESS;

    uint16_t toReceive = IV_PACKET_LENGTH_FRAME; // Haelt die zu empfangende Anzahl Bytes
    uint16_t receivedCount = 0;                  // Haelt die Gesamtanzahl gelesener Bytes in _tmp
    uint16_t headSearches = 0;                   // Haelt die Anzahl der Leseversuche eines Packetanfanges beim Timeboy

    uint16_t receivePos = 0; // Position des Tmp - Lesepuffers
    uint16_t packetPos = 0;  // Position des Paketpuffers

    // Enum fuer Kommunikationsstatus.
    enum 
    {
        begin,         // Beginn der Paketauswertung.
        headSearch,    // Suche des Paketbeginns '~'
        headReceive,   // Daten des Paketkopfes werden empfangen.
        headComplete,  // Paketkopf komplett, CRC-Pruefung war OK.
        bodyReceive,   // Paketdaten werden empfangen.
        complete,      // Paket komplett empfangen
        failed
    } receiveStatus; // Aktueller Empfangsstatus

    // Aktuellen Empfangsstatus setzen.
    receiveStatus = begin;

    // Solange Zeichen zu empfangen sind.
    for ( ; receiveStatus != complete && receiveStatus != failed; )
    {

        // Je nach aktuellem Empfangsstatus.
        switch ( receiveStatus )
        {
        case begin: // Beginn der Paketauswertung.

            // Temp - Lesepuffer zuruecksetzen
            receivePos = 0;
            receivedCount = 0;
            // Auswertungspuffer zuruecksetzen
            packetPos = 0;

            // Zuerst zu lesende Menge auf Maximale Puffergroesse setzen.
            // Die Dockingstation liest an der Folgeadresse das naechste Datenpaket aus dem Timeboy!
            // Wir hier eine kleine Paketgroesse angegeben werden alle Folgepakete mit dieser Laenge geliefert.
            toReceive = MB_SLOT_BUFFER_DATA_SIZE_SMALL;//GetDockingBufferSize(getDeviceIdRemote());

            // Empfangsstatus wechseln
            receiveStatus = headSearch;
            break;

        case headSearch:

            // Erstes Datenbyte uebernehmen
            buffer[0] = m_temp[0];

            // Erstes Zeichen muss Startzeichen sein.
            if (buffer[0] != IV_PACKET_SIGN_START)
            {	
                headSearches++;
                trace( "%s(%i):WARNING(1): %s() [ Paket start wait (%i): 0x%02X ]\n", __FILE__, __LINE__, __func__, headSearches, buffer[0] );

                // Abbruch nach mehrfachem Nachlesen.
                if ( m_timeboyBaseAddress == 0x70000000 && headSearches > 12 && clock() > clockEnd ) // 10 HeadSearches entspricht ca. 2500ms
                {
                    // Zu viele Versuche,
                    res = -IV_ERROR_SKIPCOUNT;
                    // Fehlerzustand annehmen.
                    receiveStatus = failed;
                    continue;
                }
                else if ( m_timeboyBaseAddress == 0x70010000 && headSearches > 50 && clock() > clockEndActive )
                {
                    // Zu viele Versuche,
                    res = -IV_ERROR_SKIPCOUNT;
                    // Fehlerzustand annehmen.
                    receiveStatus = failed;
                    continue;
                }

                // Zeit fuer Versuche abgelaufen
                //if (clock() > clockEnd)
                //{
                //	// Zu viele Versuche,
                //	res = -IV_ERROR_RECEIVE;
                //	// Fehlerzustand annehmen.
                //	receiveStatus = failed;
                //	continue;
                //}

                // Erneut den Paketstart lesen.
                receiveStatus = begin;
                continue;
            }

            // Paketanfang erfolgt
            receiveStatus = headReceive;
            continue;

        case headReceive:

            packetPos++;
            receivePos++;
            // Naechsten Wert fuer Auswertung uebernehmen
            buffer[packetPos] = m_temp[receivePos];

            // Paketdaten auswerten
            // Es gibt 2 Auswertesituationen.
            // 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Paketdaten setzen.
            // 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen

            // Paketkopf auswerten
            if ( (packetPos + 1) == IV_PACKET_LENGTH_HEAD ) 
            {
                // CRC pruefen
                if ( crc16( buffer, IV_PACKET_LENGTH_HEAD ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad head-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_HEAD );

                    int p = 0; // Fehlerhafte Kopfdaten ausgeben
                    for (; p < packetPos; p++) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip head checked Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[p] );
                    }

                    res = -IV_ERROR_CRC_HEAD;
                    // Schnittstelle bereinigen
                    receiveStatus = failed;
                    continue;
                }

                // Paketgroesenangabe pruefen
                if ( *(uint16_t *)&buffer[4] < IV_PACKET_LENGTH_FRAME || *(uint16_t *)&buffer[4] >= IV_PACKET_LENGTH_MAX )
                {
                    res = -IV_ERROR_PACKETLENGTH;
                    // Schnittstelle bereinigen
                    receiveStatus = failed;
                    continue;
                }

                // aktueller Status ist Paketkopf gueltig empfangen.
                receiveStatus = headComplete;
                continue;
            }
            break;

        case headComplete:
            // Anzahl nachzulesender Datenbytes setzen?
            if ( *(uint16_t *)&buffer[4] > receivedCount )
            {
                // Neue Empfangsgroesse setzen
                toReceive = *(uint16_t *)&buffer[4];// - receivedCount;
            }

            // Empfangsstatus anpassen
            receiveStatus = bodyReceive;
            break;

        case bodyReceive:

            packetPos++;
            receivePos++;
            // Naechsten Wert fuer Auswertung uebernehmen
            buffer[packetPos] = m_temp[receivePos];

            if ( (packetPos + 1) == *(uint16_t *)&buffer[4] )
            {
                // Gesamt - CRC pruefen
                if ( crc16( buffer, packetPos + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );
                    res = -IV_ERROR_CRC_PACKET;
                    // Emfpang zuruecksetzen
                    receiveStatus = failed;
                    continue;
                }

                if ( crcLast != crc32( buffer, packetPos + 1 ) )
                {
                    // Letzten CRC von Packet merken.
                    crcLast = crc32( buffer, packetPos + 1 );
                    crcRepeated = 0;
                }
                else
                {
                    // Wiederholt gleiches Paket gelesen
                    if ( crcRepeated++ > 3 )
                    {
                        crcLast = 0;
                        crcRepeated = 0;
                        // Dreimalig gleiches Paket gelesen.
                        trace( "%s(%i):ERROR(1): %s() [ Repeated packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );
                        res = -IV_ERROR_SKIPCOUNT;
                        // Emfpang zuruecksetzen
                        receiveStatus = failed;
                        continue;
                    }

                }
                // Paket komplett empfangen
                receiveStatus = complete;
                continue;
            } 
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ receiveStatus unbekannt: %i ]\n", __FILE__, __LINE__, __func__, receiveStatus );
        }

        // Sind Zeichen zu lesen?
        if ( toReceive > 0 )
        {
            // Einstieg zum lesen eines Teilpaketes
            received = 0;

            if ( (res = m_timeboyStation.readBytes( deviceNumberReceiver(), m_timeboyBaseAddress, receivedCount, m_temp, toReceive )) != IV_ERROR_SUCCESS )
            {
                trace( "%s(%i):ERROR(1): %s() [ readBytes fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                return res;
            }

            // Empfangsanzahl angleichen
            received = toReceive;

            // Gesamtanzahl gelesener Bytes halten
            receivedCount += received;

            // Anzahl zu lesender Bytes wurde erreicht, somit ist diese nun 0
            toReceive -= received;

            // Statistik: Empfangene Anzahl Zeichen erhoehen
            connection().addSizeRead( received );
        }
    }

    // Gesamtzeit des Lesevorgangs fuer Statistik halten
    m_clockRead = clock() - clockStart;
    m_clocksRead += m_clockRead;

    // Ggf. Paketdaten raustracen.
    if ( traceMode() >= TraceBase::TraceMode::packets )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }

    return res;
}

// ##################################################################
/**
* @brief Lesen eines Pakets aus der Kommunikationsschnittstelle.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
// Empfangen und verarbeiten eines Pakets
//  deviceAddress -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int32_t Datafox::Communication::ProtocolBase::readPacket( uint8_t deviceAddress, uint8_t &type, uint16_t &module, uint16_t &func, uint8_t *buffer, uint16_t &count, uint8_t readTrials )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u )\n", __FILE__, __LINE__, __func__, deviceAddress, module, func, buffer, count );

    int err = IV_ERROR_SUCCESS;

    // Adresse setzen
    setDeviceNumberSender( deviceAddress );

    /* 
    1.	(  8 Bit): Startzeichen ("~", 0x7E)
    2.	(  8 Bit): Empfängeradresse
    3.	(  8 Bit): Absenderadresse
    4.	(  8 Bit): Steuerinformationslänge
    5.	( 16 Bit): Paketgesamtlänge
    6.	(  8 Bit): Flags - für Verschlüsselung und Flashprüfung (z. B. für aktive Verschlüsselung)
    7.	(  8 Bit): Reserve
    8.	( 16 Bit): CRC über die ersten 7 Felder
    9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
    10.	(n*8 Bit): Zu transportierende Daten
    11.	( 16 Bit): CRC über Felder 1 - 10.
    */

receive_again:
    // Wird ueber Dockingstation kommuniziert?
    if ( isTimeboySlot( deviceNumberReceiver() ) )
    {
        // Protokoll-Paket empfangen
        if ( (err = readTimeboyStation( m_packetReceive, readTrials )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ readTimeboyStation error: %i ]\n", __FILE__, __LINE__, __func__, err );
            return err;
        }
    }
    else
    {
        // Protokoll-Paket empfangen
        if ( (err = readProtocol( m_packetReceive, readTrials )) != IV_ERROR_SUCCESS ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ readProtocol error: %i ]\n", __FILE__, __LINE__, __func__, err );
            return err;
        }
    }

#if _DEBUG && 0
    if ( m_previousModule == 0x1 && m_previousFunction == 0xA )
    {
        m_previousModule = 0;
        trace( "%s(%i):ERROR(1): %s() [ getFlashFake error ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_CRC_PACKET;
    }
#endif

    //	Empfaengeradresse pruefen
    if ( IV_PACKET_DEVICEID(m_packetReceive) != deviceNumberSender() ) 
    {
        err = -IV_ERROR_ADDRESS;
        trace( "%s(%i):WARNING(1): %s()  [ Wrong deviceAddress %i for %i ]\n", __FILE__, __LINE__, __func__, IV_PACKET_DEVICEID(m_packetReceive), deviceNumberSender() );

        // Handelt es sich um keine Docking,
        // weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
        if ( isTimeboySlot( deviceNumberReceiver() ) )
        {
            return err;
        }

        goto receive_again;
    }

    // Variable für die Flags in Member halten
    m_lastFlags = IV_PACKET_FLAGS( m_packetReceive );
    /////////////////////////////////
    // Die Flags der Flashprüfen prüfen
    if ( m_lastFlags & IV_PACKET_FLAG_FLASH_SERVICE )
    {	// Flash Service: Globaler Wear-Level ist >= 6
        trace( "%s(%i):WARNING(1): %s() [ Flash Service Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, m_lastFlags );
        // Flash Service occurce
    }
    if ( m_lastFlags & IV_PACKET_FLAG_FLASH_WARNING )
    {	// Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu übertragen)
        trace( "%s(%i):WARNING(1): %s() [ Flash Warning Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, m_lastFlags );
        // Flash Warning occurce
    }
    if ( m_lastFlags & IV_PACKET_FLAG_FLASH_LOCKED )
    {	// Flash Error: Keine freien Seiten mehr verfügbar, Gerät befindet sich im Bedienstopp.
        trace( "%s(%i):WARNING(1): %s() [ Flash Error Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, m_lastFlags );
        // Flash Error occurce
    }

    // Zu Punkt 9: Steuerinformation Paketzaehler
    // Steuerinformation verfuegbar?
    if ( m_packetReceive[3] > 0 )
    {
        // Ist Paket eine Antwort auf Anfrage?
        if ( m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3]] & IV_PACKET_ANSWER )
        {
            // Ist Paketnummer der letzten Anfrage?
            // Oder handelt es sich um einen Fehler vom Typ "fehlerhafter gesamt-crc", "fehlerhafte Entschluesselung",
            // hierbei wird der Packetcounter auf 0 gesetzt das Modul und die Funktion ist 0xFFFF,
            // weil diese Daten nicht sichergestellt werden koennen.
            if ( m_packetReceive[IV_PACKET_LENGTH_HEAD] == packetNumber()
                || (m_packetReceive[IV_PACKET_LENGTH_HEAD] == 0 && m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3]] & IV_PACKET_ERROR
                && *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t )] == 0xFFFF
                && *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t ) + sizeof( uint16_t )] == 0xFFFF) )
            {
                // Paketnummer erhoehen
                incPacketNumber();
            }
            else
            {
                err = -IV_ERROR_ADDRESS;
                trace( "%s(%i):WARNING(1): %s() [ Skip wrong packet number %i instead %i ]\n", __FILE__, __LINE__, __func__, m_packetReceive[IV_PACKET_LENGTH_HEAD], packetNumber() );
                // Ggf. Paketdaten raustracen.
                if ( traceMode() >= TraceBase::TraceMode::packets )
                    tracePacket( m_packetReceive );

                // Handelt es sich um keine Docking,
                // weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
                if ( isTimeboySlot( deviceNumberReceiver() ) )
                {
                    return err;
                }

                goto receive_again;
            }
        }
    }

    // Empfangene Daten ggf. entschluesseln
    if ( (err = readProtocolCrypt()) != IV_ERROR_SUCCESS ) 
    {
        trace( "%s(%i):ERROR(1): %s() [ readProtocolCrypt error: %i ]\n", __FILE__, __LINE__, __func__, err );
        return err;
    }

    // Pakettyp uebernehmen
    type = m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3]];
    // Anzahl der Datenbyte
    count = *(uint16_t *)&m_packetReceive[4] - IV_PACKET_LENGTH_FRAME - m_packetReceive[3] - sizeof( uint8_t ) - sizeof( uint16_t ) * 2;

    // Modul und Funktion uebernehmen
    module = *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t )];
    func = *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t ) + sizeof( uint16_t )];

    // Daten uebernehmen
    memcpy( buffer, &m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t ) + sizeof( uint16_t ) * 2], count );

    // Datenpuffer mit einer Nullterminierung versehen um einen Zugriff auf einen eventl. nicht Nullterminierten letzten Parameterwert zu schützen.
    buffer[count] = 0;

    trace( "%s(%i):INFO(1): %s() (id: %i, %i <- %i, X: %02X, B: %02X, M: %i, F: %i, PacketNummer: %i)\n", __FILE__, __LINE__, __func__, connectionNumber(), deviceNumberSender(), m_packetReceive[2],
        m_packetReceive[6],
        m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3]], 
        *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t )],
        *(uint16_t *)&m_packetReceive[IV_PACKET_LENGTH_HEAD + m_packetReceive[3] + sizeof( uint8_t ) + sizeof( uint16_t )],
        (m_packetReceive[3]) ? m_packetReceive[IV_PACKET_LENGTH_HEAD] : 0 );

    // Zeitpunkt des Packetempfangs halten
    m_timeReadPacket = time( nullptr );

    // Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
    m_packetReadCounter++;
    m_packetReadPayloadSize += count;

    // remoteCall-Back Funktion aufrufen, ggf. abbruch.
    if ( m_callBackFunction != nullptr ) 
    {
        if ( !m_callBackFunction() ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ CallBack abbruch ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_BREAK;
        }
    }

    return err;
}

// ##################################################################
/**
* @brief Remoteaufruf einer lokalen Modulfunktion.
*/
void Datafox::Communication::ProtocolBase::tracePacket( uint8_t *packetBuffer, bool crlf )
{
    uint16_t len = *(uint16_t *)&packetBuffer[4];
    if ( len <= IV_PACKET_LENGTH_MAX )
    {
        char tmp[10];
        char *buf = new char[len * 3 + 10];
        memset( buf, 0, len * 3 + 10 );
        for ( len = 0; len < *(uint16_t *)&packetBuffer[4]; ++len )
        {
            if ( len == 0 )
                sprintf( tmp, "[%02X ", packetBuffer[len] );
            else if ( len == (*(uint16_t *)&packetBuffer[4] - 1) )
                sprintf( tmp, "%02X]%c", packetBuffer[len], crlf ? '\n' : ' ' );
            else
                sprintf( tmp, "%02X ", packetBuffer[len]);

            strcat( buf, tmp );
        }

        trace( buf );
        delete[] buf;
    }
}

// ##################################################################
/**
* @brief Ausfuehrung einer Modulfunktion in ihrem jeweiligen Kontext.
*
*/
int32_t Datafox::Communication::ProtocolBase::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u, %i )\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, contextValue );

    int trials;
    int receiveTrials;
    int retries;
    int continues = 0;
    uint16_t txCount = 0;
    uint16_t rxCount;
    uint8_t txTmp[IV_PACKET_LENGTH_DATA];
    uint8_t rxTmp[IV_PACKET_LENGTH_DATA];
    uint16_t rxModule;
    uint16_t rxFunction;
    uint8_t type = 0;
    int err = IV_ERROR_SUCCESS;
    std::chrono::steady_clock::time_point last_write;

    switch ( context )
    {
    case IV_CONTEXT_TRANSMIT_ANSWER: // Einfacher Versand eines Paketes
    case IV_CONTEXT_TRANSMIT_ANSWER_ERROR: // Einfacher Versand eines Paketes
        // Temppuffer fuer Transmit-Paket mit Daten fuellen.
        txCount = count;
        std::copy_n( buffer, txCount, txTmp );
        retries = (contextValue == -1) ? ConnectionBase::defaultWriteTrials() : contextValue;
        receiveTrials = 0;
        m_timeboyBaseAddress = 0x70010000;
        break;

    case IV_CONTEXT_MASTER: // Ausführung (Versant und Empfang) eines Befehls
        // Temppuffer fuer Transmit-Paket, falls ein Request des Geraetes dazwischen kommt und
        // die selben Daten nochmals gesendet werden.
        txCount = count;
        std::copy_n( buffer, txCount, txTmp );
        retries = (contextValue == -1) ? ConnectionBase::defaultWriteTrials() : contextValue;
        receiveTrials = ConnectionBase::defaultReadTrials();
        m_timeboyBaseAddress = 0x70000000;
#ifdef _DEBUG
        m_previousModule = module;
        m_previousFunction = func;
#endif
        break;

    case IV_CONTEXT_SLAVE: // Empfang und Ausführung eines Befehls
        retries = (contextValue == -1) ? ConnectionBase::defaultReadTrials() : contextValue; 
        receiveTrials = 1;
        m_timeboyBaseAddress = 0x70010000;
        break;

    default:
        trace( "%s(%i):ERROR(1): %s() [ Unbekannter Context ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FAIL;
    }

    // Einen Versuch sicherstellen
    if ( retries <= 0 )
    {
        retries = 1;
    }

    for ( trials = 0; trials < retries && continues < 10; ++trials ) 
    {
        // Ist ein Kommunikationsbefehl abzuarbeiten?
        if ( context == IV_CONTEXT_MASTER || context == IV_CONTEXT_TRANSMIT_ANSWER || context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR ) 
        {
            uint8_t type = (context == IV_CONTEXT_MASTER) ? IV_PACKET_REQUEST : IV_PACKET_ANSWER;
            if ( context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR )
                type |= IV_PACKET_ERROR;

            // Befehlspaket senden
            if ( (err = writePacket( deviceNumberReceiver(), type, module, func, txTmp, txCount )) != IV_ERROR_SUCCESS ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ writePacket context: %i, error: %i ]\n", __FILE__, __LINE__, __func__, context, err );
                break;
            }

            // Kein lesen erforderlich?
            if ( context != IV_CONTEXT_MASTER )
            {
                break;
            }

            // Zeitpunkt des Paketversands halten.
            // -------------------------------------
            // Falls ein oder mehrere Zwischenbefehle kommen, muss nach Ablauf des Timeouts die Anfrage wiederholt werden.
            last_write = std::chrono::steady_clock::now();
        } 

receive_again:
        // Befehlspacket empfangen
        if ( (err = readPacket( deviceNumberSender(), type, rxModule, rxFunction, rxTmp, rxCount, receiveTrials )) != IV_ERROR_SUCCESS ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ readPacket error: %i ]\n", __FILE__, __LINE__, __func__, err );
            if ( err == -IV_ERROR_BREAK )
            {
                break; // Abbruch durch remoteCall-Back (Anwender)
            }

            if ( context == IV_CONTEXT_MASTER )
            {
                // Anzahl unbeantworteter Pakete fuer Statistik halten
                m_packetLostCounter++;
            }
            else if ( context == IV_CONTEXT_SLAVE )
            {
                switch ( err )
                {
                case IV_ERROR_CRC_PACKET: // Gesamt-CRC des Pakets ist falsch.
                    break;

                case IV_ERROR_PACKETLENGTH: // Gesamt-Laenge des Pakets ist falsch.
                    break;
                }
            }

            if ( err == -IV_ERROR_CHANNEL_OPEN_REMOTE ) // Remote hat die Verbindung geschlossen
            {
                break; 
            }

            continue;
        }

        // Packetbefehl auswerten
        switch ( type ) 
        {
        case (IV_PACKET_VERSION_4 | IV_PACKET_REQUEST):
            {
                uint8_t deviceAddress = setDeviceNumberReceiver( m_packetReceive[2] );

                // BUGFIX
                // Der Bootloader liefert als aktive Antwort auf IV_FUNCTION_BEGIN_SET_SEGMENT anstatt
                // IV_FUNCTION_BEGIN_SET_SEGMENT_OK ebenfalls den Wert IV_FUNCTION_BEGIN_SET_SEGMENT
                // Ist aktuell Warten auf diese Aktive Antwort?
                if ( context == IV_CONTEXT_SLAVE && module == 0x0001/*IV_MODULE_SYSTEM*/ && func == 0x0010/*IV_FUNCTION_BEGIN_SET_SEGMENT_OK*/ ) 
                {
                    // Korrektur des Funktionswertes, wenn nötig
                    rxFunction = func;
                }

                // Funktion der entsprechenden Modulklasse aufrufen.
                // Die Routinen senden die Antwort selber, nur Fehler werden nachfolgend fuer
                // alle gemeinsam abgehandelt.
                if ( (err = remoteCall( rxModule, rxFunction, rxTmp, rxCount )) != IV_ERROR_SUCCESS ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ remoteCall( %u, %u, %p, %u ) ]\n", __FILE__, __LINE__, __func__, rxModule, rxFunction, rxTmp, rxCount );

                    // Fehler als Antwort senden.
                    *(uint16_t *)rxTmp = (uint16_t)err;
                    if ( (err = writePacket( deviceNumberReceiver(), IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, rxTmp, sizeof( uint16_t ) )) != IV_ERROR_SUCCESS ) 
                    {
                        trace( "%s(%i):ERROR(1): %s() [ writePacket error: %i ]\n", __FILE__, __LINE__, __func__, err );
                        setDeviceNumberReceiver( deviceAddress );
                        return err;
                    }
                }

                // Auf Vorgaenger wechseln.
                setDeviceNumberReceiver( deviceAddress );

                // Aktive Antwort OK?
                if ( (context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func) ) 
                {
                    // Datenkopie uebernehmen
                    memcpy( buffer, rxTmp, rxCount );
                    count = rxCount;
                    err = IV_ERROR_SUCCESS;
                } 
                else if ( (context == IV_CONTEXT_MASTER) || (context == IV_CONTEXT_SLAVE && module != IV_MODULE_BASE) )
                {
                    // Bei einem Zwischenbefehl muss sichergestellt werden, dass nach Ablauf des Timeouts eine erneute Anfrage gesendet oder Abgebrochen wird.
                    if ( context == IV_CONTEXT_MASTER 
                        && std::chrono::steady_clock::now() > last_write + std::chrono::milliseconds( connection().readTimeout() ) )
                    {
                        trace( "%s(%i):INFO(1): %s() [ timeout while request ]\n", __FILE__, __LINE__, __func__ );
                        continue;
                    }

                    // Wurde Befehl gesendet und die Antwort ist noch zu lesen,
                    // oder wird noch auf die richtige aktive Antwort gewartet.
                    trace( "%s(%i):INFO(1): %s() [ Receive again ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }
                break;
            }

        case (IV_PACKET_VERSION_4 | IV_PACKET_REQUEST | IV_PACKET_ERROR):
            {

                if ( rxCount != sizeof( uint16_t ) ) 
                {
                    // Falsche Parameteranzahl
                    trace( "%s(%i):ERROR(1): %s() [ Fehlerrequest: falsche Parameteranzahl ]\n", __FILE__, __LINE__, __func__ );
                }

                // Fehlercode zuweisen
                uint16_t e = *(uint16_t *)rxTmp;

                // Fehler beantworten.
                if ( (err = writePacket( m_packetReceive[2], IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, rxTmp, 0 )) != IV_ERROR_SUCCESS ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ writePacket error: %u ]\n", __FILE__, __LINE__, __func__, err );
                    return err;
                }

                // Aktive Antwort OK?
                if ( (context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func) ) 
                {
                    // Nur als Aktive Antwort sinnvoll!!!
                    return e;
                } 
                else 
                {
                    // Unerwartete Anfrage erhalten
                    trace( "%s(%i):ERROR(1): %s() [ Fehlerrequest: Unerwartete Anfrage ]\n", __FILE__, __LINE__, __func__ );
                }
                break;
            }

        case (IV_PACKET_VERSION_4 | IV_PACKET_ANSWER):

            // Antwort auf zuvor gesendeten Befehl?
            if ( context == IV_CONTEXT_MASTER ) 
            {
                // Antwort von gefordertem Absender?
                if ( IV_PACKET_DEVICEID_REMOTE(m_packetReceive) != deviceNumberReceiver() ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falscher Absender ]\n", __FILE__, __LINE__, __func__ );
                    return -IV_ERROR_ADDRESS;
                }

                // Antwort auf aktuelles Modul/Function
                if ( rxModule != module || rxFunction != func ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falsche Module/Function ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }

                // Daten in Buffer der Applikation kopieren
                memcpy( buffer, rxTmp, rxCount );
                count = rxCount;
                err = IV_ERROR_SUCCESS;
            } 
            else 
            {
                // Unerwartete Antwort erhalten
                trace( "%s(%i):ERROR(1): %s() [ Unerwartete Antwort ]\n", __FILE__, __LINE__, __func__ );
            }
            break;

        case (IV_PACKET_VERSION_4 | IV_PACKET_ERROR):
        case (IV_PACKET_VERSION_4 | IV_PACKET_ANSWER | IV_PACKET_ERROR):

            // Antwort auf zuvor gesendeten Befehl?
            if ( context == IV_CONTEXT_MASTER ) 
            {
                // Antwort von gefordertem Absender?
                if ( IV_PACKET_DEVICEID_REMOTE(m_packetReceive) != deviceNumberReceiver() ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falscher Absender ]\n", __FILE__, __LINE__, __func__ );
                    return -IV_ERROR_ADDRESS;
                }

                // Wenn der Gesamt-CRC fehlerhaft ist oder die Entschluesselung nicht durchgefuehrt werden konnte,
                // dann wird in der Antwort Modul und Funktion auf -1 gesetzt.
                if ( rxModule == 0xFFFF && rxFunction == 0xFFFF )
                {

                }
                else if ( rxModule != module || rxFunction != func )
                { 
                    // Antwort auf aktuelles Modul/Function
                    trace( "%s(%i):ERROR(1): %s() [ Falsche Module/Function ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }

                // Fehlercode uebernehmen
                err = *(uint16_t *)rxTmp;

                // Pruefen ob Modulfehler vorliegt und die Anfrage erneut gesendet werden muss.
                // PROT_MODULE_ERROR_CRC                    (0x0700) // Gesamt CRC von Paket ist falsch
                // PROT_MODULE_ERROR_CONTROL                (0x0800) // Zu wenig Steuerinformationen
                // PROT_MODULE_ERROR_PROTOCOL               (0x0900) // Protokollversion passt nicht
                // PROT_MODULE_ERROR_NO_QUERY               (0x0A00) // Anfrage erwartet, Antwort erhalten
                // PROT_MODULE_ERROR_BUSY                   (0x0B00) // Antwort erwartet, Anfrage erhalten
                // PROT_MODULE_ERROR_SIZE                   (0x0C00) // Paketgroesse ueberschreitet reservierte Puffergroesse
                // PROT_MODULE_ERROR_DECRYPT                (0x0D00) // Entschluesselung des Paketinhaltes bring Fehler.
                // PROT_MODULE_ERROR_CRYPTED                (0x0E00) // Geraet arbeitet verschluesselt und Anfrage war unverschluesselt.

                if ( err == 0x0700 || err == 0x0B00 ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Modulfehler %04X, wiederholtes senden. ]\n", __FILE__, __LINE__, __func__, err );
                    trials--; // Nochmaliges senden des Pakets
                    continues++; // Abbruch nach n-maligem wiederholen.
                    continue;
                }
            } 
            else 
            {
                // Unerwartete Antwort erhalten
                trace( "%s(%i):ERROR(1): %s() [ Unerwartete Antwort. context: %u, modul: %u, func: %u ]\n", __FILE__, __LINE__, __func__, context, module, func );
            }
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Packettyp: %02X ]\n", __FILE__, __LINE__, __func__, type );
            goto receive_again;
        }

        // Schleife nach erfolgreicher Abarbeitung verlassen.
        break;
    }

    if ( context == IV_CONTEXT_MASTER ) 
    {
        double tm = static_cast<double>( (double)m_clockWrite + (double)m_clockRead ) / (double)CLOCKS_PER_SEC;
        //double clockTx = (double)_clockWrite / (double)CLOCKS_PER_SEC;
        //double clockRx = (double)_clockRead / (double)CLOCKS_PER_SEC;
        m_qs = (double)((double)m_clocksWrite + (double)m_clocksRead) / (double)CLOCKS_PER_SEC;;

        if ( !(m_packetReadCounter % 10) ) 
        {
            double saveVal1 = (connection().sizeWritten()) ? ((double)(connection().sizeWritten() - m_packetWritePayloadSize) / (double)connection().sizeWritten()) * 100 : 0;
            double saveVal2 = (connection().sizeRead()) ? ((double)(connection().sizeRead() - m_packetReadPayloadSize) / (double)connection().sizeRead()) * 100 : 0;
            double saveVal3 = ((connection().sizeWritten() + connection().sizeRead()) == 0) ? 0 : double((connection().sizeWritten() - m_packetWritePayloadSize) + (connection().sizeRead() - m_packetReadPayloadSize)) / (double)(connection().sizeWritten() + connection().sizeRead()) * 100;
            double saveVal4 = (m_packetReadCounter == 0) ? 0 : (double)m_qs / (double)m_packetReadCounter;

            //Trace("%s(%i):INFO(1): %s [ QS txBuf(%i ; %i / %i) [sendezeit+empfangszeit: %f][ empfangszeit: %f]{gesamt(empfangszeit/packetanzahl): %f} ]\n", __FILE__, __LINE__, __func__, (receivedCount+masked), (double)((double)ct2 - (double)m_ct) / (double)CLOCKS_PER_SEC, tm, (double)qs / (double)count);
            trace( "%s(%i):INFO(4): %s() [ txBuf: %i; %i/%i; %.2f; rxBuf: %i, %i; %i/%i; %.2f; s(txBuf/rxBuf): %i/%i; %.2f; t(txBuf+rxBuf): %f; t(rxBuf): %f; s(t(rxBuf+txBuf)/pa): %f; ]\n", __FILE__, __LINE__, __func__, 
                m_packetWriteCounter, connection().sizeWritten(), m_packetWritePayloadSize, saveVal1, 
                m_packetReadCounter, m_packetLostCounter, connection().sizeRead(), m_packetReadPayloadSize, saveVal2, 
                connection().sizeWritten() + connection().sizeRead(), m_packetWritePayloadSize + m_packetReadPayloadSize, saveVal3,
                tm, (double)m_clockRead / (double)CLOCKS_PER_SEC, saveVal4);
            /*
            double val1 = (m_transmittedSize == 0) ? 0 : ((double)(m_transmittedSize - m_transmittedSizeData) / (double)m_transmittedSize) * 100;
            double val2 = (m_receivedSize == 0) ? 0 : ((double)(m_receivedSize - _packetReadPayloadSize) / (double)m_receivedSize) * 100;
            double val3 = ((m_transmittedSize + m_receivedSize) == 0) ? 0 : double((m_transmittedSize - m_transmittedSizeData) + (m_receivedSize - _packetReadPayloadSize)) / (double)(m_transmittedSize + m_receivedSize) * 100;
            double val4 = (_packetReadCounter == 0) ? 0 : (double)_qs / (double)_packetReadCounter;

            Trace("%s(%i):INFO(4): %s() [ transmit { time %f, packetcount %i; size %i / datasize %i; overhead in prozent %.2f%% } ]\n", __FILE__, __LINE__, __func__, clockTx, _packetWriteCounter, m_transmittedSize, m_transmittedSizeData, val1);
            Trace("%s(%i):INFO(4): %s() [ receive { time %f, packetcount %i, packetcount unbeantwortet %i; size %i / datasize %i; overhead %.2f%% } ]\n", __FILE__, __LINE__, __func__, clockRx, _packetReadCounter, _packetLostCounter, m_receivedSize, _packetReadPayloadSize, val2);
            Trace("%s(%i):INFO(4): %s() [ summe(transmit + receive) { size %i (%.3fkByte) / datasize %i (%.3fkByte); overhead %.2f%% } ]\n", __FILE__, __LINE__, __func__, m_transmittedSize + m_receivedSize, (double)(m_transmittedSize + m_receivedSize) / (double)1024, m_transmittedSizeData + _packetReadPayloadSize, (double)(m_transmittedSizeData + _packetReadPayloadSize) / (double)1024, val3);
            Trace("%s(%i):INFO(4): %s() [ time summe(transmit + receive) { %f; durchschnittszeit pro einzelbefehl: %f } ]\n", __FILE__, __LINE__, __func__, tm, val4);
            */
        }
    }

    // Wurde die maximale Anzahl moeglicher Versuche erreicht?
    if ( trials == retries ) 
    {
        trace( "%s(%i):ERROR(1): %s() [ Maximalanzahl der Versuche erreicht. ]\n", __FILE__, __LINE__, __func__ );

        // Paketzaehler erhoehen!
        m_packetNumber++;

        //		return -IV_ERROR_NO_REQUEST; // Die aktive Antwort kam nicht nach der vorgegebenen Dauer
    }

    return err;
}

//#########################################################
//####################### ERROR ###########################
//#########################################################

// ##################################################################
/**
* @brief Die gesamte Fehlerbehandlung zuruecksetzen.
*/
void Datafox::Communication::ProtocolBase::resetError()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Fehlernummern loeschen
    m_errorCommunicationNumber = 0;
    // Puffer die in Gebrauch waren, loeschen.
    if ( m_errorCommunicationText.size() )
        m_errorCommunicationText.clear();

}


// ##################################################################
/**
* @brief Ermittlung des Fehlertextes zum Fehlercode fuer Kommunikationsfehler.
*/
void Datafox::Communication::ProtocolBase::errorCommunication( uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %p )\n", __FILE__, __LINE__, __func__, error, text.c_str() );

    switch ( error )
    {
    case IV_ERROR_SUCCESS:
        text = boost::str( boost::format( "( %i ): Kein Fehler" ) % error );
        break;

    case IV_ERROR_CHANNEL_OPEN:
        text = boost::str( boost::format( "( %i ): Schnittstelle geschlossen oder konnte nicht geöffnet werden." ) % error );
        break;

    case IV_ERROR_UNREACHABLE:
        text = boost::str( boost::format( "( %i ): Gerät antwortet nicht auf Anfrage." ) % error );
        break;

    case IV_ERROR_TRANSMIT:
        text = boost::str( boost::format( "( %i ): Es konnten keine Daten in die Schnittstelle geschrieben werden." ) % error );
        break;

    //case IV_ERROR_RECEIVE:
        //text = boost::str( boost::format( "( %i ): Es konnten keine Daten von der Schnittstelle gelesen werden." ) % error );
        //break;

    case IV_ERROR_FILE_OPEN:
        text = boost::str( boost::format( "( %i ): Datei konnte nicht geöffnet werden." ) % error );
        break;

    case IV_ERROR_CRC_PACKET:
        text = boost::str( boost::format( "( %i ): Prüfsumme (CRC) von Protokollpaket stimmt nicht." ) % error );
        break;

    case IV_ERROR_ADDRESS:
        text = boost::str( boost::format( "( %i ): Die Adressen (Sender/Empfänger) des Protokollpaketes sind fehlerhaft." ) % error );
        break;

    case IV_ERROR_COMMAND:
        text = boost::str( boost::format( "( %i ): Protokollbefehl der vierten Schicht ist falsch oder unbekannt." ) % error );
        break;

    //case IV_ERROR_PACKETLENGTH:
        //text = boost::str( boost::format( "( %i ): Die angegebene Paketlaenge stimmt nicht ueberein." ) % error );
        //break;

    case IV_ERROR_BREAK:
        text = boost::str( boost::format( "( %i ): Die Kommunikation wurde durch den Anwender (CallBack-Funktion) abgebrochen." ) % error );
        break;

    case IV_ERROR_FUNCTION_UNAVAILABLE:
        text = boost::str( boost::format( "( %i ): Funktion steht im vorliegenden Fall nicht zur Verfuegung." ) % error );
        break;

    case IV_ERROR_NEED_PASSWORD:
        text = boost::str( boost::format( "( %i ): Das angegebene Passwort stimmt nicht mit dem benötigten überein. Es wird ein Passwort gefordert." ) % error );
        break;

    case IV_ERROR_PASSWORD_LOCK:
        text = boost::str( boost::format( "( %i ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt." ) % error );
        break;

    case IV_ERROR_NEED_USER_PASSWORD:
        text = boost::str( boost::format( "( %i ): Fuer die verschluesselte Kommunikation mit dem Geraet wird ein Benutzer-Passwort benoetigt." ) % error );
        break;
    }

    if ( text.empty() )
    {
        text = boost::str( boost::format( "( %i ): Nicht abgefangener Kommunikationsfehler!" ) % error );
    }

    // Letzte Abfrage halten.
    m_errorCommunicationNumber = error;
    m_errorCommunicationText = text;
}

