/**
* @file  filedevicetext.cpp
* @brief Implementierung der Klasse FileDeviceText.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_device_text.h"
#include <ctype.h>
#include <string>
#include <cstring>
#include <fstream>

FileDeviceText::FileDeviceText()
{
    // Initialisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::text );
}

FileDeviceText::~FileDeviceText()
{
    // Speicher freigeben
    resetData();
}

void FileDeviceText::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Member Reseten
    bufferDefaultDFL_.clear();
}

bool FileDeviceText::load( const std::string &filePath )
{
    // Lesen der Hexdatei
    if ( FileIntelHex::load( filePath ) )
    {
        // Member durch gelesene Daten initialisieren und pruefen
        if ( m_buffer.size() )
        {
            TEXTDATA *td = reinterpret_cast<TEXTDATA *>(m_buffer.data());
            // CRC32 pruefen
            if ( td->crc != GetCRC32( &m_buffer[td->ptrIDTable], td->cnt * sizeof(unsigned short) + td->size ) )
            {
                // Fehler zuweisen
                setErrorState( Error_state::crc_fail );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
        }
        else
        {
            // Fehler zuweisen
            setErrorState( Error_state::no_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileDeviceText::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ )
{
    return FileIntelHex::save( filePath, appendToFile, appendEndLine );
}

char *FileDeviceText::DefaultDFL()
{
    return reinterpret_cast<char *>(bufferDefaultDFL_.data());
}

// Import einer Textdatei mit folgendem Zeilenaufbau
// #define TEXT_SERIAL_NUMBER 10  ///< 1.2.3.1.1. Menuetext 1 fuer Seriennummer "serial number"
// ---------------------------**-------**************************************--*************-
int FileDeviceText::Import( const std::string &filePath )
{
    // Lesen der Header oder Default Datei
    std::ifstream fs( filePath );
    std::string data;
    std::string dataTmp;
    unsigned short id;
    unsigned int flags;
    unsigned int numLine = 0;
    char *format;
    char *text;
    // Ist Headerfile?
    int fromHeader = (filePath[filePath.size() - 1] == 'h');

    // Datenspeicher freigeben.
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fs.is_open() == false )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( Error_state::file_open );
        setErrorDetail( str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        // Daten ermitteln
        unsigned short maxID = 0;
        size_t size = 0;
        unsigned int sizeDefaultDFL = 0;
        int res;

        // Analyse.
        // Maximalanzahl Texte, Gesamtgroesse Speicher
        while( fs.eof() == false && (errorState() == Error_state::none) )
        {
            if ( std::getline( fs, data ) )
            {
                // Zeilennummer
                numLine++;

                res = fromHeader ? ParseLineFromHeader( &data[0], id, &format, &text, flags ) : ParseLineFromDefault( &data[0], id, &format, &text, flags );
                switch ( res )
                {
                case 0: // Keine Auszuwertende Datenzeile
                    break;

                case 1: // Ausgewertete Datenzeile
                    // Dummies sind mit -1 (65535) angegeben
                    if ( (id != 65535) && id > maxID )
                        maxID = id;

                    //sprintf(text, "%c%u", fromHeader ? 'D' : 'U', id);

                    // Groesse fuer Default.dfl mitberechnen
                    dataTmp = boost::str( boost::format( "%u\t%s\t\"%s\"\r\n" ) % id % format % text );
                    sizeDefaultDFL += static_cast<unsigned int>( dataTmp.length() );

                    // Steuerzeichen binaer codieren
                    ConvertControlCharacters( &text );

                    // Nur Texte mit mindestens einem Zeichen werden auch uebertragen.
                    if ( (id != 65535) && strlen( text ) )
                    { 
                        // Zu lange Texte abfangen, Maximal 130 Zeichen erlaubt.
                        if ( strlen( text ) >= 130 ) {
                            text[130] = 0;
                        }

                        size += strlen(text) + 1;
                    }
                    break;

                default: // Fehler bei Durchfuehrung
                    // Fehler zuweisen
                    setErrorState( Error_state::line_format );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return 0;
                }
            }
        }
        // Auf Dateianfang wechseln
        fs.clear();
        fs.seekg( 0 );

        // Speicher fuer Textdaten gesamt allocieren
        m_buffer.resize( sizeof(TEXTDATA) + (maxID + 1) * sizeof(unsigned short) + size );
        std::fill( m_buffer.begin(), m_buffer.end(), '\0' );

        // Speicher fuer Default.dfl allocieren
        bufferDefaultDFL_.resize( sizeDefaultDFL + 1 );
        std::fill( bufferDefaultDFL_.begin(), bufferDefaultDFL_.end(), '\0' );
        sizeDefaultDFL = 0;

        // Werte einlesen und aufbereiten
        TEXTDATA *tn = reinterpret_cast<TEXTDATA *>(m_buffer.data());
        tn->version = 1;
        strcpy(tn->name, "Textdaten");
        tn->type = 2;
        tn->cnt = maxID + 1;
        tn->size = (unsigned int)size;
        tn->ptrIDTable = sizeof(TEXTDATA);
        tn->ptrTextdata = tn->ptrIDTable + tn->cnt*sizeof(unsigned short);
        // Zeiger auf ersten freien Text
        unsigned short ptrText = (unsigned short)tn->ptrTextdata;

        while( fs.eof() == false && errorState() == Error_state::none )
        {
            if ( std::getline( fs, data) )
            {
                res = fromHeader ? ParseLineFromHeader( &data[0], id, &format, &text, flags) : ParseLineFromDefault( &data[0], id, &format, &text, flags );
                switch ( res )
                {
                case 0: // Keine auszuwertende Datenzeile
                    break;

                case 1: // Ausgewertete Datenzeile
                    // Dummies sind mit -1 (65535) angegeben
                    if (id != 65535) {

                        //sprintf(text, "%c%u", fromHeader ? 'D' : 'U', id);

                        // Daten gleich als Default.dfl aufbereiten
                        dataTmp = boost::str( boost::format( "%u\t%s\t\"%s\"\r\n" ) % id % format % text );
                        bufferDefaultDFL_.insert( bufferDefaultDFL_.begin() + sizeDefaultDFL, dataTmp.begin(), dataTmp.end() );
                        sizeDefaultDFL += static_cast<unsigned int>( dataTmp.length() );
                        //sizeDefaultDFL += sprintf( &bufferDefaultDFL_[sizeDefaultDFL], "%u\t%s\t\"%s\"\r\n", id, format, text );

                        // Steuerzeichen binaer codieren
                        ConvertControlCharacters(&text);

                        if (strlen(text)) {

                            // Zu lange Texte abfangen, Maximal 130 Zeichen erlaubt.
                            if (strlen(text) >= 130) {
                                text[130] = 0;
                            }

                            // Texte mit mindestens einem Zeichen Uebernehmen
                            *(unsigned short*)&m_buffer[tn->ptrIDTable + id*sizeof(unsigned short)] = ptrText;
                            strcpy((char*)&m_buffer[ptrText], text);
                            ptrText += (unsigned short)strlen(text) + 1;

                        } else {

                            // Text als nicht vorhanden uebernehmen
                            *(unsigned short*)&m_buffer[tn->ptrIDTable + id*sizeof(unsigned short)] = 0;
                        }

                    }
                    break;

                default: // Fehler bei Durchfuehrung
                    // Fehler zuweisen
                    setErrorState( Error_state::line_format );
                    setErrorDetail( str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return 0;
                }
            }
        }

        // CRC berechnen
        tn->crc = GetCRC32(&m_buffer[tn->ptrIDTable], tn->cnt*sizeof(unsigned short) + tn->size);

        if ( errorState() == Error_state::none )
        {
            // Die Datei wird fuer einen eventuellen Import in den Firmwarecode als Binaerdatei gespeichert.
            std::string fName( filePath );
            fName.append( ".text.bin" );
            std::ofstream binOut( fName, std::ios::binary );

            // Datei als Binär rausschreiben
            if ( binOut.is_open() == true )
            {
                binOut.write( reinterpret_cast<char *>(m_buffer.data()), m_buffer.size() );
                binOut.close();
            }
        }
    }

    return (errorState() == Error_state::none);
}

// Die Datenzeile hinter data wird manipuliert!!!
int FileDeviceText::ParseLineFromHeader(char *data, unsigned short &id, char **format, char **text, unsigned int &flags)
{
    // Rueckgabeparameter initialisieren
    id = 65535;
    *format = 0;
    *text = 0;
    flags = 0;

    // Per Definition muss folgendes zutreffen
    // Zeile Beginnt mit #define, beinhaltet ///< und endet auf "\0
    if (data != strstr(data, "#define") )//|| strstr(data, "///<") == NULL || data[strlen(data) - 1] != '\"') 
    {
        // Keine gueltige Zeile fuer Textdaten
        return 0;
    }

    // suche nach Bloecken (id, format, text)
    char *pos, *textBegin;
    textBegin = NULL;

    // Aktuelle Position ist hinter #define 
    pos = &data[8];
    // weiterlaufen bis whitespace
    while(!isspace((unsigned char)*pos)) pos++;
    // weiterlaufen bis nach allen whitespaces
    while(isspace((unsigned char)*pos)) pos++;
    // ID extrahieren
    textBegin = pos; // Beginn der Zahl merken
    // weiterlaufen bis whitespace
    while(!isspace((unsigned char)*pos)) pos++;
    *pos = 0; // Terminierung hinter zahl
    id = (unsigned short)atoi(textBegin);
    // hinter dieser muss nun format folgen
    pos++;
    pos = strstr(pos, "///<");
    if (*pos == 0) {
        // Formatfehler
#ifdef _MSC_VER
        //		TRACE("ERROR: Zeile beinhaltet \"///<\" an falscher Position.\n");
#endif
        return -1;
    }
    pos += 4;
    // weiterlaufen bis nach allen whitespaces
    while(*pos && isspace((unsigned char)*pos)) pos++;
    *format = pos; // Beginn als Format halten
    if (!isdigit((unsigned char)*pos)) {
        // Formatfehler
#ifdef _MSC_VER
        //		TRACE("ERROR: Erstes Zeichen nach \"///<\" ist keine Ziffer.\n");
#endif
        return -1;
    }

    // Textbestand steht hinten
    pos = &pos[strlen(pos) - 1];
    *pos = 0; // Anfuehrungzeichen mit nullterminierung ueberschreiben
    pos--;
    // weiterlaufen bis " ohne \ davor
    while(*pos && (*pos != '"' || (*pos == '"' && *(pos-1) == '\\'))) pos--;
    if (*pos == 0) {
        // Formatfehler
#ifdef _MSC_VER
        //		TRACE("ERROR: Zeile endet nicht mit gueltigem String.\n");
#endif
        return -1;
    }
    *pos = 0; // Anfuehrungzeichen mit Nullterminierung ueberschreiben
    // Text extrahieren
    *text = pos + 1;

    // folgende whitespaces loeschen
    pos--;
    while(*pos && isspace((unsigned char)*pos)) { 
        *pos = 0; 
        pos--; 
    }

#ifdef _MSC_VER
    //	TRACE("ID: [%i], Format: [%s], Text: [%s]\n", id, *format, *text);
#endif

    return 1;
}

// Die Datenzeile hinter data wird manipuliert!!!
int FileDeviceText::ParseLineFromDefault(char *data, unsigned short &id, char **format, char **text, unsigned int &flags)
{
    // Rueckgabeparameter initialisieren
    id = 65535;
    *format = 0;
    *text = 0;
    flags = 0;

    // Per Definition muss folgendes zutreffen
    // Zeile Beginnt mit Ziffer und endet auf "\0
    if (!isdigit((unsigned char)*data) || data[strlen(data) - 1] != '\"')
    {
        // Keine gueltige Zeile fuer Textdaten
        return 0;
    }

    // suche nach Bloecken (id, format, text)
    char *pos, *textBegin;
    textBegin = NULL;

    // Aktuelle Position
    pos = data;
    // ID extrahieren
    textBegin = pos; // Beginn der Zahl merken
    // weiterlaufen bis whitespace
    while(!isspace((unsigned char)*pos)) pos++;
    *pos = 0; // Terminierung hinter zahl
    id = (unsigned short)atoi(textBegin);
    // hinter dieser muss nun format folgen
    pos++;
    // weiterlaufen bis nach allen whitespaces
    while(*pos && isspace((unsigned char)*pos)) pos++;
    *format = pos; // Beginn als Format halten
    if (!isdigit((unsigned char)*pos)) {
        // Formatfehler
#ifdef _MSC_VER
        //		TRACE("ERROR: Erstes Zeichen des Formats ist keine Ziffer.\n");
#endif
        return -1;
    }

    // Textbestand steht hinten
    pos = &pos[strlen(pos) - 1];
    *pos = 0; // Anfuehrungzeichen mit nullterminierung ueberschreiben
    pos--;
    // weiterlaufen bis " ohne \ davor
    while(*pos && (*pos != '"' || (*pos == '"' && *(pos-1) == '\\'))) pos--;
    if (*pos == 0) {
        // Formatfehler
        return -1;
    }
    *pos = 0; // Anfuehrungzeichen mit Nullterminierung ueberschreiben
    // Text extrahieren
    *text = pos + 1;

    // folgende whitespaces loeschen
    pos--;
    while(*pos && isspace((unsigned char)*pos)) { 
        *pos = 0; 
        pos--; 
    }

#ifdef _MSC_VER
    //	TRACE("ID: [%i], Format: [%s], Text: [%s]\n", id, *format, *text);
#endif

    return 1;
}

void FileDeviceText::ConvertControlCharacters(char **text)
{	
    // Kontrollsequenz vorhanden?
    if (strstr(*text, "\\r") != NULL || 
        strstr(*text, "\\t") != NULL) {

            // Replace von \r und \t
            char *src, *dest;
            dest = src = *text;

            while (*src) {

                if (*src == '\\') {

                    if (*(src+1) == 'r') {

                        *src = 0; src++; *src = 0;
                        // Steuerzeichen \r
                        *dest = '\r';
                    } else if (*(src+1) == 't') {

                        *src = 0; src++; *src = 0;
                        // Steuerzeichen \t
                        *dest = '\t';
                    } else {

                        // Kein Steuerzeichen
                        *dest = *src;
                        if (src > dest) *src = 0;
                    }
                } else {

                    *dest = *src;
                    if (src > dest) *src = 0;
                }

                dest++;
                src++;
            }
    }
}

bool FileDeviceText::loadBin( std::ifstream &fs, std::size_t size )
{
    // Umgebung zuruecksetzen
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( size == 0 )
    {
        // Fehler zuweisen
        setErrorState( Error_state::no_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else if ( fs.is_open() == false )
    {
        // Stream ist nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        try
        {
            m_buffer.resize( size );
            std::fill( m_buffer.begin(), m_buffer.end(), 0 );
            fs.read( reinterpret_cast<char *>( m_buffer.data() ), m_buffer.size() );

            // Member durch gelesene Daten initialisieren und pruefen
            TEXTDATA *td = reinterpret_cast<TEXTDATA *>(m_buffer.data());
            // CRC32 pruefen
            if ( td->crc != GetCRC32( &m_buffer[td->ptrIDTable], td->cnt * sizeof(unsigned short) + td->size ) )
            {
                // Fehler zuweisen
                setErrorState( Error_state::crc_fail );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_read );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileDeviceText::saveBin( std::ofstream &fs )
{
    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( fs.is_open() == false )
    {
        // Stream ist nicht geoffnet.
        setErrorState( Error_state::file_open );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }
    else
    {
        try
        {
            // Textdaten uebertragen.
            fs.write( reinterpret_cast<char *>(m_buffer.data()), m_buffer.size() );
        }
        catch ( ... )
        {
            // Fehler zuweisen
            setErrorState( Error_state::file_write );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }


    return (errorState() == Error_state::none);
}
