/**
* @file  file_touch_keys.cpp
* @brief Implementierung der Klasse FileTouchKeys.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_touch_keys.h"
#include "common/software_version.h"

FileTouchKeys::FileTouchKeys()
{
    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::touch_key );

    // Initialisieren
    resetData();

    // Fehlerobjekt zuruecksetzen
    resetError();

}

FileTouchKeys::~FileTouchKeys()
{
    // Speicher freigeben.
    resetData();
}

void FileTouchKeys::resetData()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Member initialisieren
    m_touchKeysNode_v1 = nullptr;
    m_touchKeysNode_v2 = nullptr;
    for ( auto &key : m_touchKeys_v1 )
    {
        key = nullptr;
    }
    for ( auto &key : m_touchKeys_v2 )
    {
        key = nullptr;
    }
}

bool FileTouchKeys::load( const std::string &filePath )
{
    // Lesen der Hexdatei
    if ( FileIntelHex::load( filePath ) )
    {
        // Member durch gelesene Daten initialisieren und pruefen
        if ( !m_buffer.empty() )
        {
            // Hilfszeiger auf Struktur
            uint8_t version = m_buffer[0];
            uint8_t count = 0;
            switch ( version )
            {
                case 1:
                {
                    m_touchKeysNode_v1 = (TOUCHKEYNODE_V1 *)&m_buffer[0];
                    m_touchKeysNode_v2 = nullptr;

                    // Korrigiere DeviceType
                    // In der Version 1 können nur EVO 4.3 konfiguriert worden sein. Eine fehlende Definition des DeviceType wird angepasst.
                    if ( m_touchKeysNode_v1->deviceType != svt_evo_4_3 )
                    {
                        m_touchKeysNode_v1->deviceType = svt_evo_4_3;
                    }

                    count = m_touchKeysNode_v1->cnt;
                    break;
                }
                case 2:
                {
                    m_touchKeysNode_v1 = nullptr;
                    m_touchKeysNode_v2 = (TOUCHKEYNODE_V2 *)&m_buffer[0];

                    // Korrigiere DeviceType
                    // DeviceType 0 weist auf ein EVO 4.3 hin. Der Typ wird angepasst.
                    if ( m_touchKeysNode_v2->deviceType == 0 )
                    {
                        m_touchKeysNode_v2->deviceType = svt_evo_4_3;
                    }

                    count = m_touchKeysNode_v2->cnt;
                    break;
                }

                // Eine Version > 2 wird zur Zeit nicht verarbeitet und ist daher ungültig.
                default:
                {
                    return false;
                }
            }

            // Touchkeys prüfen
            // Zeigerarray und Touchkeystrukturzeiger initialisieren
            if ( count > max_touch_keys )
            {
                // Fehler zuweisen
                setErrorState( Error_state::format );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            }
            else
            {
                for ( std::size_t i = 0; i < count; ++i )
                {
                    if ( version == 1 )
                    {
                        try
                        {
                            m_touchKeys_v1[i] = (TOUCHKEY_V1 *)&m_buffer.at( m_touchKeysNode_v1->ptrTouchkeyArray + count * sizeof( uint32_t ) + i * sizeof( TOUCHKEY_V1 ) );
                        }
                        catch ( std::out_of_range e )
                        {
                            setErrorState( Error_state::format );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                            break;
                        }
                    }
                    else
                    {
                        try
                        {
                            m_touchKeys_v2[i] = (TOUCHKEY_V2 *)&m_buffer.at( m_touchKeysNode_v2->ptrTouchkeyArray + count * sizeof( uint32_t ) + i * sizeof( TOUCHKEY_V2 ) );
                        }
                        catch ( std::out_of_range e )
                        {
                            setErrorState( Error_state::format );
                            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                            break;
                        }
                    }
                }
            }
        }
        else
        {
            // Fehler zuweisen
            setErrorState( Error_state::no_data );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        }
    }

    return (errorState() == Error_state::none);
}

bool FileTouchKeys::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ )
{
    return FileIntelHex::save( filePath, appendToFile, appendEndLine );
}

void FileTouchKeys::init( uint8_t count, uint8_t deviceType )
{
    // Speicher reservieren
    m_buffer.clear();
    if ( deviceType == svt_evo_4_3 )
    {
        m_buffer.resize( sizeof( TOUCHKEYNODE_V1 ) + count * sizeof( unsigned int ) + count * sizeof( TOUCHKEY_V1 ) );
    
        // Hilfszeiger auf Struktur
        m_touchKeysNode_v1 = (TOUCHKEYNODE_V1 *)&m_buffer[0];
        m_touchKeysNode_v2 = nullptr;

        // Grundstruktur aufsetzen
        m_touchKeysNode_v1->version = 1;
        m_touchKeysNode_v1->cnt = count;
        m_touchKeysNode_v1->deviceType = deviceType;
        m_touchKeysNode_v1->temp = 0;
        m_touchKeysNode_v1->ptrTouchkeyArray = sizeof( TOUCHKEYNODE_V1 );

        // Zeigerarray und Touchkeysstrukturzeiger initialisieren
        uint32_t i = m_touchKeysNode_v1->cnt;
        for ( uint32_t j = 0; j < i; ++j )
        {
            // Strukturoffset auf Touchkeystruktur
            *(uint32_t *)&m_buffer[m_touchKeysNode_v1->ptrTouchkeyArray + j * sizeof( uint32_t )] =
                m_touchKeysNode_v1->ptrTouchkeyArray + i * sizeof( uint32_t ) + j * sizeof( TOUCHKEY_V1 );

            m_touchKeys_v1[j] = (TOUCHKEY_V1 *)&m_buffer[m_touchKeysNode_v1->ptrTouchkeyArray + i * sizeof( uint32_t ) + j * sizeof( TOUCHKEY_V1 )];
            m_touchKeys_v2[j] = nullptr;
        }
    }
    else
    {
        m_buffer.resize( sizeof( TOUCHKEYNODE_V2 ) + count * sizeof( unsigned int ) + count * sizeof( TOUCHKEY_V2 ) );
        
        // Hilfszeiger auf Struktur
        m_touchKeysNode_v1 = nullptr;
        m_touchKeysNode_v2 = (TOUCHKEYNODE_V2 *)&m_buffer[0];

        // Grundstruktur aufsetzen
        m_touchKeysNode_v2->version = 2;
        m_touchKeysNode_v2->cnt = count;
        m_touchKeysNode_v2->deviceType = deviceType;
        m_touchKeysNode_v2->dpiX = 300;
        m_touchKeysNode_v2->dpiY = 300;
        m_touchKeysNode_v2->ptrTouchkeyArray = sizeof( TOUCHKEYNODE_V2 );
 
        // Zeigerarray und Touchkeysstrukturzeiger initialisieren
        uint32_t i = m_touchKeysNode_v2->cnt;
        for ( uint32_t j = 0; j < i; ++j )
        {
            // Strukturoffset auf Touchkeystruktur
            *(uint32_t *)&m_buffer[m_touchKeysNode_v2->ptrTouchkeyArray + j * sizeof( uint32_t )] =
                m_touchKeysNode_v2->ptrTouchkeyArray + i * sizeof( uint32_t ) + j * sizeof( TOUCHKEY_V2 );

            m_touchKeys_v1[j] = nullptr;
            m_touchKeys_v2[j] = (TOUCHKEY_V2 *)&m_buffer[m_touchKeysNode_v2->ptrTouchkeyArray + i * sizeof( uint32_t ) + j * sizeof( TOUCHKEY_V2 )];
        }
    }
}

void FileTouchKeys::setTouchKey( uint8_t idx, const TOUCHKEY_V1 &tk )
{
    if ( (m_touchKeys_v1.size() > idx) )
    {
        TOUCHKEY_V1 *key = m_touchKeys_v1[idx];
        if ( key != nullptr )
        {
            *key = tk;
        }
    }
}

void FileTouchKeys::setTouchKey( uint8_t idx, const TOUCHKEY_V2 &tk )
{
    if ( (m_touchKeys_v2.size() > idx) )
    {
        TOUCHKEY_V2 *key = m_touchKeys_v2[idx];
        if ( key != nullptr )
        {
            *key = tk;
        }
    }
}

const TOUCHKEY_V1 &FileTouchKeys::touchKey_v1( uint8_t idx ) const
{
    static TOUCHKEY_V1 dummy;
    if ( (m_touchKeys_v1.size() > idx) )
    {
        TOUCHKEY_V1 *key = m_touchKeys_v1[idx];
        if ( key != nullptr )
        {
            return *key;
        }
    }
    return dummy;
}

const TOUCHKEY_V2 &FileTouchKeys::touchKey_v2( uint8_t idx ) const
{
    static TOUCHKEY_V2 dummy;
    if ( (m_touchKeys_v1.size() > idx) )
    {
        TOUCHKEY_V2 *key = m_touchKeys_v2[idx];
        if ( key != nullptr )
        {
            return *key;
        }
    }
    return dummy;
}

uint8_t FileTouchKeys::touchKeyCount() const
{
    if ( m_touchKeysNode_v1 != nullptr )
    {
        return m_touchKeysNode_v1->cnt;
    }
    else if ( m_touchKeysNode_v2 != nullptr )
    {
        return m_touchKeysNode_v2->cnt;
    }
    return 0;
}

uint8_t FileTouchKeys::deviceType() const
{
    if ( m_touchKeysNode_v1 != nullptr )
    {
        return m_touchKeysNode_v1->deviceType;
    }
    else if ( m_touchKeysNode_v2 != nullptr )
    {
        return m_touchKeysNode_v2->deviceType;
    }
    return enum_software_version_type::svt_unknown;
}

void FileTouchKeys::setDeviceType( uint8_t deviceType )
{
    if ( m_touchKeysNode_v1 != nullptr )
    {
        m_touchKeysNode_v1->deviceType = deviceType;
    }
    else if ( m_touchKeysNode_v2 != nullptr )
    {
        m_touchKeysNode_v2->deviceType = deviceType;
    }
}

uint8_t FileTouchKeys::version() const
{
    if ( m_touchKeysNode_v1 != nullptr )
    {
        return m_touchKeysNode_v1->version;
    }
    else if ( m_touchKeysNode_v2 != nullptr )
    {
        return m_touchKeysNode_v2->version;
    }
    return 0;
}