/**
* @file  protocol_docking_station_v1.cpp
* @brief Implementierung der Klasse TimeboyStation.
*/
//////////////////////////////////////////////////////////////////////

#include "datafox/communication/protocol_docking_station_v1.h"
#include <memory.h>

// ##################################################################
/**
* @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
*
*/
Datafox::Communication::ProtocolDockingstationV1::~ProtocolDockingstationV1() 
{
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

//#########################################################
//######################### CRC ###########################
//#########################################################

// ##################################################################
/**
* @brief Erzeugt einen 16bit CRC Wert fuer die uebergebenen Daten.
*/
uint16_t Datafox::Communication::ProtocolDockingstationV1::crc16( uint8_t *data, uint32_t size, uint16_t crc )
{
    connection().trace( "%s(%i):INFO(3): %s [ 0x%08X, %i, 0x%04X ]\n", __FILE__, __LINE__, __func__, data, size, crc );

    uint32_t i;
    uint8_t j;

    for ( i = 0; i < size; ++i )
    {
        crc = crc ^ data[i];

        for ( j = 0; j < 8; ++j )
        {
            if ( crc & 0x1 )
            {
                crc >>= 1; 
                crc = crc ^ 0xA001;
            }
            else 
            {
                crc >>= 1;
            }
        }
    }

    return crc;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
uint16_t Datafox::Communication::ProtocolDockingstationV1::readSlotFirmwareVersion( uint8_t deviceNumber )
{
    uint16_t ver;
    int32_t res;
    int32_t autodetect = 1;

    // 254 kann nur Einzeldocking sein und nutzt MasterIV Protokoll.
    // Bereich pruefen
    if ( deviceNumber < 1 || deviceNumber >= 254 )
    {
        return 0;
    }

    switch ( m_timeboySlot[deviceNumber].version() ) 
    {
    case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
        break;

    case 0: 
        { // Ermitteln der FWVersionProtokollversion
            Datafox::Communication::ConnectionBase::ConnectionType type;
            std::string commString;
            uint32_t commVal;
            uint16_t commTO;
            uint32_t commValOld = 0;
            connection().openParameters( type, commString, commVal, commTO );

            while ( autodetect )
            {
                // Bei Seriell ist es wahrscheinlicher, dass mit 9600bd kommuniziert wird, deswegen zuerst damit Pruefen und dann mit 38400bd.
                if ( type == Datafox::Communication::ConnectionBase::ConnectionType::serial && commVal != 9600 )
                {
                    // Schnittstellenbaud anpassen
                    commValOld = commVal;
                    commVal = 9600;
                    connection().changeSerialState( connection().serialPort(), 9600, commTO );
                }
                else 
                {
                    // Wurde Baud per autodetect abgeaendert?
                    if (commValOld)
                    {
                        connection().changeSerialState( connection().serialPort(), commValOld, commTO );
                    }

                    autodetect = 0;
                }

                // In der Version 1.25 (08.2007) wurde die Timeboyerkennung per Kommunikationspolling der Docking auf TB vorgenommen.
                // Ab 1.26 (09.2007) wurde die Unterstützung der Timeboyerkennung per Komperatorschaltung integriert.
                // De-/Aktivierung über 0x8000000C (Wert 2 kann dort verwendet werden), abfrage des Zustandes über 0x8000E000.

                if ( (res = readBytes( deviceNumber, MB_SLOT_VERSION, 0, (uint8_t*)&ver, sizeof( ver ) )) == IV_ERROR_SUCCESS )
                {
                    // < 1.31	Paketpuffer ist 45 Byte
                    // >= 1.31	Paketpuffer ist 150 Byte
                    m_timeboySlot[0].setVersion( commVal ); // Im 0 Byte die verwendete Baudrate halten.
                    m_timeboySlot[deviceNumber].setVersion( binToUintMsbFirst( (uint8_t*)&ver ) );
                    return m_timeboySlot[deviceNumber].version();
                }
                else
                {
                    m_timeboySlot[deviceNumber].setVersion( 0xFFFF );
                }
            }
            break;
        }

    default:
        return m_timeboySlot[deviceNumber].version();
    }

    return 0;
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::ProtocolDockingstationV1::resetSlot( uint8_t deviceNumber )
{
    if ( deviceNumber >= 254 )
    {
        return;
    }

    m_timeboySlot[deviceNumber].reset();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::ProtocolDockingstationV1::resetSlots()
{
    uint8_t deviceNumber;
    for ( deviceNumber = 0; deviceNumber < 255; ++deviceNumber )
    {
        m_timeboySlot[deviceNumber].reset();
    }
}

int32_t Datafox::Communication::ProtocolDockingstationV1::callDocking( uint8_t* sendBuffer, uint8_t sendCount, uint8_t *receiveBuffer, uint8_t *receiveCount )
{
    int32_t res;
    uint32_t nb;
    uint32_t rcnt = 0;
    uint32_t len;

    // Autobaud
    // Stimmt die aktuell Verwendete Baudrate nicht mit der benoetigten ueberein?
    if ( connection().type() == Datafox::Communication::ConnectionBase::ConnectionType::serial
        && m_timeboySlot[0].version() && connection().portNumber() != m_timeboySlot[0].version() )
    {
        connection().changeSerialState( connection().serialPort(), m_timeboySlot[0].version(), connection().readTimeout() );
    }

    // CRC noch an Paket anfuegen
    *(uint16_t*)&sendBuffer[sendCount] = crc16( sendBuffer, sendCount );
    sendCount += sizeof( uint16_t );

    // Empfangsanzahl initialisieren
    *receiveCount = 0;

#ifdef TRACE_DOCKING
    connection().trace( "%s(%i):INFO(1): %s.send( %i ) ( %s[ 0x%02X ], Add: 0x%08X  Cnt: %i )\n", __FILE__, __LINE__, __func__, sendBuffer[0], (sendBuffer[3] == MB_READ) ? "READ" : "WRITE", sendBuffer[3], binToUlongMsbFirst(&sendBuffer[4]), (sendBuffer[3] == MB_READ) ? sendBuffer[9] : sendBuffer[2] - 0x0A );
#endif

    // Paketdaten versenden.
    if ( (res = connection().writeData( sendBuffer, sendCount, nb, 1 )) != IV_ERROR_SUCCESS )
    {
        connection().trace( "%s(%i):ERROR(1): %s.transmit [ TransmitData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        resetSlot( deviceNumberReceiver() );
        return res;
    }
    // Alle Zeichen versendet?
    if ( nb != (int32_t)sendCount )
    {
        return -IV_ERROR_TRANSMIT;
    }

    // Kopf des Paketes empfangen.
    if ( (res = connection().readData( receiveBuffer, 4, nb, 1 )) != IV_ERROR_SUCCESS )
    {
        connection().trace( "%s(%i):ERROR(1): %s [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        resetSlot( deviceNumberReceiver() );
        return res;
    }
    // Alle Zeichen empfangen?
    if ( nb < 4 ) 
    {
        connection().trace( "%s(%i):ERROR(1): %s [ %i von %i Byte empfangen ]\n", __FILE__, __LINE__, __func__, nb, 4 );
        return -IV_ERROR_RECEIVE;
    }
    rcnt = nb;

    // Busadresse vergleichen
    if ( sendBuffer[0] != receiveBuffer[0] )
    {
        //#########################
        // Bugfix fuer Dockingstation FW Version 1.24
        // Naechstes Byte ist Adresse?
        if ( sendBuffer[0] == receiveBuffer[1] )
        {
            //getChannelBase().Trace("%s(%i):ERROR(1): %s [ Skip wrong start byte {0x%02X} ]\n", __FILE__, __LINE__, __func__, rxBuf[0]);
            receiveBuffer[0] = receiveBuffer[1];
            receiveBuffer[1] = receiveBuffer[2];
            receiveBuffer[2] = receiveBuffer[3];
            rcnt--;
        }
        else
        {
            // Delay bei TCP/IP ueber Com-Server
            if ( connection().type() == Datafox::Communication::ConnectionBase::ConnectionType::socket )
            {
                std::this_thread::sleep_for( std::chrono::milliseconds( 200 ) );
            }

            connection().trace( "%s(%i):ERROR(1): %s [ address wrong %i != %i ]\n", __FILE__, __LINE__, __func__, sendBuffer[0], receiveBuffer[0] );
#ifdef _DEBUG
            uint32_t i;
            for ( i = 0; i < nb; ++i ) 
            {
                connection().trace( "%02X ", receiveBuffer[i] );
            }
            connection().trace( "\n" );
#endif
            connection().flush();
            return -IV_ERROR_ADDRESS;
        }
    }

    // Solllaenge Paket aus Kopf ermitteln
    len = binToUintMsbFirst( &receiveBuffer[1] );
    if ( len < 7 || len > 200 ) 
    {
        // Delay bei TCP/IP ueber Com-Server
        if ( connection().type() == Datafox::Communication::ConnectionBase::ConnectionType::socket )
        {
            std::this_thread::sleep_for( std::chrono::milliseconds( 200 ) );
        }

        connection().trace( "%s(%i):ERROR(1): %s [ packet length < 7 || > 200 {%i}]\n", __FILE__, __LINE__, __func__, len );
#ifdef _DEBUG
        uint32_t i;
        for ( i = 0; i < nb; ++i ) 
        {
            connection().trace( "%02X ", receiveBuffer[i] );
        }
        connection().trace( "\n" );
#endif
        connection().flush();
        return -IV_ERROR_PACKETLENGTH;
    }

    // Body des Paketes empfangen.
    if ( (res = connection().readData( &receiveBuffer[rcnt], len-rcnt, nb, 1 )) != IV_ERROR_SUCCESS ) 
    {
        connection().trace( "%s(%i):ERROR(1): %s [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
        // Neuerkennung notwendig.
        resetSlot( deviceNumberReceiver() );
        return res;
    }
    // Alle Zeichen empfangen?
    if ( nb < (len - rcnt) ) 
    {
        connection().trace( "%s(%i):ERROR(1): %s [ Nachlesen %i von %i Byte empfangen ]\n", __FILE__, __LINE__, __func__, nb, (len - rcnt) );
        connection().flush();
        return -IV_ERROR_RECEIVE;
    }
    rcnt += nb;

    // Alle Zeichen empfangen?
    if ( crc16( receiveBuffer, rcnt ) ) 
    {
        connection().trace( "%s(%i):ERROR(1): %s [ Paket CRC ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_CRC_PACKET;
    }

    // Handelt es sich um eine Steuerantwort der Dockingstation?
    if ( rcnt == 7 )
    { 
#ifdef TRACE_DOCKING
        connection().trace( "%s(%i):INFO(1): %s.receive ( Befehl: 0x%02X, Fehler: 0x%02X )\n", __FILE__, __LINE__, __func__, receiveBuffer[3], receiveBuffer[4] );
#endif
    }
    else
    {
#ifdef TRACE_DOCKING
        connection().trace( "%s(%i):INFO(1): %s.receive ( Befehl: 0x%02X, Länge: %i )\n", __FILE__, __LINE__, __func__, receiveBuffer[3], rcnt );
#endif
    }

    // Zeitstempel der Antwort halten
    m_timeboySlot[deviceNumberReceiver()].setReceiveStampSlot( clock() );

    // Anzahl empfangener Bytes
    *receiveCount = rcnt;

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Prueft die Erreichbarkeit des Kommunikationspartners.
*
* @param deviceID Anzusprechender Kommunikationspartner.
* @param model Erreichtes Timeboymodel.
*/
int32_t Datafox::Communication::ProtocolDockingstationV1::readTimeboyModel( uint8_t deviceNumber, int32_t &model )
{
    connection().trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceNumber );
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t data[2];
    memset( data, 0, sizeof( data ) );

    // Ist die schnelle Steck-Erkennung verfuegbar?
    if ( m_timeboySlot[deviceNumber].detectionState() == 2 )
    {
        // Pruefung ob der Slot auch meldet, dass ein Timeboy steckt.
        if ( (res = readBytes( deviceNumber, MB_SLOT_STATE, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            return res;
        }
        if ( *(uint16_t*)data != 0x0100 )
        {
            // Timeboy ist nicht erreichbar, da nicht gesteckt!
            //			return -IV_ERROR_UNREACHABLE;
            // KORREKTUR (2012_05_14):
            // Da ab der TimeboyIV Version 04.02.02.40 die Ladeabschaltung voll unterstuetzt wird,
            // Kommt es durch die Abfrageroutine zu Fehlinformationen, da die Ladeschaltung meldet,
            // das kein Timeboy steckt obwohl dieses der Fall ist.
        }
    }

    do 
    {
        // Timeboy ansprechen.
        if ( (res = readBytes( deviceNumber, 0, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Wurde ermittelt, dass ein Timeboy steckt? -> Nochmal versuchen.
            if ( *(uint16_t*)data == 0x0100 )
            {
                *(uint16_t*)data = 0;
                continue;
            }

            // Fehler ausgeben
            return res;
        }
    } while ( res != IV_ERROR_SUCCESS );

    if ( data[0] == '~' && data[1] == '~' )
    {
        model = 4;
    }
    else
    {
        model = 3;
    }

    // Unabhaengig vom Timeboymodel
    // Pruefung auf die Verfuegbarkeit der Timeboyerkennung durchfuehren.
    if ( m_timeboySlot[deviceNumber].version() >= 0x0126 && m_timeboySlot[deviceNumber].detectionState() == 0 )
    {
        if ( (res = readBytes( deviceNumber, MB_SLOT_CHECK, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
        {
            // Fehler ausgeben
            return res;
        }

        if ( *(uint16_t*)data == 0x0002 )
        {
            // Wird unterstuetzt
            m_timeboySlot[deviceNumber].setDetectionState( 2 );
        }
        else if ( *(uint16_t*)data == 0x0001 )
        {
            // Pruefung ob der Slot auch meldet, dass ein Timeboy steckt.
            if ( (res = readBytes( deviceNumber, MB_SLOT_STATE, 0, data, 2, 1 )) != IV_ERROR_SUCCESS )
            {
                // Fehler ausgeben
                return res;
            }

            if ( *(uint16_t*)data == 0x0100 )
            {
                // Timeboy wurde erkannt und der Slot meldet, dass ein Timeboy steckt -> Komperatorschaltung bestueckt.
                *(uint16_t*)data = 0x0002;
                //if ((m_result = writeBytes(deviceID, MB_SLOT_CHECK, data, 2, 1)) != IV_ERROR_SUCCESS)
                //{
                //	// Fehler ausgeben
                //	return m_result;
                //}

                m_timeboySlot[deviceNumber].setDetectionState( 2 );
            }
            else
            {
                m_timeboySlot[deviceNumber].setDetectionState( 1 );
            }
        }
        else
        {
            m_timeboySlot[deviceNumber].setDetectionState( 1 );
        }
    }

    return res;
}

// ##################################################################
/**
* @brief Setzt den Zustand der Roten LED der Mehrfachdocking.
*
* @param deviceID Anzusprechender Kommunikationspartner.
*/
int32_t Datafox::Communication::ProtocolDockingstationV1::setDockingLed( uint8_t deviceNumber, int32_t state )
{
    connection().trace( "%s(%i):INFO(3): %s [ %i ]\n", __FILE__, __LINE__, __func__, deviceNumber );
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t data[1];

    if ( state != 0 )
    {
        data[0] = 0x00;
    }
    else
    {
        data[0] = 0xFF;
    }

    if ( (res = writeBytes( deviceNumber, MB_SLOT_LED, 0, data, 1 )) != IV_ERROR_SUCCESS )
    {
        // Fehler ausgeben
        return res;
    }

#ifdef TRACE_DOCKING
    connection().trace( "%s(%i):INFO(1): %s [ Letzter Wechsel vor %0.3f Sekunden ]\n", __FILE__, __LINE__, __func__, float(clock() - m_timeboySlot[deviceNumber].ledStamp()) / float(CLOCKS_PER_SEC) );
#endif

    // Zustand der LED und Zeitpunkt des wechsels halten
    m_timeboySlot[deviceNumber].setLedState( (state != 0) ? 0x01 : 0xFF );
    m_timeboySlot[deviceNumber].setLedStamp( clock() );

    return res;
}

// ##################################################################
/**
* @brief Setzt den Kommunikationszustand eines Slots der Mehrfachdocking.
*
* @param deviceID Anzusprechender Kommunikationspartner.
* @param state Zu setzender Kommunikationszustand.
*/
int32_t Datafox::Communication::ProtocolDockingstationV1::setTimeboySlotState( uint8_t deviceNumber, int32_t state, int32_t commandResult )
{
    (void)commandResult;
    connection().trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, deviceNumber, state );

    if ( isSlot( deviceNumber ) )
    {
        // Ist der geforderte Zustand gleich dem aktuellen?
        if  (m_timeboySlot[deviceNumber].ledState() == state )
        {
            return IV_ERROR_SUCCESS;
        }

        if ( state != 0 )
        {
            int32_t model = 0;
            // Es muss ein gueltiges Timeboymodel vorliegen
            if ( readTimeboyModel( deviceNumber, model ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }

            if ( model != 4 )
            {
                return -IV_ERROR_TIMEBOY_I_II_III;
            }

            // Rote LED einschalten
            if ( setDockingLed( deviceNumber, 1 ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }
        }
        else
        {
            // Rote LED ausschalten
            if( setDockingLed( deviceNumber, 0 ) != IV_ERROR_SUCCESS )
            {
                return -IV_ERROR_UNREACHABLE;
            }
        }
    }
    else
    {
        // Bei der angesprochenen ID handelt es sich um keinen Slot einer TimeboyStation.
        return -IV_ERROR_SLOT_NONE;
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Entgegennehmen der Verbindung und starten eines Verarbeitungsthreads.
*/
int32_t Datafox::Communication::ProtocolDockingstationV1::startCommunicationThread() 
{
    connection().trace( "%s(%i):INFO(3): %s [  ]\n", __FILE__, __LINE__, __func__ );

    // Hauptthread starten
    if ( m_communicationThread.joinable() == false )
    {
        m_communicationThread = std::thread( [=] { communicationThread(); } );

        int32_t i;
        // Hochlauf des Workers pruefen
        for ( i = 0; m_communicationThread.joinable() && (threadState() != ts_idle) && (i < 60); i++ )
        {
            std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );
        }
    }

    return m_communicationThread.joinable();
}

// ##################################################################
/**
* @brief Beendigungssignal im Workerthread setzen und auf dessen beendigung warten.
*/
void Datafox::Communication::ProtocolDockingstationV1::stopCommunicationThread()
{
    connection().trace( "%s(%i):INFO(3): %s", __FILE__, __LINE__, __func__ );

    // Socket-Thread deaktivieren, wenn aktiv
    if ( m_communicationThread.joinable() )
    {
        // WorkerThread fuer Socket beenden.
        setThreadCommand( tc_exit );

        // Thread Beendigung abwarten (communicationThread stellt m_endSignal auf FALSE)
        m_communicationThread.join();
    }
}

// ##################################################################
/**
* @brief Workerthread fuer die Verarbeitung der Kommunikationsaufgaben.
*/
void Datafox::Communication::ProtocolDockingstationV1::communicationThread()
{
    connection().trace( "%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

    // Threadzustand setzen
    m_threadState = ts_begin;
    m_threadCommand = tc_none;

    // Hauptschleife ausfuehren?
    bool run = true;

    // Hauptschleife des Worker
    while ( run )
    {
        switch ( m_threadState )
        {
        case ts_begin:
            // Threadzustand setzen
            m_threadState = ts_idle;
            break;

        case ts_idle: {
            if ( m_threadCommand != tc_none )
            {
                // Threadzustand setzen
                m_threadState = ts_execute_command;
            }
            else
            {
                //// Pruefung ob LED augeschaltet werden muss.
                //if (IsFlag(TSF_AUTO_LED))
                //{
                //	for (deviceID=1; deviceID < 255; deviceID++)
                //	{
                //		if (_timeboySlot[deviceID].getLEDState() && ((clock() - _timeboySlot[deviceID].getLEDStamp()) > 3))
                //		{
                //			setDockingLED(deviceID, 0);
                //		}
                //	}
                //}

                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
            break;}

        case ts_execute_command: {

            switch ( m_threadCommand )
            {
            case tc_none:
                break;

            case tc_exit:
                // Hauptschleife verlassen
                run = false;
                break;

            case tc_write:
                m_commandParameter.res = threadWriteBytes( m_commandParameter.deviceNumberReceiver, m_commandParameter.baseAddress, m_commandParameter.offset, m_commandParameter.buffer, m_commandParameter.count, m_commandParameter.trials );
                break;

            case tc_read:
                m_commandParameter.res = threadReadBytes( m_commandParameter.deviceNumberReceiver, m_commandParameter.baseAddress, m_commandParameter.offset, m_commandParameter.buffer, m_commandParameter.count, m_commandParameter.trials );
                break;
            }

            // Aufgabe abgearbeitet
            m_threadState = ts_idle;
            m_threadCommand = tc_none;
            break;}

                                //case none:
        case ts_end:
        case ts_dummy:
            break;
        }
    }

    // Threadzustand setzen
    m_threadState = ts_end;

    connection().trace( "%s(%i):INFO(1): %s [ End ]\n", __FILE__, __LINE__, __func__ );
}

//#########################################################
//################## Members ##############################
//#########################################################

/*############################################################################*/
//  FUNKTION:    writeBytes
/** Schreibbefehl, um Bytes in den Timeboy zu schreiben.

Autor                 : Michael Wicher            @n
erstellt am           : 02.03.2007                @n

@param[in] address Zieladresse im Timeboy.
@param[in] buf Zeiger auf Daten, die gesendet werden sollen.
@param[in] count Anzahl Bytes zum Senden.
@param[in] slot Steckplatz der Docking [1..10].

@return #TRUE -> Erfolg, #FALSE -> Miszerfolg.

@b Aufgaben:
- Bytes an den Timeboy senden.

@b Aenderungen:
- 17.01.2008 (MW) Neuer Transferbefehl, um die Kommunikation durch einen Mutex
zu schützen.
- 01.02.2008 (MW) Bei allen Befehlen für den Timeboy wird zuvor bei der Docking
angefragt, ob auch ein Timeboy steckt.
Weiterhin wird hier geprüft, ob sich der Task terminieren soll.
Falls kein Timeboy steckt, wird hier auch die Startphase
abgeschloszen, so dasz die Funktion #DOCKINGControl beendet wird.
- 27.03.2008 (MW) Max. fünf Versuche, den Befehl abzusetzen. Dazwischen wird
immer geprüft, ob der Timeboy laut Docking steckt.
- 19.05.2008 (MW) Vor Abarbeitung des Befehls wird geprüft, ob der ausgewählte
Steckplatz der Docking noch erreichbar ist.
Bei Befehlen dirket für die Docking werden nun 2 Versuche
durchgeführt.
Die CRC-Berechnung verschoben.
*/
/*############################################################################*/

int32_t Datafox::Communication::ProtocolDockingstationV1::writeBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t writeTrials )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Ueber den Kommunikationsthread kommunizieren?
    if ( m_communicationThread.joinable() )
    {
        m_commandParameter.deviceNumberReceiver = deviceNumber;
        m_commandParameter.baseAddress = baseAddress;
        m_commandParameter.offset = offset;
        m_commandParameter.buffer = buffer;
        m_commandParameter.count = count;
        m_commandParameter.trials = writeTrials;

        // Schreibvorgang starten.
        if ( setThreadCommand( tc_write ) )
        {
            while ( threadCommand() != tc_none )
            {
                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
        }
        else
        {
            return -IV_ERROR_MODULE_BUSY;
        }

        return m_commandParameter.res;
    }

    // Direkte Befehlsausfuehrung
    res = threadWriteBytes( deviceNumber, baseAddress, offset, buffer, count, writeTrials );

    if ( res == IV_ERROR_SUCCESS )
    {
        // Nach erfolgreicher Kommunikation automatisch den
        // Kommunikationsthread starten, wenn noch nicht geschehen.
        //		startCommunicationThread();
    }

    return res;
}

int32_t Datafox::Communication::ProtocolDockingstationV1::threadWriteBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t writeTrials )
{
    static uint8_t pre[MB_SLOT_BUFFER_SIZE_LARGE];

    int32_t res = IV_ERROR_SUCCESS;
    uint8_t current[MB_SLOT_BUFFER_SIZE_LARGE];
    uint8_t i;
    uint8_t limit;
    uint8_t send[MB_SLOT_BUFFER_SIZE_LARGE];
    uint16_t pause;
    uint8_t receive[MB_SLOT_BUFFER_SIZE_LARGE];
    uint8_t repeat;
    uint8_t receive_count;
    uint8_t sendPre;
    int32_t executeFails = 0;
    uint32_t address = baseAddress + offset;

    // Remoteadresse setzen.
    setDeviceNumberReceiver( deviceNumber );

    uint8_t bufSize = dataBufferSize( deviceNumberReceiver() );

    // prüfe, ob die Busnummer verstellt ist
    //DOCKINGBusnumVerify(slot);

    for ( ; count; ) 
    {
        sendPre = FALSE;

        for ( limit = 20; limit; limit >>= 1 ) 
        {
            // bei allen Befehlen für den Timeboy bei der Docking anfragen, dasz auch ein Timeboy steckt
            //if ((((UINT8 *)&address)[3] & 0xF0) == 0x40) {

            //	KERNELCheckForTerminate(DOCKINGStop);

            //	// falls kein Timeboy steckt, macht der weitere Ablauf keinen Sinn und man kann abbrechen
            //	if (DOCKINGCheckTimeboy(slot) != DOCKING_TIMEBOY_PUSHED) {

            //		// Startphase abgeschloszen, jetzt darf auch der IO-Task loslaufen
            //		globalDOCKING.ready = TRUE;

            //		return (FALSE);

            //	}

            //} else {

            //	limit = 1;

            //}

            current[0] = (uint8_t)deviceNumberReceiver();
            current[1] = 0x00;
            current[2] = 0x0A + ((count > bufSize) ? bufSize : count);
            current[3] = 0x22;
            current[4] = ((uint8_t *)&address)[3];
            current[5] = ((uint8_t *)&address)[2];
            current[6] = ((uint8_t *)&address)[1];
            current[7] = ((uint8_t *)&address)[0];

            memcpy( (uint8_t *)&current[8], buffer, (count > bufSize) ? bufSize : count );

            repeat = limit;

            // je nach Timeboy oder Docking als Empfänger
            if ( current[4] == 0x80 ) 
            {
                // beim Stellen der Busnummer nur ein Versuch
                if ( (current[6] == 0x00) && (current[7] == 0x00) ) 
                {
                    repeat = 1;
                    pause = 0;

                    // bei allen anderen Anfragen an die Docking zwei Versuche!
                } 
                else 
                {
                    repeat = 2;
                    pause = 80;
                }
            } 
            else 
            {
                pause = 2000 / repeat;
            }

            for ( i = 0; i < repeat; ++i ) 
            {
                //if (i) DEBUGString(TRACE_DEVICE_DOCKING, "repeat write <%u>.", i);

                if ( sendPre == TRUE ) 
                {
                    // lade, falls nötig,  den Vorgänger
                    memcpy( send, pre, sizeof( send ) );
                } 
                else 
                {
                    // aktuelles Paket laden
                    memcpy( send, current, sizeof( send ) );
                }

                if ( (res = callDocking( send, send[2] - 2, receive, &receive_count )) == IV_ERROR_SUCCESS ) 
                {
                    // Anzahl fehlversuche zurücksetzen
                    executeFails = 0;

                    // Paket für Docking?
                    if ( send[4] == 0x80 ) 
                    {
                        if ( *(uint32_t *)&receive[4] == *(uint32_t *)&send[4] ) 
                        {
                            break;
                        }
                    } 
                    else // Paket fuer Timeboy
                    {
                        if ( receive_count == 7 ) 
                        {
                            //DEBUGString(TRACE_DEVICE_DOCKING, "%02X, %08lX", receive[4], CONVSwapLong(*(UINT32 *)&transmit[4]));

                            if ( receive[4] == 0x22 ) 
                            {
                                //KERNELMilliSleep(100);
                                std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );

                                // falls nicht der Vorgänger gesendet wurde
                                if ( sendPre == FALSE ) 
                                {
                                    // gesendetes Paket zur evtl. Wiederholung merken
                                    memcpy( pre, send, sizeof( pre ) );

                                    // fertig
                                    break;
                                }

                                // Limit neu festlegen
                                repeat = limit;

                                // jetzt das aktuelle Paket nochmal senden
                                sendPre = FALSE;
                            } 
                            else if ( (receive[4] == 0x21) || (receive[4] == 0x24) ) 
                            {
                                //DEBUGString(TRACE_DEVICE_DOCKING, "repeat pre.");

                                // falls eine Wiederholung begonnen wird
                                if ( sendPre == FALSE ) 
                                {
                                    // lege Limit neu fest
                                    repeat = limit;
                                }

                                sendPre = TRUE;
                            }
                        }
                        else
                        {
                            //if (*(uint32_t *)&receive[4] == *(uint32_t *)&transmit[4]) 
                            //{
                            //	break;
                            //}
                        }
                    }
                }
                else
                {
                    // Fehlerhafte Ausfuehrungen halten
                    executeFails++;
                    if ( executeFails >= writeTrials )
                    {
                        return res;
                    }
                }

                if ( pause ) 
                {
                    std::this_thread::sleep_for( std::chrono::milliseconds( pause ) );
                }
            }

            // bei erfolgreichem Senden keine Wiederholung
            if ( i < repeat ) 
            {
                break;
            }
        }

        // falls das Limit erreicht wurde, konnte der Befehl nicht erfolgreich gesendet werden!
        if ( limit == 0 ) 
        {
            return -IV_ERROR_UNREACHABLE;
        }

        address += bufSize;
        buffer += bufSize;
        count -= (count > bufSize ? bufSize : count);
    }

    // Zeitpunkt der letzten Timeboykommunikation halten
    if ( (address & 0x80000000) == 0 )
    {
        m_timeboySlot[deviceNumberReceiver()].setReceiveStampTimeboy( clock() );
    }

    return res;
}

/*############################################################################*/
//  FUNKTION:    readBytes
/** Lesebefehl, um Bytes aus dem Timeboy zu lesen.

Autor                 : Michael Wicher            @n
erstellt am           : 02.03.2007                @n

@param[in] buf Zeiger, wo gelesene Daten hinkopiert werden sollen.
@param[in] address Quelladresse im Timeboy.
@param[in] count Anzahl Bytes zum Lesen.
@param[in] slot Steckplatz der Docking [1..10].

@return #TRUE -> Erfolg, #FALSE -> Miszerfolg.

@b Aufgaben:
- Bytes vom Timeboy lesen.

@b Aenderungen:
- 17.01.2008 (MW) Neuer Transferbefehl, um die Kommunikation durch einen Mutex
zu schützen.
- 01.02.2008 (MW) Bei allen Befehlen für den Timeboy wird zuvor bei der Docking
angefragt, ob auch ein Timeboy steckt.
Weiterhin wird hier geprüft, ob sich der Task terminieren soll.
Falls kein Timeboy steckt, wird hier auch die Startphase
abgeschloszen, so dasz die Funktion #DOCKINGControl beendet wird.
- 27.03.2008 (MW) Max. fünf Versuche, den Befehl abzusetzen. Dazwischen wird
immer geprüft, ob der Timeboy laut Docking steckt.
- 19.05.2008 (MW) Vor Abarbeitung des Befehls wird geprüft, ob der ausgewählte
Steckplatz der Docking noch erreichbar ist.
Bei Befehlen dirket für die Docking werden nun 2 Versuche
durchgeführt.
Die CRC-Berechnung verschoben.
*/
/*############################################################################*/

int32_t Datafox::Communication::ProtocolDockingstationV1::readBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t readTrials )
{
    int32_t res = IV_ERROR_SUCCESS;
    // Ueber den Kommunikationsthread kommunizieren?
    if ( m_communicationThread.joinable() )
    {
        m_commandParameter.deviceNumberReceiver = deviceNumber;
        m_commandParameter.baseAddress = baseAddress;
        m_commandParameter.offset = offset;
        m_commandParameter.buffer = buffer;
        m_commandParameter.count = count;
        m_commandParameter.trials = readTrials;

        // Lesevorgang starten.
        if ( setThreadCommand( tc_read ) )
        {
            while ( threadCommand() != tc_none )
            {
                //connection().trace("%s(%i):INFO(1): %s [ wait ]\n", __FILE__, __LINE__, __func__);
                std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
            }
        }
        else
        {
            return -IV_ERROR_MODULE_BUSY;
        }

        return m_commandParameter.res;
    }

    // Direkte Befehlsausfuehrung
    res = threadReadBytes( deviceNumber, baseAddress, offset, buffer, count, readTrials );

    if ( res == IV_ERROR_SUCCESS )
    {
        // Nach erfolgreicher Kommunikation automatisch den
        // Kommunikationsthread starten, wenn noch nicht geschehen.
        //		startCommunicationThread();
    }

    return res;
}

int32_t Datafox::Communication::ProtocolDockingstationV1::threadReadBytes( uint8_t deviceNumber, uint32_t baseAddress, int32_t offset, uint8_t *buffer, int32_t count, int32_t readTrials )
{
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t i;
    uint8_t limit;
    uint8_t send[MB_SLOT_BUFFER_SIZE_LARGE];
    uint16_t pause;
    uint8_t receive[MB_SLOT_BUFFER_SIZE_LARGE];
    uint8_t repeat;
    uint8_t receive_count;
    int32_t executeFails = 0;
    int32_t error21Cnt = 0;
    uint32_t address = baseAddress + offset;

    // Remoteadresse setzen.
    setDeviceNumberReceiver( deviceNumber );

    uint8_t bufSize = dataBufferSize( deviceNumberReceiver() );

    if ( offset == MB_SLOT_BUFFER_DATA_SIZE_SMALL )
    {
        address = baseAddress;
        if ( bufSize == MB_SLOT_BUFFER_DATA_SIZE_SMALL || (count - MB_SLOT_BUFFER_DATA_SIZE_SMALL) <= MB_SLOT_BUFFER_DATA_SIZE_SMALL )
        {
            // Nur wenn der moegliche Lesepuffer groesser ist, macht es sinn die bereits gelesenen Daten nochmals mitzulesen?
            // Die Dockingstation liest ein Paket vor und sendet immer die zuletzt angeforderte Paketgroesse in Folge.
            address += MB_SLOT_BUFFER_DATA_SIZE_SMALL;
            // bereits gelesene Daten von der geforderten Groesse abziehen
            count -= MB_SLOT_BUFFER_DATA_SIZE_SMALL;
            // Puffer weitersetzen
            buffer += MB_SLOT_BUFFER_DATA_SIZE_SMALL;
        }
    }

    // prüfe, ob die Busnummer verstellt ist
    //DOCKINGBusnumVerify(slot);

    for ( ; count; ) 
    {
        for ( limit = 20; limit; limit >>= 1 ) 
        {
            // bei allen Befehlen für den Timeboy bei der Docking anfragen, dasz auch ein Timeboy steckt
            //if ((((UINT8 *)&address)[3] & 0xF0) == 0x40) {

            //	KERNELCheckForTerminate(DOCKINGStop);

            //	// falls kein Timeboy steckt, macht der weitere Ablauf keinen Sinn und man kann abbrechen
            //	if (DOCKINGCheckTimeboy(slot) != DOCKING_TIMEBOY_PUSHED) {

            //		// Startphase abgeschloszen, jetzt darf auch der IO-Task loslaufen
            //		globalDOCKING.ready = TRUE;

            //		return (FALSE);

            //	}

            //} else {

            //	limit = 1;

            //}

            send[0] = (uint8_t)deviceNumberReceiver();
            send[1] = 0x00;
            send[2] = 0x0C;
            send[3] = 0x21;
            send[4] = ((uint8_t *)&address)[3];
            send[5] = ((uint8_t *)&address)[2];
            send[6] = ((uint8_t *)&address)[1];
            send[7] = ((uint8_t *)&address)[0];
            send[8] = 0x00;
            send[9] = (count > bufSize) ? bufSize : count;

            repeat = limit;

            // je nach Timeboy oder Docking als Empfänger
            if ( send[4] == 0x80 ) 
            {
                repeat = 2;
                pause = 80;
            } 
            else 
            {
                pause = 2000 / repeat;
            }

            for ( i = 0; i < repeat; ++i ) 
            {
                //if (i) DEBUGString(TRACE_DEVICE_DOCKING, "repeat <%u>.", i);

                if ( (res = callDocking( send, 0x0A, receive, &receive_count )) == IV_ERROR_SUCCESS ) 
                {
                    // Fehlerhafte Ausfuehrungen zuruecksetzen
                    executeFails = 0;

                    if ( send[4] == 0x80 )
                    {
                        // wurde ein gültiges Antwortpaket empfangen, prüfen, ob auch die Länge paszt
                        if ( *(uint32_t *)&receive[4] == *(uint32_t *)&send[4] ) 
                        {
                            break;
                        }
                    }
                    else
                    {
                        if ( receive_count == 7 )
                        {
                            switch ( receive[4] )
                            {
                            case 0x22:
                            case 0x23:
                                break;
                            case 0x21:
                            case 0x24:
                                {
                                    // Erster Befehl kann aufgrund von Sleep unbeantwortet bleiben.
                                    error21Cnt++;
                                    if ( error21Cnt > 1 )
                                    {
                                        return -IV_ERROR_UNREACHABLE;
                                    }
                                }
                                break;
                            }
                        }
                        else
                        {
                            // Fehlerzaehler ruecksetzen.
                            error21Cnt = 0;

                            // wurde ein gueltiges Antwortpaket empfangen, pruefen, ob auch die Adresse passt
                            if ( *(uint32_t *)&receive[4] == *(uint32_t *)&send[4] ) 
                            {
                                break;
                            }
                        }
                    }
                }
                else
                {
                    // Fehlerhafte Ausfuehrung
                    executeFails++;
                    if ( executeFails >= readTrials )
                    {
                        return res;
                    }
                }

                //KERNELMilliSleep(pause);
                std::this_thread::sleep_for( std::chrono::milliseconds( pause ) );
            }

            // bei erfolgreichem Senden keine Wiederholung
            if ( i < repeat ) 
            {
                break;
            }
        }

        // falls das Limit erreicht wurde, konnte der Befehl nicht erfolgreich gesendet werden!
        if ( limit == 0 ) 
        {
            return -IV_ERROR_UNREACHABLE;
        }

        memcpy( buffer, (uint8_t *)&receive[8], (count > bufSize) ? bufSize : count );

        buffer += bufSize;
        address += bufSize;
        count -= ((count > bufSize) ? bufSize : count);
    }

    // Zeitpunkt der letzten Timeboykommunikation halten
    if ( (address & 0x80000000) == 0 )
    {
        m_timeboySlot[deviceNumberReceiver()].setReceiveStampTimeboy( clock() );
    }

    return res;
}
