#include "channel/channel_4.h"
using namespace Datafox::Communication;

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <array>
#include <bitset>
#include <boost/algorithm/string.hpp>
#include <cstring>
#include <algorithm>

#if defined(_MSC_VER) || defined(__MINGW32__)
#define stat _stat
#endif

// Versionspruefung
#include "common/dfcom_utils.h"
#include "DatafoxLibraryIV/include/dfcom_version.h"

#include "file/file_device_text.h"
#include "file/file_touch_keys.h"
#include "file/file_display_colors.h"
#include "file/file_wlan_config.h"

#include "document_studio.h"
#include "file/dfcom_ini_helper.h"

#include "common/utils.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf	    _snprintf
#endif

// ##################################################################
/**
* @brief Initialisiert die Membervariablen und
*        benoetigte Umgebung.
*
* @param connectionNumber Verbindungsnummer die für Log usw. verwendet wird. Sie identifiziert
*           die Klasse.
*/
Channel4::Channel4( int32_t connectionNumber, uint8_t deviceAddress )
    : ModuleTable( connectionNumber, deviceAddress )
{
    trace( "%s(%i):INFO(3): %s [ %i, %i ]\n", __FILE__, __LINE__, __func__, connectionNumber, deviceAddress );

    /*	//#####################################################
    // SOCKETS
    // Benutzung der Socket - Funktionalitaet initialisieren
    WSADATA WindowsData;
    WORD    Version;

    Version = MAKEWORD(2, 2);
    WSAStartup(Version, &WindowsData);
    //#####################################################
    */

    // Membervariablen setzen
    setCommunicationThreadStopSignal( false );
    setActiveModeState( ams_unused );
    setCommunicationThreadError( IV_ERROR_SUCCESS );

    // Sockets initialisieren
    m_listenSocket = nullptr;

    // Callback fuer Nachricht auf Datensaetze
    m_callBackNotifyRecordAvailable = nullptr;

    // Melden von Datensaetzen aktiviert?
    m_notify = 0;

    // Hardwareinfo aufsetzen
    memset( &m_hardwareInfo, 0, sizeof( StructHardwareInfo ) );
    m_hardwareInfo.typeDevice = IV_VALUE_TYPE_DEVICE_PZE;
    m_hardwareInfo.typeBoard = IV_VALUE_TYPE_BOARD_2_1;
    m_hardwareInfo.componentRS485_0 = 0;
    m_hardwareInfo.typeTCPIP = IV_VALUE_TYPE_TCPIP_XPORT;
    m_hardwareInfo.typeModemMC = IV_VALUE_TYPE_MODEM_MC_NONE;
    m_hardwareInfo.componentRS485_1 = 0;
    m_hardwareInfo.typeTransponder = IV_VALUE_TYPE_TRANSPONDER_TSR32;
    m_hardwareInfo.typeDisplay = IV_VALUE_TYPE_DISPLAY_SMALL;
    m_hardwareInfo.typeKeypad = IV_VALUE_TYPE_KEYPAD_9;
    m_hardwareInfo.typeMMC = IV_VALUE_TYPE_MMC_NONE;
    m_hardwareInfo.typeGPS = IV_VALUE_TYPE_GPS_NONE;
    m_hardwareInfo.componentUSB = 0;
    m_hardwareInfo.typeRadio = IV_VALUE_TYPE_RADIO_NONE;
    m_hardwareInfo.typeDCF77 = IV_VALUE_TYPE_DCF77_NONE;
    m_hardwareInfo.componentInAnalog = IV_VALUE_COMPONENT_IN_ANALOG_NONE;
    m_hardwareInfo.componentInDigital = IV_VALUE_COMPONENT_IN_DIGITAL_1;
    m_hardwareInfo.componentOutDigital = IV_VALUE_COMPONENT_OUT_DIGITAL_1;
    m_hardwareInfo.typeFingerprint = IV_VALUE_TYPE_FINGERPRINT_NONE;
    m_hardwareInfo.typeExternalReader = IV_VALUE_TYPE_EXTERNAL_READER_UNDEFINED;
    m_hardwareInfo.optionIOModule = 0;
    m_hardwareInfo.optionTBAdaption = 0;

    // Textdaten als Default uebertragen
    m_loadAsDefaultLanguage = 0;

    m_deviceUpdate = std::unique_ptr<DeviceUpdate>( new DeviceUpdate() );
}

// ##################################################################
/**
* @brief Destruktor bereinigt die verwendete Umgebung.
*/
Channel4::~Channel4()
{
    // Beendigung sicherstellen, auch wenn er bereits beendet wurde.
    if ( m_communicationThread.joinable() )
    {
        setCommunicationThreadStopSignal();
        m_communicationThread.join();
    }

    //#####################################################
    // SOCKETS

    // Socket - Funktionalitaet exit
    //	WSACleanup();
    //#####################################################
}

// ##################################################################
/**
* @brief Entgegennehmen der Verbindung und starten eines Verarbeitungsthreads.
*/
int32_t Channel4::connect( SOCKET *listen_socket, int32_t connectionNumber, uint8_t deviceAddress, int32_t timeout, int32_t aliveTimeout, int32_t infoFlags )
{
    trace( "%s(%i):INFO(3): %s [ 0x08X, %i, %i, %i, %i ]\n", __FILE__, __LINE__, __func__, listen_socket, connectionNumber, deviceAddress, timeout, aliveTimeout );

    // Hauptthread starten
    if ( activeModeState() == ams_unused )
    {
        // Verbindungsobjekt ist nun in Verwendung.
        setActiveModeState( ams_starting );
        trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_starting ]\n", __FILE__, __LINE__, __func__, deviceAddress );

        // Wenn das Kanalobjekt bereits verwendet wurde, dann einen join ausführen damit das Thread-Objekt ungueltig wird.
        if ( m_communicationThread.joinable() )
        {
            setCommunicationThreadStopSignal();
            m_communicationThread.join();
        }

        // Kanalnummer und DeviceID uebernehmen.
        setConnectionNumber( connectionNumber );
        setDeviceNumberSender( deviceAddress );

        // Listensocket von dem die Verbindungsanforderung kommt wird 
        // ueber lokale Membervariable uebergeben.
        m_listenSocket = listen_socket;
        if ( timeout < 3000 )
        {
            trace( "%s(%i):WARNING(1): %s(%i) [ Timeout von %i auf mindestens 3000ms gesetzt. ]\n", __FILE__, __LINE__, __func__, deviceAddress, timeout );
            timeout = 3000;
        }
        else if ( timeout > 60000 )
        {
            trace( "%s(%i):WARNING(1): %s(%i) [ Timeout von %i auf maximal 60000ms gesetzt. ]\n", __FILE__, __LINE__, __func__, deviceAddress, timeout );
            timeout = 60000;
        }
        connection().setReadTimeout( timeout );
        m_aliveTimeout = aliveTimeout;
        m_infoFlags = infoFlags;

        // Kommunikationsworker
        setCommunicationThreadStartSignal();
        m_communicationThread = std::thread( [=]
        {
            communicationThread();
        } );
        // Thread umschalten
        std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );

        int32_t loops = 0;
        // Start des Threads pruefen.
        while ( communicationThreadStartSignal() == false )
        {
            if ( loops == 30 )
            {
                // Nach 3 Sekunden sollte der Threadstart und die Verbindungsannahme beendet sein.
                if ( m_communicationThread.joinable() )
                {
                    setCommunicationThreadStopSignal();
                    m_communicationThread.join();
                }

                // Fehler beim Starten des Threads da immer noch im Zustand is_starting?
                if ( activeModeState() == ams_starting )
                    setActiveModeState( ams_unused );

                break;
            }

            std::this_thread::sleep_for( std::chrono::milliseconds( 100 ) );
            loops++;
        }
    }

    return (activeModeState() != ams_unused);
}

// ##################################################################
/**
* @brief Gewünschte Info über die Active Verbindung erstellen.
*/
bool Channel4::activeConnectionInfo( char *info, int32_t *infoLength )
{
    trace( "%s(%i):INFO(3): %s [ 0x%p, 0x%p ]\n", __FILE__, __LINE__, __func__, info, infoLength );

    if ( activeModeState() == ams_active )
    {
        char infoText[255];
        int32_t infoPos = 0;
        int32_t infoPosMax = 0;
        memset( infoText, 0, sizeof( infoText ) );
        std::bitset<32> flags( m_infoFlags );

        for ( int32_t paramNumber = 0; infoPos != -1 && infoPos < *infoLength; ++paramNumber )
        {
            infoPosMax = infoPos;

            if ( paramNumber == 0 ) // Gerätetyp
            {
                infoPos = snprintf( infoText, sizeof( infoText ), "%u\r", m_hardwareInfo.typeDevice );
            }
            else if ( paramNumber == 1 ) // Seriennummer
            {
                infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "%u\r", static_cast<uint32_t>(property( "device.serial", -1 )) );
            }
            else if ( paramNumber == 2 ) // Firmwareversion optional
            {
                // Wenn keine Flags gesetzt wurden, dann Ende.
                if ( !flags.count() )
                    break;

                if ( flags.test( 0 ) )
                    infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "%s\r", m_versionFirmware.c_str() );
                else
                    infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "\r" );
            }
            else if ( paramNumber == 3 ) // TCPIP Adresse und Port optional
            {
                // Wenn keine weiteren Flags gesetzt wurden, dann Ende.
                if ( !flags.test( 1 ) && !flags.test( 2 ) )
                    break;

                if ( flags.test( 1 ) )
                    infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "%s:%i\r", m_tcpipInfo.ip.c_str(), m_tcpipInfo.port );
                else
                    infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "\r" );
            }
            else if ( paramNumber == 4 ) // Mac optional
            {
                if ( flags.test( 2 ) )
                    infoPos += snprintf( &infoText[infoPos], sizeof( infoText ) - infoPos, "%s\r", m_tcpipInfo.mac.c_str() );
            }
            else
                break;
        }
        // Pruefen ob die Werte kopiert werden koennen; die Puffergroesse muss passen!
        if ( infoPosMax < *infoLength )
        {
            snprintf( info, infoPosMax, "%s", infoText );
            info[infoPosMax] = 0;
            *infoLength = infoPosMax;
            return true;
        }
        else
        {
            *infoLength = 0;
        }
    }

    return false;
}

bool Channel4::communicationThreadStartSignal()
{
    return m_communicationThreadStartSignal;
}

void Channel4::setCommunicationThreadStartSignal( bool signal )
{
    m_communicationThreadStartSignal = signal;
}

bool Channel4::communicationThreadStopSignal()
{
    return m_communicationThreadStopSignal;
}

void Channel4::setCommunicationThreadStopSignal( bool signal )
{
    m_communicationThreadStopSignal = signal;
}

int32_t Channel4::communicationThreadError()
{
    return m_communicationThreadError;
}

void Channel4::setCommunicationThreadError( int32_t error )
{
    m_communicationThreadError = error;
}

// ##################################################################
/**
* @brief Workerthread fuer die Verarbeitung der Kommunikationsaufgaben.
*/
void Channel4::communicationThread()
{
    trace( "%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__ );

#if defined(_MSC_VER)
#ifndef MYCONST
#define MYCONST const
#endif
#else
#define MYCONST
#endif

    // Zustand auf Startet.
    setCommunicationThreadStopSignal( false );
    setCommunicationThreadError( IV_ERROR_SUCCESS );

    // Hauptschleife ausfuehren?
    bool run = false;

    int32_t res;
    int32_t failCount = 0;

    SOCKADDR_IN sa;
    struct timeval seltime;
    fd_set readfds;

    uint16_t count;
    uint8_t data[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe

    int32_t len;

    // Timeout fuer Abfrage auf Daten
    seltime.tv_sec = 0;
    seltime.tv_usec = 10000;

    // Zustand setzen.
    setActiveModeState( ams_accepting );
    trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_accepting ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

    // Verbindungsgesuch annehmen.
    len = sizeof( sa );
    connection().setSocket( accept( *m_listenSocket, (SOCKADDR *)&sa, (socklen_t *)&len ) );

    // Signal fuer den Erfolgreichen Start und den Annahmeversuch setzen.
    setCommunicationThreadStartSignal( true );

    if ( connection().socket() == INVALID_SOCKET )
    {
        trace( "%s(%i):ERROR(1): %s [ accept fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
        // Fehler setzen.
        setCommunicationThreadError( IV_ERROR_FAIL );
    }
    else
    {
        // Verbindungsmeldung ausgeben
        trace( "%s(%i):INFO(1): %s [ (%i) Connected to %s, port %d ]\n", __FILE__, __LINE__, __func__, connectionNumber(), inet_ntoa( sa.sin_addr ), htons( sa.sin_port ) );

        // Mindesttimeout sicherstellen
        if ( connection().readTimeout() < 3000 )
            connection().setReadTimeout( 3000 );

        // Socket zugewiesen
        connection().setType( ConnectionBase::ConnectionType::socket );

        // Verbindungsentgegennahme laeuft.
        setActiveModeState( ams_detecting );
        trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_detecting ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

        // Grunddaten zur Verbindung abfragen
        for ( failCount = 0; failCount < 4; ++failCount )
        {
            // Paketabarbeitung einleiten
            if ( (res = ping( 254, nullptr, nullptr )) != IV_ERROR_SUCCESS )
            {
                if ( (res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2) )
                {
                    // Wenn Daten gemeldet werden,
                    // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                    ModuleTable::closeConnection();

                    trace( "%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

                    // Fehler setzen.
                    setCommunicationThreadError( IV_ERROR_FAIL );

                    // Abbruch, zu viele Fehlversuche oder Gegenstelle hat geschlossen.
                    break;
                }
            }
            else
            {
                // Atwort auf Ping erhalten. Verbindung OK...
                break;
            }
        }

        if ( communicationThreadError() == IV_ERROR_SUCCESS && communicationThreadStopSignal() == false )
        {
            // com.                  Praefix fuer Kommunikationsvariablen
            // active                Aktivieren der Aktiven Verbindung
            // notify                notify => Aktivieren der Datensatzmeldung
            // prio                  Prioritaet des Geraetes
            // host                  Hostname des Remoterechners
            // port                  Portnummer des Remoterechners
            // retry                 Anzahl Versuche fuer Verbindungsaufbau
            // timeout               Auszeit zwischen den Versuchsbloecken
            // repeat                Auszeit zur wiederholten Datensatzmeldung, wenn diese nicht abgerufen wurden.
            // alive                 Pinganfrage wenn fuer diese Zeit keine Kommunikation zustande kam.
            // Alle Zeitangaben in Sekunden und 32bit

            std::string val;
            std::string notify;
            std::string status;
            uint32_t serial;

            val = std::to_string( 3 + static_cast<int32_t>(ceil( static_cast<double>(m_aliveTimeout) * 1.05 )) ); // 5% fuer Sync. dazugeben
            // Aliveparameter berechnen und setzen.
            // Geraetetyp und Seriennummer als GeraeteID abfragen
            if ( (res = writeSystemVariable( 254, "COM.ALIVE", val )) != IV_ERROR_SUCCESS ||
                 (res = readSerialNumber( 254, serial )) != IV_ERROR_SUCCESS ||
                 (res = readSystemVariable( 254, "COM.NOTIFY", notify )) != IV_ERROR_SUCCESS ||
                 (res = readFirmwareVersion( 254, m_versionFirmware )) != IV_ERROR_SUCCESS ||
                 // Der Systemstatus wird nur bei aktivem Trace gelesen. Bei aktivem Produktschutz wird er nicht geliefert.
                 (connection().traceMode() >= TraceBase::TraceMode::on && (res = readSystemVariable( 254, "SYSTEM.STATUS", status )) != IV_ERROR_SUCCESS && res != 28) )
            {
                // Verbindung schliessen
                ModuleTable::closeConnection();

                trace( "%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

                // Fehler setzen.
                setCommunicationThreadError( IV_ERROR_FAIL );
            }
            else
            {
                // Gelesene Seriennummer setzen.
                setProperty( "device.serial", serial );

                // Geraetetyp ermitteln.
                SoftwareVersion versionFirmware( m_versionFirmware );
                m_hardwareInfo.typeDevice = versionFirmware.typeDevice();

                // Melden von Datensaetzen
                try
                {
                    m_notify = std::stoi( notify );
                }
                catch ( ... )
                {
                    m_notify = 0;
                }

                // Bereits ermittelte Info ausgeben
                trace( "%s(%i):INFO(1): %s [ getTypeDevice( %i ), getSerial( %u ) ]\n", __FILE__, __LINE__, __func__, m_hardwareInfo.typeDevice, property( "device.serial", -1 ) );
                // Systemstatus auswerten
                if ( connection().traceMode() >= TraceBase::TraceMode::on && status.size() )
                {
                    char *token = strtok( &status[0], "\r" );
                    while ( token != nullptr )
                    {
                        trace( "%s(%i):INFO(1): %s [ SystemStatusLine( %s ) ]\n", __FILE__, __LINE__, __func__, token );
                        token = strtok( nullptr, "\r" );
                    }
                }

                // Daten fuer das TCPIP-Modul ermitteln
                m_tcpipInfo.reset();

                // Informationen zum TCP/IP abrufen
                // Erst nachtraeglich hinzugefuegt, deswegen separat ausfuehren.
                if ( (res = readModuleInfo( 254, "COM.TCPIP", "", m_tcpipInfo.data )) != IV_ERROR_SUCCESS )
                {
                    // Nur der Fehler "Majorname unbekannt" ist hier gueltig. Dann wird der Befehl noch nicht unterstuetzt.
                    if ( (res & 0xFF) != 1 )
                    {
                        // Wenn Daten gemeldet werden,
                        // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                        ModuleTable::closeConnection();

                        trace( "%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

                        // Fehler setzen.
                        setCommunicationThreadError( IV_ERROR_FAIL );
                    }
                }
                else
                {
                    // Werte aufschluesseln
                    m_tcpipInfo.parse();

                    // Bereits ermittelte Info ausgeben
                    trace( "%s(%i):INFO(1): %s [ GetInfoTCPIP( %i ), ModulVersion( %s ), IP( %s ), Port( %i ), MAC( %s ) ]\n", __FILE__, __LINE__, __func__, connectionNumber(), m_tcpipInfo.version.c_str(), m_tcpipInfo.ip.c_str(), m_tcpipInfo.port, m_tcpipInfo.mac.c_str() );
                }

                // Gewuenschte Zusatzinfos abrufen
                if ( m_infoFlags & 0x01 )
                {

                }
            }

            // Bedingung fuer Start der Hauptschleife erfuellt.
            if ( communicationThreadError() == IV_ERROR_SUCCESS && communicationThreadStopSignal() == false )
            {
                // Ermittelbare Daten zur Verbindung erfolgreich ermittelt.
                setActiveModeState( ams_usable );
                trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_usable ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

                // Hauptschleife ausfuehren.
                run = true;
            }
        }
    }

    // Fehlerzaehler neu initialisieren
    failCount = 0;

    // Hauptschleife des Worker
    while ( run )
    {
        // Beendigungssignal abfragen
        if ( communicationThreadStopSignal() )
        {
            // Verbindung herunterfahren
            if ( connection().socket() != INVALID_SOCKET )
            {
                // Befehlsausfuehrung geschuetzt durch Mutex.
                lockGuardCall guard( m_mutexCall );

                // Senden am Socket verhindern
                if ( shutdown( connection().socket(), 0x01/*SD_SEND*/ ) == SOCKET_ERROR )
                {
                    trace( "%s(%i):ERROR(1): %s [ shutdown fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                }
                else
                {
                    // Alle anstehenden Daten lesen
                    do
                    {

                        FD_ZERO( (fd_set FAR *) & readfds );
                        FD_SET( connection().socket(), (fd_set FAR *) & readfds );

                        if ( (res = select( FD_SETSIZE,
                                            (fd_set FAR *) & readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                                            (MYCONST struct timeval FAR *) & seltime )) == SOCKET_ERROR )
                        {
                            trace( "%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                        }
                        else if ( res )
                        {
                            char packet_receive[IV_PACKET_LENGTH_MAX];
                            // Anstehende Daten lesen
                            if ( (res = recv( connection().socket(), packet_receive, sizeof( packet_receive ), 0 )) == SOCKET_ERROR )
                            {
                                trace( "%s(%i):ERROR(1): %s [ recv fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );
                            }
                            else if ( res == 0 )
                            {
                                // Disconnect von Gegenstelle
                            }
                        }
                    }
                    while ( (res > 0) && (res != SOCKET_ERROR) );
                }

                // Verbindung schliessen
                ModuleTable::closeConnection();
            }

            // Hauptschleife verlassen
            run = false;
            continue;
        }

        // Eigener Block wegen MutexCall
        {
            // Befehlsausfuehrung geschuetzt durch Mutex.
            lockGuardCall guard( m_mutexCall );

            // Socketdaten initialisieren
            FD_ZERO( (fd_set FAR *) & readfds );
            FD_SET( connection().socket(), (fd_set FAR *) & readfds );

            // Auf eintreffende Daten warten
            if ( (res = select( FD_SETSIZE,
                                (fd_set FAR *) & readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                                (MYCONST struct timeval FAR *) & seltime )) != SOCKET_ERROR )
            {
                // Stehen Daten an? (ansonsten Timeout)
                if ( res )
                {
                    // Paketabarbeitung einleiten
                    if ( (res = call( IV_CONTEXT_SLAVE, IV_MODULE_BASE, IV_FUNCTION_BASE, data, count, 1 )) != IV_ERROR_SUCCESS )
                    {
                        if ( (res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2) )
                        {
                            // Wenn Daten gemeldet werden,
                            // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                            ModuleTable::closeConnection();

                            trace( "%s(%i):INFO(1): %s [ Disconnect ]\n", __FILE__, __LINE__, __func__, connectionNumber(), inet_ntoa( sa.sin_addr ), htons( sa.sin_port ) );

                            // Fehler setzen.
                            setCommunicationThreadError( IV_ERROR_FAIL );

                            // Hauptschleife verlassen
                            run = false;
                            continue;
                        }

                        // Fehlerzaehler erhoehen
                        failCount++;
                    }
                    else
                    {
                        // Fehlerzaehler zuruecksetzen
                        failCount = 0;
                    }
                }
            }
            else
            {
                trace( "%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError() );

                // Verbindung schliessen
                ModuleTable::closeConnection();

                trace( "%s(%i):INFO(1): %s [ Disconnect ]\n", __FILE__, __LINE__, __func__, connectionNumber(), inet_ntoa( sa.sin_addr ), htons( sa.sin_port ) );

                // Fehler setzen.
                setCommunicationThreadError( IV_ERROR_FAIL );

                // Hauptschleife verlassen
                run = false;
                continue;
            }

            // Aliveping senden?
            if ( (protocol().timeReadPacket() + m_aliveTimeout) < time( nullptr ) )
            {
                // Ping
                if ( (res = ping( 254, nullptr, nullptr )) != IV_ERROR_SUCCESS )
                {
                    if ( (res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2) )
                    {
                        // Wenn Daten gemeldet werden,
                        // die Anzahl aber 0 ist, so wurde die Verbindung beendet.
                        ModuleTable::closeConnection();

                        trace( "%s(%i):INFO(1): %s [ Disconnect (%i) ]\n", __FILE__, __LINE__, __func__, connectionNumber() );

                        // Fehler setzen.
                        setCommunicationThreadError( IV_ERROR_FAIL );

                        // Hauptschleife verlassen
                        run = FALSE;
                        continue;
                    }

                    // Fehlerzaehler erhoehen
                    failCount++;
                }
                else
                {
                    // Fehlerzaehler zuruecksetzen
                    failCount = 0;
                }

                trace( "%s(%i):INFO(1): %s [ Alive - Ping (%i) %ds ]\n", __FILE__, __LINE__, __func__, connectionNumber(), int32_t( time( nullptr ) - protocol().timeReadPacket() ) );
            }
        }

        // Kurze Ausszeit um die Systemlast zu reduzieren.
        std::this_thread::sleep_for( std::chrono::milliseconds( 10 ) );
    }

    // Eventuellen Eintrag aus Benachrichtungsschlange entfernen.
    if ( m_callBackNotifyRecordAvailable )
    {
        CNotifyRecord nr;

        // Meldungen dieses Kanals loeschen
        nr.erase = 1;

        // Aktuelle Kanalnummer eintragen
        nr.channelNumber = connectionNumber();

        // Meldung aus Warteschlange entfernen.
        m_callBackNotifyRecordAvailable( nr );
    }

    // Kanalobjekt wieder frei
    setActiveModeState( ams_unused );
    trace( "%s(%i):INFO(1): %s [ setActiveModeState(%i) is_none ]\n", __FILE__, __LINE__, __func__, connectionNumber() );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
int32_t Channel4::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue )
    // Modulbefehl senden und Antwort zurueckgeben
    //  deviceAddress -> Adresse (Geraetenummer) des Empfaengers
    //  module        -> Nummer des anzusprechenden Modules
    //  func          -> Nummer der Funktion im Module
    //  buf           -> Buffer fuer Funktionsparameter.
    //  count         -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
    //
    //  return -> <0 -> Kommunikationsfehler
    //             0 -> senden und empfangen und ausfuehren erfolgreich
    //            >0 -> module-/Funktionsfehler
{
    trace( "%s(%i):INFO(3): %s [ %u, %u, %u, %p, %u, %d ]\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, contextValue );

    return ModuleSystem::call( context, module, func, buffer, count, contextValue );
}

int32_t Channel4::generateFingerprintIdList( uint8_t deviceAddress, int32_t flags, uint32_t &current_count )
{
    int32_t res;
    std::string name( "FINGERPRINT.IDLIST" );
    std::string reserved;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    if ( (res = generateTable( deviceAddress, name, reserved )) != IV_ERROR_SUCCESS )
    {
        // Liegen im Modul keine Daten vor?
        if ( res == 133 )
        {
            return -IV_ERROR_NO_DATA_AVAILABLE;
        }
    }

    // Flag fuer anschliessende Ermittlung der Datensatzanzahl gesetzt?
    if ( flags & 0x01 )
    {
        int32_t handle;
        uint32_t unsortedCount = 0;
        uint32_t deletedCount = 0;
        current_count = 0;

        // Erstellte Tabelle oeffnen.
        if ( (res = openTable( deviceAddress, name, 0, handle, reserved )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Anzahl enthaltener Datensaetze abrufen.
        if ( (res = readTableRowCount( deviceAddress, handle, current_count, unsortedCount, deletedCount )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Tabelle schliessen.
        if ( (res = closeTable( deviceAddress, handle )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    return res;
}

// Backup der Fingerprints erstellen
int32_t Channel4::fingerprintBackup( uint8_t deviceAddress, const std::string &filePath, int32_t flags )
{
    (void)flags;
    int32_t res = IV_ERROR_SUCCESS;
    std::string name( "FINGERPRINT.IDLIST" );
    std::string reserved;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
        return -IV_ERROR_NO_MODULE;

    // Tabelle mit Fingerprintdaten erzeugen
    if ( (res = generateTable( deviceAddress, name, reserved )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    int32_t hTable;
    // Tabelle mit Fingerprintdaten oeffnen
    if ( (res = openTable( deviceAddress, name, 0, hTable, reserved )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    uint32_t recordCount = 0;
    uint32_t unsortedCount = 0;
    uint32_t deletedCount = 0;
    // Pruefen ob Datensaetze vorliegen
    if ( (res = readTableRowCount( deviceAddress, hTable, recordCount, unsortedCount, deletedCount )) != IV_ERROR_SUCCESS )
    {
        // Modul oder Funktionsfehler?
        if ( res > 0 ) closeTable( deviceAddress, hTable );
        return res;
    }
    if ( recordCount == 0 )
    {
        // Es liegen keine Daten vor
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }
    FILE *fp;
    StructFingertemplateId Id;
    uint8_t type = 3; // Saturn01 zuerst verwenden.
    char data[IV_PACKET_LENGTH_DATA];
    uint16_t size;
    uint16_t crc;

    // Datei oeffnen
    if ( (fp = fopen( filePath.c_str(), "wb" )) == nullptr )
    {
        closeTable( deviceAddress, hTable );
        return -IV_ERROR_FILE_OPEN;
    }

    // Kopfstruktur vorbereiten
    StructFTFileHeader structHeader;
    memset( &structHeader, 0, sizeof( StructFTFileHeader ) );
    // Kopfdaten reservieren
    if ( fwrite( &structHeader, 1, sizeof( StructFTFileHeader ), fp ) < sizeof( StructFTFileHeader ) )
    {
        // Daten wurden nicht geschrieben
        closeTable( deviceAddress, hTable );
        fclose( fp );
        return -IV_ERROR_FILE_WRITE;
    }
    // Kopfdaten aufsetzen. Die Datei ist erst gueltig, wenn die Kopfdaten geschrieben wurden.
    structHeader.Id = 0xDFDF;
    structHeader.offs = sizeof( StructFTFileHeader );
    structHeader.count = recordCount;

    // Fingertemplates abrufen.
    do
    {
        // Datenspeicher loeschen
        size = static_cast<uint16_t>(sizeof( data ));
        memset( data, 0, sizeof( data ) );

        // Datensatz aus Liste abrufen
        if ( (res = readCurrentTableRowDataExt( deviceAddress, hTable, 0, 0, data, size )) != IV_ERROR_SUCCESS )
        {
            // Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
            if ( res > 0 )
            {
                closeTable( deviceAddress, hTable );
            }
            fclose( fp );
            return res;
        }
        // bekannte Daten in struktur uebernehmen
        memcpy( (uint8_t *)&Id, data, sizeof( StructFingertemplateId ) );

        // Datenspeicher loeschen
        size = static_cast<uint16_t>(sizeof( data ));
        memset( data, 0, sizeof( data ) );

        // Das dazugehoerige Fingertemplate abfragen
        while ( (res = readFingerprintRecord( deviceAddress, type, (uint32_t)atof( Id.pid ), atoi( Id.fid ), reinterpret_cast<uint8_t *>(data), size )) != IV_ERROR_SUCCESS )
        {
            // Wenn das Templateformat unbekannt oder nicht dem verwendeten entspricht, dann ggf. noch anderes versuchen.
            // Funktionsfehler 1 und 3
            if ( (res == 1 || res == 3) && type != 0 )
            {
                type--; // Nächstes Format prüfen.
                continue;
            }

            // Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
            if ( res > 0 ) closeTable( deviceAddress, hTable );
            fclose( fp );
            return res;
        }

        // Groesse ist abhaengig von verwendetem Templatetyp.
        switch ( type )
        {
            case 0:
                size = sizeof( StructFingertemplateDIN );
                break;

            case 1:
                size = sizeof( StructFingertemplateIdencom );
                break;

            case 2:
                size = sizeof( StructFingertemplateIdencomStandard );
                break;

            case 3:
                size = sizeof( StructFingertemplateSaturn01 );
                break;

            default:
                size = 0;
                break;
        }
        // CRC16 ueber das Fingertemplate erstellen
        crc = GetCRC16( reinterpret_cast<uint8_t *>(data), size );
        // Datensatz in die Datei wegschreiben
        if ( fwrite( data, 1, (size_t)size, fp ) < (size_t)size )
        {
            // Daten wurden nicht geschrieben
            closeTable( deviceAddress, hTable );
            fclose( fp );
            return -IV_ERROR_FILE_WRITE;
        }
        // CRC schreiben
        if ( fwrite( &crc, 1, sizeof( uint16_t ), fp ) < sizeof( uint16_t ) )
        {
            // Daten wurden nicht geschrieben
            closeTable( deviceAddress, hTable );
            fclose( fp );
            return -IV_ERROR_FILE_WRITE;
        }

        // Naechsten Datensatz abrufen
    }
    while ( (res = writeCurrentTableRow( deviceAddress, hTable, 1, 1 )) == IV_ERROR_SUCCESS );

 // Nicht: Sprung fuehrte zu EOF?
 // Funktionsfehler: 4
    if ( res != 4 )
    {
        // Modul oder Funktionsfehler?
        if ( res > 0 ) closeTable( deviceAddress, hTable );
        fclose( fp );
        return res;
    }

    // Headerstruktur abschliessen und schreiben
    structHeader.type = type;
    structHeader.crc = GetCRC16( (uint8_t *)&structHeader, sizeof( structHeader ) - 2 );

    // Auf Dateianfang und Headerstruktur schreiben
    if ( fseek( fp, 0, 0 ) != 0 || fwrite( &structHeader, 1, sizeof( structHeader ), fp ) < sizeof( structHeader ) )
    {
        // Daten wurden nicht geschrieben
        closeTable( deviceAddress, hTable );
        fclose( fp );
        return -IV_ERROR_FILE_WRITE;
    }

    // Tabelle schliessen
    closeTable( deviceAddress, hTable );

    // Datei schliessen
    fclose( fp );

    return IV_ERROR_SUCCESS;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int32_t Channel4::fingerprintRestore( uint8_t deviceAddress, const std::string &filePath, int32_t flags )
{
    int32_t res = IV_ERROR_SUCCESS;
    FILE *fp;
    StructFingertemplateId Id;
    uint8_t data[IV_PACKET_LENGTH_DATA];
    size_t size, sizeCheck;
    uint16_t crc, crcSegment;
    int32_t recordSkiped = 0;
    int32_t recordFailed = 0;
    std::string name( "FINGERPRINT.IDLIST" );
    std::string reserved;
    uint32_t currentCountAfter = 0;
    uint32_t unorderedCountAfter = 0;
    uint32_t deletedCountAfter = 0;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Biokey verbaut ist.
    tags = hardwareInfo.moduleBiokey();
    if ( tags.size() == 0 )
        return -IV_ERROR_NO_MODULE;

    // Datei oeffnen
    if ( (fp = fopen( filePath.c_str(), "rb" )) == nullptr )
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Kopfstruktur vorbereiten
    StructFTFileHeader structHeader;
    memset( &structHeader, 0, sizeof( StructFTFileHeader ) );
    // Kopfdaten lesen
    if ( fread( &structHeader, 1, sizeof( StructFTFileHeader ), fp ) < sizeof( StructFTFileHeader ) )
    {
        // Daten wurden nicht gelesen
        fclose( fp );
        return -IV_ERROR_FILE_READ;
    }
    // Kopfdaten pruefen.
    crc = GetCRC16( (uint8_t *)&structHeader, sizeof( StructFTFileHeader ) );
    if ( structHeader.Id != 0xDFDF ||
         crc != 0 ||
         structHeader.type > 3 )
    {
        fclose( fp );
        return -IV_ERROR_FILE_FORMAT;
    }
    // Auf Datenanfang wechseln
    fseek( fp, structHeader.offs, 0 );

    // Groesse ist abhaengig von verwendetem Templatetyp.
    switch ( structHeader.type )
    {
        case 0:
            size = sizeof( StructFingertemplateDIN );
            break;

        case 1:
            size = sizeof( StructFingertemplateIdencom );
            break;

        case 2:
            size = sizeof( StructFingertemplateIdencomStandard );
            break;

        case 3:
            size = sizeof( StructFingertemplateSaturn01 );
            break;

        default:
            size = 0;
    }

    // Datensaetze loeschen?
    if ( (flags & 1) == 1 )
    {
        // Alle Datensaetze vom Modul loeschen
        if ( (res = deleteFingerprintRecord( deviceAddress, 0, 0 )) != IV_ERROR_SUCCESS )
        {
            fclose( fp );
            return res;
        }
    }

    // Fingertemplates einspielen.
    while ( !feof( fp ) && !ferror( fp ) )
    {
        // Datenspeicher loeschen
        memset( data, 0, sizeof( data ) );
        memset( &Id, 0, sizeof( StructFingertemplateId ) );

        // Template aus Datei laden
        if ( (sizeCheck = fread( data, 1, size, fp )) < size )
        {
            if ( sizeCheck == 0 && feof( fp ) )
            {
                continue;
            }
            else if ( sizeCheck && !ferror( fp ) )
            {
                fclose( fp );
                return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
            }
            // Daten wurden nicht gelesen
            fclose( fp );
            return -IV_ERROR_FILE_READ;
        }
        // CRC16 des Fingertemplates erstellen
        crc = GetCRC16( data, (uint32_t)size );
        // CRC lesen
        if ( (sizeCheck = fread( &crcSegment, 1, sizeof( uint16_t ), fp )) < sizeof( uint16_t ) )
        {
            if ( (sizeCheck == 0 && feof( fp )) || (sizeCheck && !ferror( fp )) )
            {
                fclose( fp );
                return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
            }
            // Daten wurden nicht gelesen
            fclose( fp );
            return -IV_ERROR_FILE_READ;
        }
        // CRC pruefen
        if ( crc != crcSegment )
        {
            fclose( fp );
            return -IV_ERROR_FILE_CRC;
        }

        // Das Fingertemplate anfuegen
        if ( (res = appendFingerprintRecord( deviceAddress, structHeader.type, data, static_cast<uint16_t>(size) )) != IV_ERROR_SUCCESS )
        {
            // Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
            // Funktionsfehler 134
            if ( res == 134 )
            {
                recordSkiped++;
                continue;
            }
            //else if (res == 132)
            //{
            //    // Speicher des Moduls ist voll
            //    res = -IV_ERROR_MEMORY_FULL;
            //}
            else if ( res == 1 || res == 3 )
            {
                // Falsches Templateformat
                res = -IV_ERROR_FILE_DATA_FORMAT;
            }
            else if ( res == 2 )
            {
                int32_t failLoop = 0;
                // Bei aelterer Hardware des Biokey3000 kann eine Verzoegerung nach dem loeschen der Datensaetze
                // vorkommen. Hier wird versucht das Template nochmals zu uebertragen.
                for ( ; failLoop < 2; failLoop++ )
                {
                    if ( (res = appendFingerprintRecord( deviceAddress, structHeader.type, data, static_cast<uint16_t>(size) )) == 2 )
                    {
                        // Kleine Verzoegerung
                        //m_channel4.Trace("ERROR: GetInfoVerzoegerung: %i\n", failLoop);
                        std::this_thread::sleep_for( std::chrono::seconds( 2 ) );
                        continue;
                    }
                    break;
                }

                // Konnte der Befehl ausgefuehrt werden?
                // Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
                // Funktionsfehler 134
                if ( res == 134 )
                {
                    recordSkiped++;
                    continue;
                }
                else if ( res == 132 )
                {
                    // Speicher des Moduls ist voll
                    res = -IV_ERROR_MEMORY_FULL;
                }
                else if ( res == 1 || res == 3 )
                {
                    // Falsches Templateformat
                    res = -IV_ERROR_FILE_DATA_FORMAT;
                }
                else if ( res == IV_ERROR_SUCCESS )
                {
                    //m_channel4.Trace("OK: Verzoegerung: %i\n", failLoop);
                    continue;
                }
            }
            else if ( res > 0 )
            {
                // B.Ottmann 15.11.2019
                //   Es wird das Standardverhalten so angepasst, dass Fehler der Applikationsschicht weitestgehend ignoriert werden und
                //   versucht wird alle möglichen Templates zu übertragen. Unter Rücksprache mit S.Meyer so entschieden.
                ++recordFailed;
                continue;
            }

            fclose( fp );
            return res;
        }
    }



    // Nicht: Sprung fuehrte zu EOF?
    if ( ferror( fp ) )
    {
        fclose( fp );
        return -IV_ERROR_FILE_READ;
    }

    // Datei schliessen
    fclose( fp );

    // Tabelle mit Fingerprintdaten erzeugen
    if ( (res = generateTable( deviceAddress, name, reserved )) != IV_ERROR_SUCCESS )
    {
        // Liegen im Modul keine Daten vor?
        if ( res == 133 )
        {
            res = -IV_ERROR_NO_DATA_AVAILABLE;
        }
        return res;
    }
    int32_t hTable;
    // Tabelle mit Fingerprintdaten oeffnen
    if ( (res = openTable( deviceAddress, name, 0, hTable, reserved )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Pruefen ob Datensaetze vorliegen
    if ( (res = readTableRowCount( deviceAddress, hTable, currentCountAfter, unorderedCountAfter, deletedCountAfter )) != IV_ERROR_SUCCESS )
    {
        // Modul oder Funktionsfehler?
        if ( res > 0 )
        {
            closeTable( deviceAddress, hTable );
        }

        return res;
    }
    // Tabelle schliessen
    if ( (res = closeTable( deviceAddress, hTable )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Pruefen ob die Datensatzanzahl passt.
    if ( structHeader.count > currentCountAfter )
    {
        // Es muessen mindestens die Anzahl geforderter Datensaetze vorhanden sein.		
    }

    // Falls beim Anfügen von Datensätzen ein Fehler aufgetreten ist, dann diesen allgemeinen Fehler zurückliefern.
    if ( recordFailed )
    {
        return -IV_ERROR_FINGERPRINT;
    }

    return IV_ERROR_SUCCESS;
}

int32_t Channel4::setOptionFirmware( uint8_t deviceAddress, int32_t mask, int32_t options )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Externe Leserserie
    if ( mask & OFF_EXTERNAL_READER )
    {
        if ( (res = writeExternalTransponderReaderType( deviceAddress, (options & OFF_EXTERNAL_READER) ? 2 : 1 )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    // IOModule
    if ( mask & OFF_IOMODULE )
    {
        if ( (res = writeIoModuleOption( deviceAddress, (options & OFF_IOMODULE) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Timeboy anbindung
    if ( mask & OFF_TBADAPTION )
    {
        if ( (res = writeTimeboyAdaptionOption( deviceAddress, (options & OFF_TBADAPTION) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // ComServer
    if ( mask & OFF_COMSERVER )
    {
        if ( (res = writeComServerOption( deviceAddress, (options & OFF_COMSERVER) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Transponder schreiben
    if ( mask & OFF_TPWRITE )
    {
        if ( (res = writeTransponderWriteOption( deviceAddress, (options & OFF_TPWRITE) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // HTTP-Protokoll
    if ( mask & OFF_HTTP )
    {
        if ( (res = writeHttpOption( deviceAddress, (options & OFF_HTTP) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Biokey update
    if ( mask & OFF_BIOKEY_UPDATE )
    {
        if ( (res = writeFingerprintScannerUpdateOption( deviceAddress, (options & OFF_BIOKEY_UPDATE) ? 1 : 0 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // SetupOption: Signalverarebeitung
    if ( mask & OFF_SO_IOCONTROL )
    {
        if ( (res = writeSetupOptions( deviceAddress, 0x0001, (options & OFF_SO_IOCONTROL) ? 0x0001 : 0x0000 )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 3 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }
    }

    // Systemflag fuer Setup und Listen beibehalten
    if ( mask & OFF_KEEP_SETUP_AFTER_UPDATE )
    {
        uint32_t flags = 0;
        // Systemflags
        if ( (res = readSystemFlags( deviceAddress, flags )) != IV_ERROR_SUCCESS )
        {
            // Erst ab Version 2 verfuegbar
            if ( ((res >> 8) & 0xFF) == 2 )
            {
                // Die weitere Verarbeitung abbrechen.
                return IV_ERROR_SUCCESS;
            }

            return res;
        }

        bool should_keep_setup_after_update = (options & OFF_KEEP_SETUP_AFTER_UPDATE) == OFF_KEEP_SETUP_AFTER_UPDATE;
        bool is_keep_setup_after_update = (flags & IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS) != IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;

        if ( should_keep_setup_after_update != is_keep_setup_after_update )
        {
            flags &= ~IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;
            flags |= should_keep_setup_after_update ? 0 : IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS;
            if ( (res = writeSystemFlags( deviceAddress, flags )) != IV_ERROR_SUCCESS )
            {
                // Erst ab Version 2 verfuegbar
                if ( ((res >> 8) & 0xFF) == 2 )
                {
                    // Die weitere Verarbeitung abbrechen.
                    return IV_ERROR_SUCCESS;
                }

                return res;
            }
        }
    }

    return res;
}

int32_t Channel4::getOptionFirmware( uint8_t deviceAddress, int32_t &mask, int32_t &options )
{
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t val;

    // Zuruecksetzen der Werte
    mask = options = 0;

    // Externe Leserserie
    if ( (res = readExternalTransponderReaderType( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    mask |= OFF_EXTERNAL_READER;
    options |= (val == 2) ? OFF_EXTERNAL_READER : 0;

    // IOModule
    if ( (res = readIoModuleOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    mask |= OFF_IOMODULE;
    options |= (val == 1) ? OFF_IOMODULE : 0;

    // Timeboy anbindung
    if ( (res = readTimeboyAdaptionOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    mask |= OFF_TBADAPTION;
    options |= (val == 1) ? OFF_TBADAPTION : 0;

    // ComServer
    if ( (res = readComServerOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_COMSERVER;
    options |= (val == 1) ? OFF_COMSERVER : 0;

    // Transponder schreiben
    if ( (res = readTransponderWriteOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_TPWRITE;
    options |= (val == 1) ? OFF_TPWRITE : 0;

    // Http-Protokoll
    if ( (res = readHttpOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_HTTP;
    options |= (val == 1) ? OFF_HTTP : 0;

    // Biokey Update
    if ( (res = readFingerprintScannerUpdateOption( deviceAddress, val )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_BIOKEY_UPDATE;
    options |= (val == 1) ? OFF_BIOKEY_UPDATE : 0;

    uint16_t soMask, soValue;
    // Setup Options: Signalverarbeitung
    if ( (res = readSetupOptions( deviceAddress, soMask, soValue )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_SO_IOCONTROL;
    options |= (soValue & 0x0001) ? OFF_SO_IOCONTROL : 0;

    uint32_t flags = 0;
    // Systemflags
    if ( (res = readSystemFlags( deviceAddress, flags )) != IV_ERROR_SUCCESS )
    {
        // Erst ab Version 2 verfuegbar
        if ( ((res >> 8) & 0xFF) == 2 )
        {
            // Die weitere Verarbeitung abbrechen.
            return IV_ERROR_SUCCESS;
        }

        return res;
    }
    mask |= OFF_KEEP_SETUP_AFTER_UPDATE;
    options |= ((flags & IV_VALUE_SYSTEM_FLAG_DEL_SETUP_AND_LISTS) == 0) ? OFF_KEEP_SETUP_AFTER_UPDATE : 0;

    return res;
}

int32_t Channel4::setWlanConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &file_path )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setWlanConfigData( deviceAddress, file_path );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::setWlanConfigData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    bool hasLAN = hardwareInfo.moduleLANCommunication().size() || hardwareInfo.onBoardLANCommunication().size();
    bool hasWLAN = hardwareInfo.moduleWLANCommunication().size();
    // Wenn weder LAN noch WLAN verbaut ist.   
    if ( !hasLAN && !hasWLAN )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Bei V3 Geräten nur mit bestimmtem WLAN-Modul möglich.
    if ( hardwareInfo.isMasterIV() )
    {
        // Als Aufsteckmodul auf dem Modemsteckplatz.
        if ( hardwareInfo.moduleWLANCommunication( { EnumModuleWLANCommunication::WLAN_RS9110_Plug } ).size() == 0 )
        {
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }

    // WlanConfig Daten
    FileWlanConfig fileWlanConfig;
    uint32_t address = 0;
    // WLAN Konfigurationsdatei laden
    if ( !fileWlanConfig.load( filePath.c_str() ) )
    {
        switch ( fileWlanConfig.errorState() )
        {
            case FileWlanConfig::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileWlanConfig::Error_state::file_format:
            default:
                return -IV_ERROR_FILE_FORMAT;
        }
    }

    if ( !fileWlanConfig.WlanConfigCheckCRC( -1 ) )
    {
        return -IV_ERROR_FILE_CRC;
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind WLAN Daten vorhanden?
    if ( fileWlanConfig.sizeData() )
    {
        address = 0;
        auto data = fileWlanConfig.data();
        auto sizeData = fileWlanConfig.sizeData();
        auto crc = fileWlanConfig.checksum32bit();

        // Systemstopp setzen
        res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_WLANCONFIGS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_WLANCONFIGS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // WLAN Daten erfolgreich uebertragen
    return res;
}

int32_t Channel4::getWlanConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = getWlanConfigData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::getWlanConfigData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    bool hasLAN = hardwareInfo.moduleLANCommunication().size() || hardwareInfo.onBoardLANCommunication().size();
    bool hasWLAN = hardwareInfo.moduleWLANCommunication().size();
    // Wenn weder LAN noch WLAN verbaut ist.   
    if ( !hasLAN && !hasWLAN )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Bei V3 Geräten nur mit bestimmtem WLAN-Modul möglich.
    if ( hardwareInfo.isMasterIV() )
    {
        // Als Aufsteckmodul auf dem Modemsteckplatz.
        if ( hardwareInfo.moduleWLANCommunication( { EnumModuleWLANCommunication::WLAN_RS9110_Plug } ).size() == 0 )
        {
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }

    //
    FileWlanConfig fileWlanConfig;
    uint32_t address = 0;
    uint32_t size = 0;
    uint32_t crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        if ( (res & 0xFF) == 1 )
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

    fileWlanConfig.resetData();
    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_WLANCONFIGS, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_WLANCONFIGS, address, size, crc );


    //////////////////////////////
    // Daten
    fileWlanConfig.setSizeData( size );
    // EEPROM uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileWlanConfig.data(), fileWlanConfig.sizeData() )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_WLANCONFIGS )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_WLANCONFIGS );

    // Segmentend: Bedienung stoppen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    // Datenanzahl in Datei uebernehmen.
    auto crcTest = fileWlanConfig.checksum32bit();
    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Datei speichern
    if ( !fileWlanConfig.save( filePath.c_str() ) )
    {
        trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FILE_WRITE;
    }

    return res;
}

int32_t Channel4::setHardwareConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setHardwareConfigData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::setHardwareConfigData( uint8_t deviceAddress, const std::string &filePath )
{
    // HardwareConfig Daten
    HardwareInfo hardwareInfo;
    FileHardwareInfo &fileHardwareConfig = hardwareInfo.fileHardwareInfo();
    int32_t res = IV_ERROR_SUCCESS;
    uint32_t address = 0;
    uint8_t *buf;
    uint32_t size = 0;
    uint32_t crc = 0;

    // Hardware Konfigurationsdatei laden
    if ( !fileHardwareConfig.load( filePath ) )
    {
        switch ( fileHardwareConfig.errorState() )
        {
            case FileHardwareInfo::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileHardwareInfo::Error_state::file_format:
            default:
                return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Handelt es sich um eine MasterIV - HWI Datei, dann Abbruch mit Fehler.
    if ( hardwareInfo.isMasterIV() )
    {
        trace( "%s(%i):ERROR(1): %s [ Es handelt sich um eine HIP-Datei für MasterIV. ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }

    //	memset(version, 0, sizeof(version));
    // Bootversion lesen 
    //	if ((res = m_channel4.GetVersionBootloader(deviceAddress, version)) != IV_ERROR_SUCCESS)
    //{
    //		return res;
    //	}

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind HardwareConfig Daten vorhanden?
    if ( fileHardwareConfig.sizeData() )
    {
        address = 0;
        buf = const_cast<uint8_t *>(fileHardwareConfig.data());
        size = fileHardwareConfig.sizeData();
        crc = fileHardwareConfig.checksum32bit();

        // Systemstopp setzen
        res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_HARDWARECONFIGS, address, size, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, buf, size )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_HARDWARECONFIGS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    crc = 0;
    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Hardware Daten erfolgreich uebertragen
    return res;
}

int32_t Channel4::getHardwareConfigDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = getHardwareConfigData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::getHardwareConfigData( uint8_t deviceAddress, const std::string &filePath )
{
    HardwareInfo hardwareInfo;
    int32_t res = IV_ERROR_SUCCESS;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) == IV_ERROR_SUCCESS )
    {
        // Datei speichern
        if ( !hardwareInfo.fileHardwareInfo().save( filePath ) )
        {
            trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_FILE_WRITE;
        }
    }

    return res;
}

int32_t Channel4::writeFileDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeFileData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeFileData( uint8_t deviceAddress, const std::string &filePath )
{
    // HardwareConfig Daten
    int32_t res = IV_ERROR_SUCCESS;
    int32_t errorApplication = IV_ERROR_SUCCESS;

    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        struct stat fileStat;
        std::size_t pos = filePath.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = filePath;
        }
        else
        {
            fileName = filePath.substr( pos + 1 );
        }

        if ( stat( filePath.c_str(), &fileStat ) != 0 )
        {
            return -IV_ERROR_FILE_OPEN;
        }

        FILE *fp = fopen( filePath.c_str(), "rb" );
        // Datei oeffnen
        if ( fp == nullptr )
        {
            return -IV_ERROR_FILE_OPEN;
        }

        controllerSegmentCurrent.fileData.resize( fileStat.st_size );
        if ( fread( &controllerSegmentCurrent.fileData[0], sizeof( uint8_t ), fileStat.st_size, fp ) != static_cast<size_t>(fileStat.st_size) )
        {
            fclose( fp );
            trace( "%s(%i):ERROR(1): %s [ Lesen der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            res = -IV_ERROR_FILE_READ;
            goto label_end;
        }

        fclose( fp );

        if ( (res = writeSystemVariable( deviceAddress, "SYSTEM.FILETRANSFER", fileName )) != IV_ERROR_SUCCESS )
        {
            // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
            // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
            if ( (res & 0xFF) == 3 )
            {
                // Fehler liegt bei Wert
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( (res & 0xFF) == 2 )
            {
                // Variable nicht bekannt
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            goto label_end;
        }

        // Ggf. gesetztes Streaming deaktivieren.
        setStreamTransfer( false );
        clearStreamTransfer();

        uint32_t unsusedUInt = 0;
        // Systemstopp setzen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, unsusedUInt, 0, 0 )) != IV_ERROR_SUCCESS )
        {
            goto label_end;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::active;
        controllerSegmentCurrent.transferType = ControllerSegment::btt_write_file;
        if ( controllerSegmentCurrent.allowedTransferTypes == ControllerSegment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            ControllerSegment &controllerSegmentGlobal = controllerSegment( 255 );
            if ( controllerSegmentGlobal.allowedTransferTypes != ControllerSegment::btt_none )
            {
                controllerSegmentCurrent.allowedTransferTypes = controllerSegmentGlobal.allowedTransferTypes;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transferFlags.settingsFromGlobal = true;;
            }
        }
        controllerSegmentCurrent.beginTimePoint = std::chrono::system_clock::now();

        controllerSegmentCurrent.address = 0;
        controllerSegmentCurrent.crc = crc32( &controllerSegmentCurrent.fileData[0], static_cast<uint32_t>(controllerSegmentCurrent.fileData.size()) );

        // Systemstopp setzen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FILETRANSFER, controllerSegmentCurrent.address,
                                              static_cast<uint32_t>(controllerSegmentCurrent.fileData.size()), controllerSegmentCurrent.crc )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktion wird von Firmware nicht unterstuetzt.    
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            goto label_end;
        }
    }

    if ( controllerSegmentCurrent.discard == false )
    {
        // Dateidaten uebertragen
        if ( (res = writeToFlash( deviceAddress, controllerSegmentCurrent.address + controllerSegmentCurrent.size,
                                    &controllerSegmentCurrent.fileData[controllerSegmentCurrent.size],
                                    static_cast<uint32_t>(controllerSegmentCurrent.fileData.size() - controllerSegmentCurrent.size) )) != IV_ERROR_SUCCESS )
        {
            if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
            {
                controllerSegmentCurrent.resume = true;
                return res;
            }

            // Bei Kommunikationsfehlern abbrechen,
            // bei Ausfuehrungsfehlern das Uebertragen abbrechen.
            if ( res < IV_ERROR_SUCCESS )
            {
                goto label_end;
            }
            else
            {
                errorApplication = res;
            }
        }
    }

    // Dateidaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FILETRANSFER, controllerSegmentCurrent.crc )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE_BG, 0 )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }

label_end:
    if ( controllerSegmentCurrent.fileData.size() )
    {
        controllerSegmentCurrent.fileData.clear();
    }

    // Blocktransfer beendet.
    controllerSegmentCurrent.endTimePoint = std::chrono::system_clock::now();
    controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::inactive;

    // Ggf. aufgetretenen Fehler bei setFlash zurueckgeben.
    if ( errorApplication != IV_ERROR_SUCCESS )
    {
        res = errorApplication;
    }

    return res;
}

int32_t Channel4::readFileDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = readFileData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

/*! Lesen von Dateidaten und speichern in der Datei unter file_path.

\param deviceAddress Geraeteaddresse des anzusprechenden Geraetes.
\param file_path Dateiname + Pfad unter dem die Datei abgespeichert wird, der Dateiname selbst wird gelesen.

\retval IV_ERROR_SUCCESS Erfolgreich ausgelesen und abgespeichert.
\retval > IV_ERROR_SUCCESS Fehler auf der Anwendungsschicht der Firmware.
\retval < IV_ERROR_SUCCESS Fehler in der Kommunikation.
*/
int32_t Channel4::readFileData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    int32_t errorApplication = IV_ERROR_SUCCESS;
    uint32_t address;
    uint32_t size;
    uint32_t crc;
    uint32_t crcTest;
    uint32_t addressOffset = 0;
    address = size = crc = 0;

    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        std::size_t pos = filePath.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = filePath;
        }
        else
        {
            fileName = filePath.substr( pos + 1 );
        }

        //

        if ( (res = writeSystemVariable( deviceAddress, "SYSTEM.FILETRANSFER", fileName )) != IV_ERROR_SUCCESS )
        {
            // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
            // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
            if ( IV_RESULT_FUNCTION( res ) == 3 )
            {
                // Fehler liegt bei Wert
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else if ( IV_RESULT_FUNCTION( res ) == 2 )
            {
                // Variable nicht bekannt
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }

            return res;
        }

        uint32_t unusedUInt = 0;
        // Segmentbegin: Hintergrunduebertragung
        if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG, unusedUInt, unusedUInt, unusedUInt )) != IV_ERROR_SUCCESS )
        {
            if ( IV_RESULT_FUNCTION( res ) == 1 )
            {
                // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
                // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
                return -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                return res;
            }
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

        //////////////////////////////
        // Begin

        // Segmentbegin: Dateidaten lesen
        if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_FILETRANSFER, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_FILETRANSFER, address, size, crc );

        // Liegen Daten zur Datei vor?
        if ( size == 0 )
        {
            // Datei konnte nicht geoeffnet werden.
            //@TODO: Besseren Fehler fuer die Bedeutung. Zu lesende Datei existiert nicht, oder so aehnlich.
            errorApplication = -IV_ERROR_FILE_OPEN;
            goto label_end;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::active;
        controllerSegmentCurrent.transferType = ControllerSegment::btt_read_file;
        if ( controllerSegmentCurrent.allowedTransferTypes == ControllerSegment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            ControllerSegment &controllerSegmentGlobal = controllerSegment( 255 );
            if ( controllerSegmentGlobal.allowedTransferTypes != ControllerSegment::btt_none )
            {
                controllerSegmentCurrent.allowedTransferTypes = controllerSegmentGlobal.allowedTransferTypes;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transferFlags.settingsFromGlobal = true;
            }
        }
        controllerSegmentCurrent.beginTimePoint = std::chrono::system_clock::now();

        // Ermittelte Daten halten.
        controllerSegmentCurrent.filePath = filePath;
        controllerSegmentCurrent.address = address;
        controllerSegmentCurrent.crc = crc;
        // Dateigroesse halten.
        controllerSegmentCurrent.fileData.resize( size );
    }
    else
    {
        // Zwischengespeicherte Daten wiederherstellen.
        address = controllerSegmentCurrent.address;
        crc = controllerSegmentCurrent.crc;
        addressOffset = controllerSegmentCurrent.size;
    }

    if ( controllerSegmentCurrent.discard == false && errorApplication == IV_ERROR_SUCCESS )
    {
        //////////////////////////////
        // Daten

        // EEPROM uebertragen
        if ( (res = readFromFlash( deviceAddress, address + addressOffset, &controllerSegmentCurrent.fileData[addressOffset],
                                     static_cast<uint32_t>(controllerSegmentCurrent.fileData.size() - addressOffset) )) != IV_ERROR_SUCCESS )
        {
            if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
            {
                controllerSegmentCurrent.resume = true;
                return res;
            }

            // Bei Kommunikationsfehlern abbrechen,
            // bei Ausfuehrungsfehlern das Uebertragen abbrechen.
            if ( res < IV_ERROR_SUCCESS )
            {
                goto label_end;
            }
            else
            {
                errorApplication = res;
            }
        }
    }

    // Dateidaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_FILETRANSFER )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_FILETRANSFER );

    // Segmentend: Bedienung stoppen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        goto label_end;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    if ( errorApplication == IV_ERROR_SUCCESS )
    {
        crcTest = crc32( controllerSegmentCurrent.fileData.data(), static_cast<uint32_t>(controllerSegmentCurrent.fileData.size()) );
        if ( crc != crcTest )
        {
            trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
            res = -IV_ERROR_CRC_SEGMENT;
            goto label_end;
        }

        std::size_t pos = controllerSegmentCurrent.filePath.find_last_of( "\\/" );
        std::string fileName;
        if ( pos == std::string::npos )
        {
            fileName = controllerSegmentCurrent.filePath;
        }
        else
        {
            fileName = controllerSegmentCurrent.filePath.substr( pos + 1 );
        }
        // Bei Verwendung von '*' oder '?' im Dateinamen wird ggf. die Systemvariable mit dem richtigen Namen der Datei gefüllt
        // und muss im Nachgang ausgelesen und verwendet werden.
        pos = fileName.find_first_of( "*?" );
        if ( pos != std::string::npos )
        {
            std::string tmpFileName;
            if ( (res = readSystemVariable( deviceAddress, "SYSTEM.FILETRANSFER", tmpFileName )) != IV_ERROR_SUCCESS )
            {
                // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                if ( (res & 0xFF) == 3 )
                {
                    // Fehler liegt bei Wert
                    res = -IV_ERROR_FUNCTION_UNAVAILABLE;
                }
                else if ( (res & 0xFF) == 2 )
                {
                    // Variable nicht bekannt
                    res = -IV_ERROR_FUNCTION_UNAVAILABLE;
                }

                goto label_end;
            }

            if ( fileName == tmpFileName )
            {
                // Die Musterzeichen dürfen nun nicht mehr im Namen vorhanden sein. Dateien mit diesen Musterzeichen dürfen im System nicht existieren.
                // Dieses führt bei fopen zu einem Fehler.
            }
            else
            {
                pos = controllerSegmentCurrent.filePath.find_last_of( "\\/" );
                controllerSegmentCurrent.filePath.replace( pos + 1, fileName.size(), tmpFileName );
            }
        }

        // Datenanzahl in Datei uebernehmen.
        FILE *fp = fopen( controllerSegmentCurrent.filePath.c_str(), "wb" );
        // Datei oeffnen
        if ( fp == nullptr )
        {
            res = -IV_ERROR_FILE_OPEN;
            goto label_end;
        }

        if ( fwrite( controllerSegmentCurrent.fileData.data(), sizeof( uint8_t ), controllerSegmentCurrent.fileData.size(), fp ) != size )
        {
            fclose( fp );
            trace( "%s(%i):ERROR(1): %s [ Schreiben der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
            res = -IV_ERROR_FILE_WRITE;
            goto label_end;
        }

        fclose( fp );
    }

label_end:
    // Dateidaten loeschen.
    controllerSegmentCurrent.filePath.clear();
    controllerSegmentCurrent.fileData.clear();

    // Blocktransfer beendet.
    controllerSegmentCurrent.endTimePoint = std::chrono::system_clock::now();
    controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::inactive;

    // Ggf. aufgetretenen Fehler bei Verarbeitung zurueckgeben.
    if ( errorApplication != IV_ERROR_SUCCESS )
    {
        res = errorApplication;
    }

    return res;
}

int32_t Channel4::setTextDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setTextData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::setTextData( uint8_t deviceAddress, const std::string &filePath )
{
    // Zeichensatz- und Grafikdaten
    FileDeviceText *pFileDeviceText = m_fileDevice.GetFileDeviceText();
    int32_t res = IV_ERROR_SUCCESS;
    uint32_t address = 0;

    if ( pFileDeviceText->Import( filePath ) == false )
    {
        switch ( pFileDeviceText->errorState() )
        {
            case FileIntelHex::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileIntelHex::Error_state::file_format:
            default:
                return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind Textdaten vorhanden?
    if ( pFileDeviceText->sizeData() )
    {
        // Type User setzen.
        TEXTDATA *td = reinterpret_cast<TEXTDATA *>(pFileDeviceText->data());
        td->type = (m_loadAsDefaultLanguage) ? 2 : 3; // Textdata Default/User
        m_loadAsDefaultLanguage = 0; // ggf. zuruecksetzen

        address = 0;
        // Systemstopp setzen
        res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TEXT, address, pFileDeviceText->sizeData(), pFileDeviceText->checksum32bit() );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, pFileDeviceText->data(), pFileDeviceText->sizeData() )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TEXT, pFileDeviceText->checksum32bit() )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Textdaten erfolgreich uebertragen
    return res;
}

int32_t Channel4::setTouchKeyDataWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setTouchKeyData( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::setTouchKeyData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    listItemBoard tags;
    // Nur wenn Touch verbaut ist.
    tags = hardwareInfo.moduleTouch();
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Touchkeydaten
    FileTouchKeys fileTouchKeys;;
    // Touchdatendatei laden
    if ( !fileTouchKeys.load( filePath ) )
    {
        switch ( fileTouchKeys.errorState() )
        {
            case FileTouchKeys::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileTouchKeys::Error_state::file_format:
            default:
                return -IV_ERROR_FILE_FORMAT;
        }
    }

    StructHardwareInfo hwInfo = hardwareInfo.exportStructHardwareInfo();
    if ( fileTouchKeys.deviceType() != hwInfo.typeDevice )
    {
        return -IV_ERROR_FILE_DATA_FORMAT;
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    uint32_t address = 0;
    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind TouchKeyDaten vorhanden?
    if ( fileTouchKeys.sizeData() )
    {
        address = 0;
        auto data = fileTouchKeys.data();
        auto sizeData = fileTouchKeys.sizeData();
        auto crc = fileTouchKeys.checksum32bit();

        // Systemstopp setzen
        res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TOUCHKEYS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TOUCHKEYS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // TouchKeyDaten erfolgreich uebertragen
    return res;
}

int32_t Channel4::setDisplayColorsWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = setDisplayColors( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::setDisplayColors( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    HardwareInfo hardwareInfo;

    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    if ( hardwareInfo.deviceType() != HardwareInfo::TimeboyIV )
    {
        return -IV_ERROR_DEVICE_WRONG;
    }

    listItemBoard tags;
    // Nur wenn Farbdisplay verbaut ist.
    tags = hardwareInfo.moduleDisplay( { EnumModuleDisplay::Color_1_9, EnumModuleDisplay::Color_2_0 } );
    if ( tags.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }

    // Farbdaten
    FileDisplayColors fileDisplayColors;
    // Touchdatendatei laden
    if ( !fileDisplayColors.load( filePath ) )
    {
        switch ( fileDisplayColors.errorState() )
        {
            case FileDisplayColors::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileDisplayColors::Error_state::file_format:
            default:
                return -IV_ERROR_FILE_FORMAT;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    uint32_t address = 0;
    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Sind TouchKeyDaten vorhanden?
    if ( fileDisplayColors.sizeData() )
    {
        address = 0;
        auto data = fileDisplayColors.data();
        auto sizeData = fileDisplayColors.sizeData();
        auto crc = fileDisplayColors.checksum32bit();

        // Systemstopp setzen
        res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_DISPLAYCOLORS, address, sizeData, crc );
        if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
        {
            return res;
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, data, sizeData )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_DISPLAYCOLORS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
        else
        {
            // Funktion wird von Firmware nicht unterstuetzt.
            return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
    }
    else
    {
        // Es liegen keine Daten fuer die Uebertragung vor.
        return -IV_ERROR_NO_DATA_AVAILABLE;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // DisplayColorDaten erfolgreich uebertragen
    return res;
}

int32_t Channel4::readTablesWithTimeboySlot( uint8_t deviceAddress, const std::string &typeName )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = readTables( deviceAddress, typeName );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// Datensatzbeschreibungen aus Geraet lesen
int32_t Channel4::readTables( uint8_t deviceAddress, const std::string &typeName )
{
    uint8_t tableCount;
    uint8_t fieldCount;
    uint8_t i;
    int32_t err = IV_ERROR_SUCCESS;

    StructTable2 table;
    StructTable2Field field;
    StructRecordDescription *pTable = nullptr;
    StructRecordField *pField = nullptr;

    // Struktur uebernehmen
    if ( typeName == "LIST" )
    {
        MaxListB = 0;
    }
    else if ( typeName == "RECORD" )
    {
        MaxDatB = 0;
    }
    else
    {
        return -IV_ERROR_LIST_PROCESSING;
    }

    // Alle zuvor gelesenen Beschreibungen neu initalisieren.
    for ( tableCount = 0; tableCount < MAX_RECORD_DESCRIPTION; ++tableCount )
    {
        // Struktur uebernehmen
        if ( typeName == "LIST" )
        {
            std::fill_n( reinterpret_cast<uint8_t *>(&ListenBeschreibung[tableCount]), sizeof( StructRecordDescription ), 0 );
            for ( fieldCount = 0; fieldCount < MAX_FIELDS; ++fieldCount )
            {
                std::fill_n( reinterpret_cast<uint8_t *>(&ListenFeld[tableCount][fieldCount]), sizeof( StructRecordField ), 0 );
            }
        }
        else if ( typeName == "RECORD" )
        {
            std::fill_n( reinterpret_cast<uint8_t *>(&DatensatzBeschreibung[tableCount]), sizeof( StructRecordDescription ), 0 );
            for ( fieldCount = 0; fieldCount < MAX_FIELDS; ++fieldCount )
            {
                std::fill_n( reinterpret_cast<uint8_t *>(&DatensatzFeld[tableCount][fieldCount]), sizeof( StructRecordField ), 0 );
            }
        }
    }

    // Alle Tabellenbeschreibungen lesen
    for ( tableCount = 0; tableCount < MAX_RECORD_DESCRIPTION; ++tableCount )
    {
        std::string name( typeName );
        name += "." + std::to_string( tableCount );

        if ( (err = readTableStruct( deviceAddress, name, table )) != IV_ERROR_SUCCESS )
        {
            // Nur wenn der Minorname (Index) nicht bekannt ist, 
            // liegt wirklich ein Fehler vor.
            if ( err < 0 )
            {
                // Kommunikationsfehler
                return err;
            }
            else if ( err == 2 )
            {
                // Tabellenname unbekannt
                // Fehlercode zuruecksetzen
                err = IV_ERROR_SUCCESS;
            }
            else
            {
                // Funktionsfehler
                return err;
            }

            break;
        }

        // Struktur uebernehmen
        if ( typeName == "LIST" )
        {
            pTable = &ListenBeschreibung[tableCount];
        }
        else if ( typeName == "RECORD" )
        {
            pTable = &DatensatzBeschreibung[tableCount];
        }

        strncpy( (char *)pTable->name, table.name, sizeof( pTable->name ) - 1 );
        pTable->fields = table.count;
        pTable->ptr = 0;
    }

    // Feldbeschreibung lesen
    for ( i = 0; i < tableCount; ++i )
    {
        // Struktur uebernehmen
        if ( typeName == "LIST" )
        {
            pTable = &ListenBeschreibung[i];
        }
        else if ( typeName == "RECORD" )
        {
            pTable = &DatensatzBeschreibung[i];
        }

        // Alle Felder lesen
        for ( fieldCount = 0; fieldCount < pTable->fields; ++fieldCount )
        {
            std::string name( typeName );
            name += "." + std::to_string( i ) + "." + std::to_string( fieldCount );

            if ( (err = readTableColumnStruct( deviceAddress, name, field )) != IV_ERROR_SUCCESS )
            {
                if ( err < 0 )
                {
                    // Kommunikationsfehler
                    return err;
                }
                else
                {
                    // Funktionsfehler
                    return err;
                }
            }

            // Struktur uebernehmen
            if ( typeName == "LIST" )
            {
                pField = &ListenFeld[i][fieldCount];
            }
            else if ( typeName == "RECORD" )
            {
                pField = &DatensatzFeld[i][fieldCount];
            }

            pField->type = field.type;
            pField->count = field.size;
            strncpy( (char *)pField->name, field.name, sizeof( pField->name ) - 1 );
        }
    }

    // Information uebernehmen
    if ( typeName == "LIST" )
    {
        MaxListB = tableCount;
    }
    else if ( typeName == "RECORD" )
    {
        MaxDatB = tableCount;
    }

    return err;
}

int32_t Channel4::writeSetupWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        //char val[80];
        //// Systemvariablen fuer MD5-Pruefwerte lesen.
        //// Wenn die Systemvariable definiert ist, dann wird eine andere Art der Setupuebertragung verwendent.
        //if ((res = getSystemVariable(deviceAddress, "SYSTEM.TMP1", val)) != IV_ERROR_SUCCESS) {

        //    // Bei Kommunikationsfehler, abbruch.
        //    if (res < IV_ERROR_SUCCESS)
        //        return res;

        //    // Befehlsausfuehrung starten.
        res = writeSetupData( deviceAddress, filePath );
        //} else {

        //    // Befehlsausfuehrung starten.
        //    res = writeSetupData4_2( deviceAddress );
        //}

        if ( res == IV_ERROR_SUCCESS && !isPasswordProtection() )
            res = clearRecords( deviceAddress );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeSetupData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = IV_ERROR_SUCCESS;
    int32_t fileError;
    FileAES &fileAES = m_doc->fileAES();
    FileXML &fileXML = m_doc->fileXML();

    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        std::string ext( filePath.substr( filePath.size() - 4 ) );
        boost::to_lower( ext );

        // Depending on the file format, the document is initialized.
        FileType fileType = FileType::Undefined;
        if ( ext == ".aes" || ext == ".ael" )
        {
            fileType = FileType::AES_AEL;
        }
        else if ( ext == ".xml" )
        {
            fileType = FileType::XML;
        }
        else
        {
            // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael; *.xml) oeffnen.
            return -IV_ERROR_FILE_EXTENSION;
        }

        // Set the file type of currently loaded document.
        m_doc->setFileType( fileType );

        // Reset der StudioFile-Klasse
        m_doc->SetDocumentDefault();

        switch ( fileType )
        {
            case FileType::AES_AEL:
            {
                // Laden der Dateidaten. Keine Beachtung der Studioversion.
                if ( fileAES.load( filePath, &fileError, static_cast<FileAES::Enum_flag_ignore_version>(FileAES::fiv_min_studio | FileAES::fiv_should_studio) ) == false )
                {
                    switch ( fileError )
                    {
                        case -3:
                            return -IV_ERROR_FILE_READ;

                        case -7: // Mindestversion DLL wird nicht eingehalten.
                            return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;

                        case -8: // Sollteversion DLL
                            return -IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;

                        case -9: // Format der Setupdatei
                            return -IV_ERROR_FILE_FORMAT;

                        case -4: // CRC in Dateidaten
                        case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
                            return -IV_ERROR_FILE_CRC;
                    }
                    return -IV_ERROR_FILE_OPEN;
                }

                // Library-Version aus dem Flash-Header auslesen. Sollte die Version nicht gesetzt sein (z.B. Datei mit einer älteren
                // Version des Datafox Studios erstellt), dann wird die Studioversion ohne Buildnummer verwendet.
                uint32_t libraryVersion = fileAES.m_flash_header.libraryVersion;
                if ( libraryVersion == UINT32_MAX )
                {
                    libraryVersion = fileAES.m_flash_header.studioVersion & 0xFFFFFF00;
                }

                m_doc->setVersions( fileAES.m_flash_header.studioVersion, fileAES.m_flash_header.libraryVersion );
                break;
            }
            case FileType::XML:
            {
                // Analyze the contents of the XML file to build the document.
                auto xmlDoc = fileXML.xmlDoc();
                if ( !xmlDoc || (xmlDoc->LoadFile( filePath.c_str() ) != tinyxml2::XML_SUCCESS) )
                {
                    return -IV_ERROR_FILE_READ;
                }

                // Load succesful, continue with processing.
                auto rootNode = xmlDoc->FirstChildElement( fileXML.nodeNameByType( FileXML::ContentType::MainNode ).c_str() );      ///< main node of xml: 'documentStudio'
                fileXML.setParentNode( rootNode );
                if ( rootNode == nullptr )
                {
                    return -IV_ERROR_FILE_FORMAT;
                }

                // get header information
                if ( !(fileXML.readContent( FileXML::ContentType::FlashHeader, fileAES.m_flash_header ) &&
                        fileXML.readContent( FileXML::ContentType::FileHeader, fileAES.m_file_header )) )
                {
                    return -IV_ERROR_FILE_DATA_FORMAT;
                }

                // Library-Version aus dem Flash-Header auslesen. Sollte die Version nicht gesetzt sein (z.B. Datei mit einer älteren
                // Version des Datafox Studios erstellt), dann wird die Studioversion ohne Buildnummer verwendet.
                uint32_t libraryVersion = fileAES.m_flash_header.libraryVersion;
                if ( libraryVersion == 0 )
                {
                    libraryVersion = fileAES.m_flash_header.studioVersion & 0xFFFFFF00;
                }
                m_doc->setVersions( fileAES.m_flash_header.studioVersion, fileAES.m_flash_header.libraryVersion );

                // Dokument neu aufsetzen
                if ( !m_doc->BuildRootFromImage( fileAES.m_flash_header.setupType ) )
                {
                    return -IV_ERROR_FILE_DATA_FORMAT;
                }

                // Dateiname inkl. Pfad übernehmen.
                m_doc->setFilePath( filePath );

                break;
            }
        }

        int32_t res;
        std::string version;
        if ( (res = readFirmwareVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        SoftwareVersion versionFirmware( version );
        uint8_t type = versionFirmware.typeDevice();

        // Pruefen ob die Firmwareversion unterstuetzt wird.
        if ( !fileAES.IsSupportedFirmwareVersion( version.c_str() ) )
        {
            return -IV_ERROR_VERSION_FIRMWARE;
        }

        std::map<uint8_t, SetupType> setupType;
        // Prüfen ob das Setup für den vorliegenden Gerätetypen passt.
        setupType[IV_VALUE_TYPE_DEVICE_PZE] = SetupType::PZE_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_ZK] = SetupType::ZK_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_BDE] = SetupType::BDE_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_MOBIL] = SetupType::Mobil_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_MDE] = SetupType::MDE_BoxIV;
        setupType[IV_VALUE_TYPE_DEVICE_AE] = SetupType::AE_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_TIMEBOY] = SetupType::Timeboy_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_EXKLUSIV] = SetupType::Exklusiv_LineIV;
        setupType[IV_VALUE_TYPE_DEVICE_FLEX] = SetupType::Flex_MasterIV;
        setupType[IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE] = SetupType::Timeboy_Mobil_PZE;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_2_8_PURE] = SetupType::EVO_2_8_Pure;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_4_3] = SetupType::EVO_4_3;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IO_BOX] = SetupType::IO_Box;
        setupType[IV_VALUE_TYPE_DEVICE_KYO_INLOC] = SetupType::KYO_Inloc;
        setupType[IV_VALUE_TYPE_DEVICE_MOBIL_BOX] = SetupType::Mobil_Box;
        setupType[IV_VALUE_TYPE_DEVICE_TIMEBOY_TBDOCK] = SetupType::Timeboy_TBDock;
        setupType[IV_VALUE_TYPE_DEVICE_MOBIL_TBDOCK] = SetupType::Mobil_TBDock;
        setupType[IV_VALUE_TYPE_DEVICE_KYO_ONELOC] = SetupType::KYO_Oneloc;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_3_5_UNIVERSAL] = SetupType::EVO_3_5_Universal;
        //setupType[IV_VALUE_TYPE_DEVICE_SAMS_OCL_READER]    = SetupType::;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_3_5_PURE] = SetupType::EVO_2_8_Pure;
        //setupType[IV_VALUE_TYPE_DEVICE_EVO_AGERA]          = SetupType::;
        setupType[IV_VALUE_TYPE_DEVICE_KYO_CENLOC] = SetupType::KYO_Cenloc;
        setupType[IV_VALUE_TYPE_DEVICE_VARIO_10] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_VARIO_12] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_VARIO_15] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_VARIO_17] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EVO_12_1] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EVO_15] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EVO_18_5] = SetupType::IPC_Extension;
        setupType[IV_VALUE_TYPE_DEVICE_IPC_EVO_24] = SetupType::IPC_Extension;
        //setupType[IV_VALUE_TYPE_DEVICE_EVO_INTERA_2]       = SetupType::;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_4_6_FLEX_KEY] = SetupType::EVO_4_3;
        setupType[IV_VALUE_TYPE_DEVICE_EVO_5_0_PURE] = SetupType::EVO_2_8_Pure;

        // Entweder liegt kein Setuptyp für das Gerät vor oder der Setuptyp passt nicht zu dem vorliegenden Gerät.
        if ( !setupType.count( type ) || fileAES.m_flash_header.setupType != setupType[type] )
        {
            return -IV_ERROR_DEVICE_WRONG;
        }

        // Ist ein Dateiheader verfuegbar, ist die Firmwareversion zu pruefen.
        if ( fileAES.IsHeader() )
        {
            SoftwarebuildVersion fwVersion( version );
            // Achtung:
            // Bei der Auswertung ist die Variante "SHOULD" durch die Library nicht zu berücksichtigen, da sonst ein Einspielen verhindert würde.
            switch ( fileAES.ValidFirmwareVersion( fwVersion ) )
            {
                case 0: // Kein Dateiheader, hier unmoeglich.
                    return -IV_ERROR_FILE_FORMAT;

                case -1: // Unsupported 04.01.xx
                    return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE1_VERSION;

                case -2: // Mindestversion 04.01.xx
                    return -IV_ERROR_SETUPFILE_MIN_FIRMWARE1_VERSION;

                    //case -3: // Sollteversion 04.01.xx
                    //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE1_VERSION;

                case -4: // Unsupported 04.02.xx
                    return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE2_VERSION;

                case -5: // Mindestversion 04.02.xx
                    return -IV_ERROR_SETUPFILE_MIN_FIRMWARE2_VERSION;

                    //case -6: // Sollteversion 04.02.xx
                    //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE2_VERSION;

                case -7: // Firmware wird nicht unterstuetz.
                    return -IV_ERROR_VERSION_FIRMWARE;

                case -8: // Unsupported 04.03.xx
                    return -IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE3_VERSION;

                case -9: // Mindestversion 04.03.xx
                    return -IV_ERROR_SETUPFILE_MIN_FIRMWARE3_VERSION;

                    //case -10: // Sollteversion 04.03.xx
                    //return -IV_ERROR_SETUPFILE_SHOULD_FIRMWARE3_VERSION;
            }
        }
        else
        {
            // Wenn kein Dateiheader vorliegt, wird sichergestellt, dass eine Setupdatei mit der hinterlegten
            // DatafoxStudioIV Version 04.02.xx nur auf eine Hardware >= V3 aufgespielt werden kann.

            // Boardtyp ermitteln und Pruefen ob die Setupversion eingespielt werden darf.
            if ( (res = readBoardType( deviceAddress, type )) != IV_ERROR_SUCCESS )
            {
                if ( res == -IV_ERROR_FUNCTION_UNAVAILABLE )
                {
                    // Baugruppe als 1.4 annehmen
                    type = 0x14;
                }
                else
                {
                    return res;
                }
            }

            if ( type < IV_VALUE_TYPE_BOARD_3_0 )
            {
                // Neuere Setupdateien nicht in aeltere Hardwarestaende einspielen.
                //############################
                // 2010-03-04
                auto studioVersion = fileAES.endian_reverse( fileAES.m_flash_header.studioVersion );
                if ( (studioVersion & 0xFF000000) > 4 || studioVersion > 0x04010000 )
                {
                    return -IV_ERROR_VERSION_SETUPFILE;
                }
            }
        }

        if ( (res = checkData( deviceAddress )) == -1 )
        {
            return -IV_ERROR_UNREACHABLE;
        }
        else if ( res )
        {
            return -IV_ERROR_DATA_AVAILABLE;
        }
    }

    res = writeSetupData4( deviceAddress );

    if ( res == IV_ERROR_SUCCESS && !isPasswordProtection() )
    {
        //res = ModuleTable::clearRecords(deviceAddress);
    }

    return res;
}

int32_t Channel4::writeSetupData4( uint8_t deviceAddress )
{
    int32_t i;
    int32_t res = IV_ERROR_SUCCESS;
    FileAES &fileAES = m_doc->fileAES();

    // BlockTransfer durchfuehren.
    //controller_segment &controllerSegmentCurrent = get_controller_segment( deviceAddress );

    //if (controllerSegmentCurrent.blockTransferState == ControllerSegment::Inactive)
    //{    

    //@@: Uebergangsweise zur 04.01.01 release
    //uint8_t *buf = nullptr;
    uint8_t countTable;
    uint8_t countField;
    uint32_t sizeAccess;

    //////////////////////////////
    // Vorbereitung (Berechnung von Groessen und CRC)
    uint32_t sizeSetup = GetFlashSetupSize();

    // Vorhandene zur moeglichen Groesse pruefen
    // 	if (sizeSetup > (EEP_SIZE_SU + sizeof(StructConfig) + sizeof(dummyConfig) + FLASH_SIZE_SU)) {
    // 
    // 		return -IV_ERROR_FILE_FORMAT;
    // 	}

    uint32_t sizeEK = GetEntrance2EKFlashSize();
    // Speicher reservieren
    std::vector<uint8_t> buf;
    try
    {
        buf.resize( sizeEK, 0 );
        std::fill( buf.begin(), buf.end(), 0 );
    }
    catch ( ... )
    {
        return IV_ERROR_OVERFLOW;
    }

    //@@: Uebergangsweise zur 04.01.01 release
    // Zeiger auf Eingabeketten ohne FLASH_KONFIG_START
    // und die Zeiger der Eingabeketten mit einem Offset versehen.
    PrepareEntrance2EK( sizeSetup, &buf[0] );

    // Groesse der Flashdaten in EepMem eintragen
    fileAES.m_flash_header.flashSizeSUS = fileAES.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittseingabeketten eintragen
    fileAES.m_flash_header.inputChainAccess = fileAES.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittskontrolllisten eintragen
    fileAES.m_flash_header.listAccess = fileAES.endian_reverse( sizeSetup + sizeEK );

    // Groesse der MDE - Eingabeketten in EepMem eintragen.
    fileAES.m_flash_header.inputChainSizeMDE = GetMDEEKFlashSize();

    // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
    fileAES.m_flash_header.inputChainSizeEntrance = fileAES.endian_reverse( sizeEK );

    // Version der Kommunikationsbiliothek eintragen.
    fileAES.m_flash_header.libraryVersion = g_dfcomVersion;

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    //
    uint32_t address = 0;
    uint32_t size = 0;
    uint32_t crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, address, size, crc );

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = writeStream( deviceAddress, Stream_type::Update, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        setStreamTransfer( true );
    }

    ///////////////
    // Grunddaten
    // Groesse darf nicht veraendert werden!!!
    // Berechnung 0x80 - sizeof(StructConfig)
    std::array<uint8_t, 0x1C> dummyConfig;
    std::fill( dummyConfig.begin(), dummyConfig.end(), 0 );

    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = crc32( fileAES.m_eep_mem.data(), fileAES.m_eep_mem.size(), crc, false );
    crc = crc32( fileAES.m_config_mem.data(), fileAES.m_config_mem.size(), crc, false );
    crc = crc32( &dummyConfig[0], dummyConfig.size(), crc, false );
    crc = crc32( fileAES.m_flash_mem.data(), sizeSetup, crc, false );
    //@@: Uebergangsweise zur 04.01.01 release
    crc = crc32( &buf[0], sizeEK, crc, false );

    ///////////////
    // ZK II

    // Anzahl der Zutrittslisten ermitteln
    for ( i = 0, countTable = 0; countTable < (sizeof( m_entrance2Pointer ) / sizeof( uint32_t )); ++countTable )
    {
        if ( m_entrance2Pointer[countTable] == 0 )
        {
            break;
        }

        // Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
        i += m_entrance2Table[countTable].count;
    }
    sizeAccess = sizeof( m_entrance2Pointer );
    sizeAccess += countTable * sizeof( StructTable2 );
    sizeAccess += i * sizeof( StructTable2Field );

    // CRC der Zutrittslisten ermitteln
    crc = crc32( reinterpret_cast<uint8_t *>(m_entrance2Pointer), sizeof( m_entrance2Pointer ), crc, false );

    for ( i = 0; i < countTable; ++i )
    {
        crc = crc32( reinterpret_cast<uint8_t *>(&m_entrance2Table[i]), sizeof( StructTable2 ), crc, false );

        for ( countField = 0; countField < m_entrance2Table[i].count; ++countField )
        {
            // Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
            crc = crc32( reinterpret_cast<uint8_t *>(&m_entrance2TableField[i][countField]), sizeof( StructTable2Field ), crc,
                             ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))) );
        }
    }

    //////////////////////////////
    // Begin
    //	double now = (double)clock() / (double)CLOCKS_PER_SEC;
    //	int32_t nr = 0;
    //	Trace("Elapse (%i): %f\n", nr++, ((double)clock() / (double)CLOCKS_PER_SEC) - now);

    // Gesamtgroesze berechnen
    size = EEP_SIZE_SU + sizeof( StructConfig ) + static_cast<uint32_t>(dummyConfig.size()) + sizeSetup + sizeEK + sizeAccess;

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        // Segmentbeginn: Setup uebertragen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }

        trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, address, size, crc );
    }
    //}
    //else
    {
        // Zwischengepeicherte Daten fuer Wiederaufnahme setzen.

    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_eep_mem.data(), static_cast<uint32_t>(fileAES.m_eep_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG uebertragen (100 Byte)
    address += EEP_SIZE_SU;
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_config_mem.data(), static_cast<uint32_t>(fileAES.m_config_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG dummyConfig (28 Byte)
    address += sizeof( StructConfig );
    if ( (res = writeToFlash( deviceAddress, address, &dummyConfig[0], static_cast<uint32_t>(dummyConfig.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // FLASH uebertragen
    address += static_cast<uint32_t>(dummyConfig.size());
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_flash_mem.data(), sizeSetup )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    //@@: Uebergangsweise zur 04.01.01 release
    //////////////////////////////
    // Pointerarray + Strukturen der Zutrittslisten uebertragen

    // Eingabekette der Zutrittskontrolle
    address += sizeSetup;
    if ( (res = writeToFlash( deviceAddress, address, &buf[0], sizeEK )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    address += sizeEK;
    // Pointerarray uebertragen
    if ( (res = writeToFlash( deviceAddress, address, reinterpret_cast<uint8_t *>(m_entrance2Pointer), sizeof( m_entrance2Pointer ) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    address += sizeof( m_entrance2Pointer );

    // Strukturen uebertragen
    for ( i = 0; i < countTable; ++i )
    {
        // Tabellenstruktur uebertragen
        if ( (res = writeToFlash( deviceAddress, address, reinterpret_cast<uint8_t *>(&m_entrance2Table[i]), sizeof( StructTable2 ) )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += sizeof( StructTable2 );

        // Feldstrukturen uebertragen
        if ( (res = writeToFlash( deviceAddress, address, reinterpret_cast<uint8_t *>(m_entrance2TableField[i]), m_entrance2Table[i].count * sizeof( StructTable2Field ) )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += m_entrance2Table[i].count * sizeof( StructTable2Field );
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( isStreamTransfer() && m_streamBuffer.size() )
    {
        if ( (res = writeStream( deviceAddress, Stream_type::Setup, m_streamBuffer.data(), (uint32_t)m_streamBuffer.size() )) != IV_ERROR_SUCCESS )
        {
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }

        // Gepufferte Daten loeschen.
        clearStreamTransfer();
    }

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        //////////////////////////////
        // Ende
        // Segmentend: Setup uebertragen
        if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, crc );
    }

    // Extendedstruktur lesen
    StructExtendedFlash extBegin = *reinterpret_cast<StructExtendedFlash *>(fileAES.m_flashExtended);

    ///////////////
    // Extended - Speicher Vorbereitung

    // Zeigeroffsets anpassen (substrahieren von Offset)
    AdjustExtendedMemPointer();

    // Groesse MDE und dessen Eingabeketten
    size = extBegin.offsetFreeScope;

    // CRC berechnen
    crc = 0xFFFFFFFF;
    // IOModul - Konfiguration
    crc = crc32( fileAES.m_flashExtended, size, crc );

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        //////////////////////////////
        // Begin

        // Segmentbegin: Setup uebertragen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktionsfehler Typ IV_VALUE_SEGMENT_SETUP_EXTENDED unbekannt.
            }
            else
            {
                return res;
            }
        }

        trace( "%s(%i):INFO(1): %s [ (%i)beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, res, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc );
    }

    if ( res == IV_ERROR_SUCCESS )
    {
        //////////////////////////////
        // Extended - Speicher Uebertragen

        if ( (res = writeToFlash( deviceAddress, address, fileAES.m_flashExtended, size )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        address += size;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( isStreamTransfer() && m_streamBuffer.size() )
        {
            if ( (res = writeStream( deviceAddress, Stream_type::Setup_extended, m_streamBuffer.data(), (uint32_t)m_streamBuffer.size() )) != IV_ERROR_SUCCESS )
            {
                // Gepufferte Daten loeschen.
                clearStreamTransfer();
                return res;
            }

            // Gepufferte Daten loeschen.
            clearStreamTransfer();
        }

        // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
        if ( !isStreamTransfer() )
        {
            //////////////////////////////
            // Ende

            // Segmentend: Setup uebertragen
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc );
        }
    }

    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    // Pause wegen Verarbeitungsporblemen bis 4.1.5
    std::this_thread::sleep_for( std::chrono::seconds( 1 ) );

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, crc );


    // Pruefen und setzen des Status von Passwortschutz.
    writeSetupPassword( deviceAddress, reinterpret_cast<const char *>("ISPW") );
    ////////////////////////////////////////////////////////////////////
    // Der folgende Teil ist nur bei Versionen unter 04.01.07 noetig.
    // Ab dieser Version ist auch der Passwortschutz moeglich.
    // Es ist hier nicht moeglich, ein Setup mit einem gesetzten Passwort zu uebertragen 
    // und darauffolgend gleich weitere Funktionen aufrufen, weil ansonsten der
    // Schluessel fuer die Passwortverschluesselung offen gelegt werden muesste.
    if ( !isPasswordProtection() )
    {
        // GV setzen?
        if ( extBegin.offsetGlobalVariableValues )
        {
            StructGlobalVar *globVar = reinterpret_cast<StructGlobalVar *>(fileAES.m_flash_mem.data());
            char name[64];
            memset( name, 0, sizeof( name ) );

            // Struktur lesen
            StructValueGlobalVar ValueGVx = *reinterpret_cast<StructValueGlobalVar *>(&fileAES.m_flashExtended[extBegin.offsetGlobalVariableValues]);

            for ( i = 0; i < 8; ++i )
            {
                if ( ValueGVx.flags & (1 << i) )
                {
                    std::string name( "SETUP." );
                    name += globVar->names[i];
                    if ( (res = writeSetupVariable( deviceAddress, name.c_str(), (char *)&ValueGVx.value[i] )) != IV_ERROR_SUCCESS )
                    {
                        // Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
                        // und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
                        if ( (res & 0xFF) == 3 )
                        {
                            // Fehler liegt bei Wert
                            //res = -IV_ERROR_VARIABLE_VALUE;
                        }
                        else if ( (res & 0xFF) == 2 )
                        {
                            // Variable nicht bekannt
                            //res = -IV_ERROR_VARIABLE_UNDEFINED;
                        }

                        return res;
                    }
                }
            }
        }

        // Standardwerte fuer Systemvariable
        //############################
        // 2005-09-19: Erweiterte Struktur fuer Parameter
        auto studioVersion = fileAES.endian_reverse( fileAES.m_flash_header.studioVersion );
        if ( ((studioVersion & 0xFFFF0000) == 0x03010000 && (studioVersion & 0x0000FFFF) >= 0x00003406)
             || ((studioVersion & 0xFFFF0000) == 0x04010000 && (studioVersion & 0x0000FFFF) > 0x00000002) )
        {
            StructExtendedFlash extBegin = *(StructExtendedFlash *)fileAES.m_flashExtended;

            // Liegt die Struktur vor?
            if ( extBegin.offsetExtendedStruct )
            {
                StructExtended extParam = *(StructExtended *)&fileAES.m_flashExtended[extBegin.offsetExtendedStruct];

                std::map<std::string, std::string> systemVarMap;

                // Buchungswiederholsperrdauer
                systemVarMap["_SERVICE.DBLINPUT"] = boost::str( boost::format( "%1%" ) % std::min<uint16_t>( 65499, extParam.doubleInputTime ) );

                // Abblendedauer fuer Bestaetigungsfeld
                systemVarMap["_SERVICE.DEMANDTIMER"] = boost::str( boost::format( "%1%" ) % std::min<uint16_t>( 255, extParam.confirmationTime ) );

                // Tiefentladungsschutz
                systemVarMap["POWER.LOW"] = boost::str( boost::format( "%1%" ) % extParam.energieOption.powerLow );

                // Timerfaktor (Zeitbasis)
                systemVarMap["io.timer.divisor"] = boost::str( boost::format( "%1%" ) % std::min<uint8_t>( 10, extParam.timerTimeBaseDivisor ) );

                for ( const auto &item : systemVarMap )
                {
                    writeSystemVariable( deviceAddress, item.first, item.second );
                }
            }
        }
    }

    return res;
}

int32_t Channel4::writeSetupData4_2( uint8_t deviceAddress )
{
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t dummyConfig[0x1C]; // Groesse darf nicht veraendert werden!!!
    FileAES &fileAES = m_doc->fileAES();
    // Berechnung 0x80 - sizeof(StructConfig)

    //////////////////////////////
    // Vorbereitung (Berechnung von Groessen und CRC)
    uint32_t sizeSetup = GetFlashSetupSize();

    // Extendedstruktur
    StructExtendedFlash extBegin = *(StructExtendedFlash *)fileAES.m_flashExtended;

    // Groesse der Flashdaten in EepMem eintragen
    fileAES.m_flash_header.flashSizeSUS = fileAES.endian_reverse( sizeSetup );

    // Anfangsadresse der Zutrittseingabeketten eintragen
    fileAES.m_flash_header.inputChainAccess = fileAES.endian_reverse( static_cast<uint32_t>(0) );

    // Anfangsadresse der Zutrittskontrolllisten eintragen
    fileAES.m_flash_header.listAccess = fileAES.endian_reverse( static_cast<uint32_t>(0) );

    // Groesse der MDE - Eingabeketten in EepMem eintragen.
    fileAES.m_flash_header.inputChainSizeMDE = GetMDEEKFlashSize();

    // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
    fileAES.m_flash_header.inputChainSizeEntrance = fileAES.endian_reverse( static_cast<uint32_t>(0) );

    // Groesse der Headerstruktur in EepMem eintragen.
    fileAES.m_flash_header.headerSize = fileAES.m_file_header.size;

    // Groesse des Erweiterten Speicherbereichs eintragen.
    fileAES.m_flash_header.extendedSize = extBegin.offsetFreeScope;

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    //
    uint32_t address, size, crc;
    address = size = crc = 0;

    // Segmentbegin: Bedienung stoppen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, address, size, crc );

    // Es wird ab der Version 04.02.04 ein MD5 ueber den Dateiheader die Listen- und Datensatzbeschreibung als Systemvariable abgelegt.
    // Vor der Uebertragung des Setups werden diese hier geloescht, falls verfuegbar.
    bool md5Sysvars = false;

    // Systemvariablen fuer MD5-Pruefwerte loeschen.
    if ( (res = writeSystemVariable( deviceAddress, (char *)"SYSTEM.TMP1", (char *)"" )) != IV_ERROR_SUCCESS )
    {
        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if ( res != 1 && res != 2 )
        {
            return res;
        }
    }
    else if ( (res = writeSystemVariable( deviceAddress, (char *)"SYSTEM.TMP2", (char *)"" )) != IV_ERROR_SUCCESS )
    {
        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if ( res != 1 && res != 2 )
        {
            return res;
        }
    }
    else if ( (res = writeSystemVariable( deviceAddress, (char *)"SYSTEM.TMP3", (char *)"" )) != IV_ERROR_SUCCESS )
    {

        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if ( res != 1 && res != 2 )
        {
            return res;
        }
    }
    else if ( (res = writeSystemVariable( deviceAddress, (char *)"SYSTEM.TMP4", (char *)"" )) != IV_ERROR_SUCCESS )
    {
        // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
        if ( res != 1 && res != 2 )
        {
            return res;
        }
    }
    else
    {

        // Die Systemvariablen wurde beide erfolgreich geleert.
        md5Sysvars = true;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = writeStream( deviceAddress, Stream_type::Update, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        setStreamTransfer( true );
    }

    ///////////////
    // Grunddaten
    memset( dummyConfig, 0, sizeof( dummyConfig ) );

    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = crc32( fileAES.m_eep_mem.data(), fileAES.m_eep_mem.size(), crc, false );
    crc = crc32( fileAES.m_config_mem.data(), fileAES.m_config_mem.size(), crc, false );
    crc = crc32( dummyConfig, sizeof( dummyConfig ), crc, false );
    crc = crc32( fileAES.m_flash_mem.data(), sizeSetup, crc, false );
    crc = crc32( fileAES.m_file_header_mem.data(), fileAES.m_file_header.size, crc );

    //////////////////////////////
    // Begin
    //	double now = (double)clock() / (double)CLOCKS_PER_SEC;
    //	int32_t nr = 0;
    //	Trace("Elapse (%i): %f\n", nr++, ((double)clock() / (double)CLOCKS_PER_SEC) - now);

    // Gesamtgroesze berechnen
    // In dem Block des Setups werden verschiedene Einzelbloecke zusammengefasst uebertragen.
    size = EEP_SIZE_SU + sizeof( StructConfig ) + sizeof( dummyConfig ) + sizeSetup + fileAES.m_file_header.size;

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        // Segmentbeginn: Setup uebertragen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, address, size, crc );
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_eep_mem.data(), static_cast<uint32_t>(fileAES.m_eep_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG uebertragen (100 Byte)
    address += EEP_SIZE_SU;
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_config_mem.data(), static_cast<uint32_t>(fileAES.m_config_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // CONFIG dummyConfig (28 Byte)
    address += sizeof( StructConfig );
    if ( (res = writeToFlash( deviceAddress, address, dummyConfig, static_cast<uint32_t>(sizeof( dummyConfig )) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // FLASH uebertragen
    address += static_cast<uint32_t>(sizeof( dummyConfig ));
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_flash_mem.data(), sizeSetup )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Header der Datei uebertragen
    address += sizeSetup;
    if ( (res = writeToFlash( deviceAddress, address, fileAES.m_file_header_mem.data(), static_cast<uint32_t>(fileAES.m_file_header.size) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( isStreamTransfer() && m_streamBuffer.size() )
    {
        if ( (res = writeStream( deviceAddress, Stream_type::Setup, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
        {
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
            return res;
        }

        // Gepufferte Daten loeschen.
        clearStreamTransfer();
    }

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        //////////////////////////////
        // Ende
        // Segmentend: Setup uebertragen
        if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP, crc );
    }

    ///////////////
    // Extended - Speicher Vorbereitung

    // Zeigeroffsets anpassen (substrahieren von Offset)
    AdjustExtendedMemPointer();

    // Groesse MDE und dessen Eingabeketten
    size = extBegin.offsetFreeScope;

    // CRC berechnen
    crc = 0xFFFFFFFF;
    // IOModul - Konfiguration
    crc = crc32( fileAES.m_flashExtended, size, crc );

    // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
    if ( !isStreamTransfer() )
    {
        //////////////////////////////
        // Begin

        // Segmentbegin: Setup uebertragen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Zu grosse Setupdatei, es ist nicht genug speicher verfuegbar.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            return res;
        }
        trace( "%s(%i):INFO(1): %s [ (%i)beginSetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, res, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc );
    }

    if ( res == IV_ERROR_SUCCESS )
    {
        //////////////////////////////
        // Extended - Speicher Uebertragen

        if ( (res = writeToFlash( deviceAddress, address, fileAES.m_flashExtended, size )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        address += size;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( isStreamTransfer() && m_streamBuffer.size() )
        {
            if ( (res = writeStream( deviceAddress, Stream_type::Setup_extended, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
            {
                // Gepufferte Daten loeschen.
                clearStreamTransfer();
                return res;
            }

            // Gepufferte Daten loeschen.
            clearStreamTransfer();
        }

        // Wenn kein Streaming moeglich, dann mit inneren Bloecken uebertragen.
        if ( !isStreamTransfer() )
        {
            //////////////////////////////
            // Ende

            // Segmentend: Setup uebertragen
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
            trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc );
        }
    }

    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    // Pause wegen Verarbeitungsporblemen bis 4.1.5
    std::this_thread::sleep_for( std::chrono::seconds( 1 ) );

    // Werden die Systemvariablen fuer die MD5 unterstuetzt?
    if ( md5Sysvars )
    {
        char *md5Tables = fileAES.GetMD5Header(); // Dateiheader
        // Systemvariablen fuer MD5 loeschen.
        if ( md5Tables && (res = writeSystemVariable( deviceAddress, "SYSTEM.TMP1", md5Tables )) != IV_ERROR_SUCCESS )
        {
            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if ( res != 1 && res != 2 )
            {
                return res;
            }
        }

        md5Tables = fileAES.GetMD5Tables( true ); // Listenbeschreibungen
        if ( md5Tables && (res = writeSystemVariable( deviceAddress, "SYSTEM.TMP2", md5Tables )) != IV_ERROR_SUCCESS )
        {
            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if ( res != 1 && res != 2 )
            {
                return res;
            }
        }

        md5Tables = fileAES.GetMD5Tables( false ); // Datensatzbeschreibungen
        if ( md5Tables && (res = writeSystemVariable( deviceAddress, "SYSTEM.TMP3", md5Tables )) != IV_ERROR_SUCCESS )
        {
            // Wenn der Fehler nicht Major oder Minorbestandteil unbekannt ist, dann hier bereits abbrechen.
            if ( res != 1 && res != 2 )
            {
                return res;
            }
        }
    }

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_UPDATE, crc );


    // Pruefen und setzen des Status von Passwortschutz.
    writeSetupPassword( deviceAddress, "ISPW" );

    return res;
}

int32_t Channel4::readSetupWithTimeboySlot( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res;
    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        //char val[80];
        //// Systemvariablen fuer MD5-Pruefwerte ermitteln.
        //// Wenn die Systemvariable definiert ist und einen gueltigen MD5 beinhaltet, 
        //// dann wird eine andere Art der Setupuebertragung verwendent.
        //if ((res = getSystemVariable(deviceAddress, "SYSTEM.TMP1", val)) == IV_ERROR_SUCCESS && IsMD5(val)) {

        //    // Befehlsausfuehrung starten.
        //    res = readSetupData4_2(deviceAddress);
        //} else {

        //    // Bei einem Kommunikationsfehler, abbrechen.
        //    if (res < IV_ERROR_SUCCESS)
        //        return res;

        //    // Befehlsausfuehrung starten.
        res = readSetupData( deviceAddress, filePath );
        //}


        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readSetupData( uint8_t deviceAddress, const std::string &filePath )
{
    int32_t res = 0;
    FileAES &fileAES = m_doc->fileAES();

    std::string ext = boost::to_lower_copy( filePath.substr( filePath.size() - 4 ) );
    if ( ext != ".aes" && ext != ".ael" && ext != ".xml" )
    {
        // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
        return -IV_ERROR_FILE_EXTENSION;
    }

    // Speicher initialisieren
    m_doc->SetDocumentDefault();

    std::string version;
    if ( (res = readFirmwareVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Pruefen ob die Firmwareversion unterstuetzt wird.
    if ( !fileAES.IsSupportedFirmwareVersion( version.c_str() ) )
    {
        return -IV_ERROR_VERSION_FIRMWARE;
    }

    // Setupdatei nach erfolgreichem Auslesen, abspeichern.
    if ( (res = readSetupData4( deviceAddress )) == IV_ERROR_SUCCESS )
    {
        // Diese Version der Firmware der Setupdatei als Mindestversion vorgeben.
        SoftwarebuildVersion fwVersion( version );
        fileAES.NeedFirmwareVersion( fwVersion );

        if ( ext == ".xml" )
        {
            if ( !m_doc->BuildRootFromImage( fileAES.m_flash_header.setupType ) )
            {
                res = -IV_ERROR_FILE_OPEN;
            }
            else
            {
                if ( m_doc->fileXML().save( filePath ) != FileXML::ErrorCodes::Success )
                {
                    res = -IV_ERROR_FILE_OPEN;
                }
            }
        }
        else
        {
            int32_t error;
            if ( !fileAES.save( filePath, &error ) )
            {
                res = -IV_ERROR_FILE_OPEN;
            }
        }
    }

    return res;
}

int32_t Channel4::readSetupData4( uint8_t deviceAddress )
{
    int32_t i;
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t dummyConfig[0x1C]; // Groesse nicht veraendern !!!
    // Berechnet sich aus 0x80 - sizeof(StructConfig)

    uint8_t countTable;
    uint8_t countField;
    uint32_t sizeAccess;

    //
    uint32_t address = 0;
    uint32_t size = 0;
    uint32_t crc = 0;
    uint32_t crcTest = 0;

    FileAES &fileAES = m_doc->fileAES();

    // Segmentbegin: Bedienung stoppen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        if ( (res & 0xFF) == 1 )
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc );

    // Vorhandene zur moeglichen Groesse pruefen
    if ( size < (EEP_SIZE_SU + (sizeof( StructConfig ) + sizeof( dummyConfig ))) )
    {
        return -IV_ERROR_PROCESSING;
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_eep_mem.data(), static_cast<uint32_t>(fileAES.m_eep_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += EEP_SIZE_SU;
    // CONFIG uebertragen (100 Byte)
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_config_mem.data(), static_cast<uint32_t>(fileAES.m_config_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += static_cast<uint32_t>(sizeof( StructConfig ));
    // CONFIG dummybytes (28 Byte)
    if ( (res = readFromFlash( deviceAddress, address, dummyConfig, static_cast<uint32_t>(sizeof( dummyConfig )) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += static_cast<uint32_t>(sizeof( dummyConfig ));

    // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
    uint32_t sizeSetup = fileAES.endian_reverse( fileAES.m_flash_header.flashSizeSUS );

    // FLASH uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_flash_mem.data(), sizeSetup )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += sizeSetup;


    // Groesse der Zutrittskontroll - Eingabeketten aus EepMem entnehmen.
    uint32_t sizeEK = fileAES.endian_reverse( fileAES.m_flash_header.inputChainSizeEntrance );

    // Speicher reservieren
    std::unique_ptr<uint8_t[]> bufEK( new uint8_t[sizeEK] );
    std::fill( bufEK.get(), bufEK.get() + sizeEK, 0 );

    //////////////////////////////
    // Pointerarray + Strukturen der Zutrittslisten uebertragen

    // Eingabekette der Zutrittskontrolle
    if ( (res = readFromFlash( deviceAddress, address, bufEK.get(), sizeEK )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += sizeEK;

    // Pointerarray lesen
    if ( (res = readFromFlash( deviceAddress, address, (uint8_t *)m_entrance2Pointer, static_cast<uint32_t>(sizeof( m_entrance2Pointer )) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += sizeof( m_entrance2Pointer );

    ///////////////
    // ZK II

    // Anzahl der Zutrittslisten ermitteln
    for ( i = 0, countTable = 0; countTable < (sizeof( m_entrance2Pointer ) / sizeof( uint32_t )); countTable++ )
    {
        if ( m_entrance2Pointer[countTable] == 0 )
        {
            break;
        }

        // Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
        i += m_entrance2Table[countTable].count;
    }
    sizeAccess = sizeof( m_entrance2Pointer );
    sizeAccess += countTable * sizeof( StructTable2 );
    sizeAccess += i * sizeof( StructTable2Field );

    // Strukturen lesen
    for ( i = 0; i < countTable; i++ )
    {
        // Tabellenstruktur uebertragen
        if ( (res = readFromFlash( deviceAddress, address, (uint8_t *)&m_entrance2Table[i], sizeof( StructTable2 ) )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        address += sizeof( StructTable2 );

        // Feldstrukturen lesen
        if ( (res = readFromFlash( deviceAddress, address, (uint8_t *)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof( StructTable2Field ) )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        address += m_entrance2Table[i].count * sizeof( StructTable2Field );
    }


    ///////////////
    // Grunddaten

    // CRC berechnen
    crcTest = 0xFFFFFFFF;
    crcTest = crc32( fileAES.m_eep_mem.data(), fileAES.m_eep_mem.size(), crcTest, false );
    crcTest = crc32( fileAES.m_config_mem.data(), fileAES.m_config_mem.size(), crcTest, false );
    crcTest = crc32( dummyConfig, sizeof( dummyConfig ), crcTest, false );
    crcTest = crc32( fileAES.m_flash_mem.data(), sizeSetup, crcTest, false );
    crcTest = crc32( bufEK.get(), sizeEK, crcTest, false );

    // Speicher wieder freigeben.
    bufEK.reset( nullptr );

    // CRC der Zutrittslisten ermitteln
    crcTest = crc32( (uint8_t *)m_entrance2Pointer, sizeof( m_entrance2Pointer ), crcTest, false );

    for ( i = 0; i < countTable; i++ )
    {
        crcTest = crc32( (uint8_t *)&m_entrance2Table[i], sizeof( StructTable2 ), crcTest, false );

        for ( countField = 0; countField < m_entrance2Table[i].count; countField++ )
        {
            // Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
            crcTest = crc32( (uint8_t *)&m_entrance2TableField[i][countField], sizeof( StructTable2Field ), crcTest, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))) );
        }
    }

    //////////////////////////////
    // Ende

    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP );

    //////////////////////////////
    // Begin
    address = size = crc = 0;
    // Segmentbegin: Setup uebertragen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc );

    //////////////////////////////
    // Extended - Speicher Lesen

    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_flashExtended, size )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += size;

    ///////////////
    // Extended - Speicher pruefen

    // Extended - Speicher
    crcTest = 0xFFFFFFFF;
    crcTest = crc32( fileAES.m_flashExtended, size, crcTest );

    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED );


    // Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    return res;
}

int32_t Channel4::readSetupData4_2( uint8_t deviceAddress )
{
    int32_t res = IV_ERROR_SUCCESS;
    uint8_t dummyConfig[0x1C]; // Groesse nicht veraendern !!!
    // Berechnet sich aus 0x80 - sizeof(StructConfig)

    uint32_t sizeSetup = 0, sizeHeader = 0, sizeExtended = 0;

    //
    uint32_t address, size, crc, crcTest = 0;
    address = size = crc = 0;

    FileAES &fileAES = m_doc->fileAES();

    // Segmentbegin: Bedienung stoppen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        if ( (res & 0xFF) == 1 )
        {
            // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
            // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
            res = -IV_ERROR_FUNCTION_UNAVAILABLE;
        }
        else
        {
            return res;
        }
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

    //////////////////////////////
    // Begin

    // Segmentbegin: Setup uebertragen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc );

    // Gelieferte zur Mindestgroesse pruefen
    if ( size < (EEP_SIZE_SU + sizeof( StructConfig ) + sizeof( dummyConfig )) )
    {
        return -IV_ERROR_PROCESSING;
    }

    //////////////////////////////
    // Grunddaten

    // EEPROM uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_eep_mem.data(), static_cast<uint32_t>(fileAES.m_eep_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += EEP_SIZE_SU;
    // CONFIG uebertragen (100 Byte)
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_config_mem.data(), static_cast<uint32_t>(fileAES.m_config_mem.size()) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += static_cast<uint32_t>(sizeof( StructConfig ));
    // CONFIG dummybytes (28 Byte)
    if ( (res = readFromFlash( deviceAddress, address, dummyConfig, static_cast<uint32_t>(sizeof( dummyConfig )) )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += static_cast<uint32_t>(sizeof( dummyConfig ));

    // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
    sizeSetup = fileAES.endian_reverse( fileAES.m_flash_header.flashSizeSUS );

    // FLASH uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_flash_mem.data(), sizeSetup )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += sizeSetup;

    // Groesse der Headerdaten aus EepMem entnehmen.
    sizeHeader = fileAES.m_flash_header.headerSize;
    // Headerdaten uebertragen
    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_file_header_mem.data(), sizeHeader )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Pruefen ob die Headerangaben passen.
    if ( !fileAES.IsHeader() )
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;

    SoftwarebuildVersion minVersion( fileAES.m_file_header.minLibraryVersion );
    // Pruefen ob die Mindestversion der benoetigten DLL erreicht bleibt.
    if ( g_dfcomVersion < minVersion )
    {

        // Mindestversion DLL wird nicht eingehalten.
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
    }

    ///////////////
    // Grunddaten
    if ( size != (EEP_SIZE_SU + sizeof( StructConfig ) + sizeof( dummyConfig ) + sizeSetup + sizeHeader) )
    {

        trace( "%s(%i):ERROR(1): %s [ Size von Segment falsch: %u, %u ]\n", __FILE__, __LINE__, __func__, size,
               (EEP_SIZE_SU + sizeof( StructConfig ) + sizeof( dummyConfig ) + sizeSetup + sizeHeader) );
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;;
    }

    // CRC berechnen
    crcTest = 0xFFFFFFFF;
    crcTest = crc32( fileAES.m_eep_mem.data(), fileAES.m_eep_mem.size(), crcTest, false );
    crcTest = crc32( fileAES.m_config_mem.data(), fileAES.m_config_mem.size(), crcTest, false );
    crcTest = crc32( dummyConfig, sizeof( dummyConfig ), crcTest, false );
    crcTest = crc32( fileAES.m_flash_mem.data(), sizeSetup, crcTest, false );
    crcTest = crc32( fileAES.m_file_header_mem.data(), sizeHeader, crcTest );

    //////////////////////////////
    // Ende

    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP );

    //////////////////////////////
    // Begin
    address = size = crc = 0;
    // Segmentbegin: Setup uebertragen
    if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc );

    //////////////////////////////
    // Extended - Speicher Lesen

    // Groesse des Extended Speichers aus eeprom ermitteln.
    sizeExtended = fileAES.m_flash_header.extendedSize;

    if ( (res = readFromFlash( deviceAddress, address, fileAES.m_flashExtended, sizeExtended )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    address += sizeExtended;

    ///////////////
    // Extended - Speicher pruefen
    if ( size != sizeExtended )
    {

        trace( "%s(%i):ERROR(1): %s [ Size von Segment falsch: %u, %u ]\n", __FILE__, __LINE__, __func__, size, sizeExtended );
        return -IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
    }

    // Extended - Speicher
    crcTest = 0xFFFFFFFF;
    crcTest = crc32( fileAES.m_flashExtended, size, crcTest );

    if ( crc != crcTest )
    {
        trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
        return -IV_ERROR_CRC_SEGMENT;
    }

    //////////////////////////////
    // Ende

    // Segmentend: Setup uebertragen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED );


    // Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
    // Zeigeroffsets anpassen (addieren von Offset)
    AdjustExtendedMemPointer();

    crc = 0;
    // Segmentend: Bedienung stoppen
    if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
    {
        return res;
    }
    trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

    return res;
}

int32_t Channel4::writeListWithTimeboySlot( uint8_t deviceAddress )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeListData( deviceAddress );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeListData( uint8_t deviceAddress )
{
    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        int32_t i;
        // Listengesamtgroesse berechnen und ggf. sortieren.
        uint32_t listSize = MAX_RECORD_DESCRIPTION * sizeof( uint32_t );
        uint32_t listPointer[MAX_RECORD_DESCRIPTION];
        for ( i = 0; i < MAX_RECORD_DESCRIPTION; ++i )
        {
            if ( m_listSize[i] )
            {
                listPointer[i] = listSize + FLASH_LIST_START;
                ChangeOrderLong( (uint8_t *)&listPointer[i] );

                listSize += m_listSize[i];
            }
            else
            {
                listPointer[i] = 0;
            }
        }

        if ( listSize == (MAX_RECORD_DESCRIPTION * sizeof( uint32_t )) )
        {
            return -IV_ERROR_LIST_PROCESSING;
            return 0; // keine Listen vorhanden
        }
    }

    int32_t res = writeListData4( deviceAddress );

    if ( (res > 0) && (((res >> 8) & 0xFF) == 3) )
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.

    return res;
}

// Fuer neue Listentechnik
int32_t Channel4::writeListData4( uint8_t deviceAddress )
{
    uint8_t tableNum;
    int32_t res = IV_ERROR_SUCCESS;

    StructTable2 table;
    char name[32];

    //
    uint16_t segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    uint32_t address, size, crc;
    uint32_t addressOffset = 0;
    address = size = crc = 0;

    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        // Ggf. gesetztes Streaming deaktivieren.
        setStreamTransfer( false );
        clearStreamTransfer();

        // Systemstopp Listen uebertragen
        while ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
            if ( ((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG) )
            {
                segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
                continue;
            }

            return res;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::active;
        controllerSegmentCurrent.transferType = ControllerSegment::btt_list;
        if ( controllerSegmentCurrent.allowedTransferTypes == ControllerSegment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            ControllerSegment &controllerSegmentGlobal = controllerSegment( 255 );
            if ( controllerSegmentGlobal.allowedTransferTypes != ControllerSegment::btt_none )
            {
                controllerSegmentCurrent.allowedTransferTypes = controllerSegmentGlobal.allowedTransferTypes;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transferFlags.settingsFromGlobal = true;
            }
        }
        controllerSegmentCurrent.beginTimePoint = std::chrono::system_clock::now();
        // UpdateVariation halten.
        controllerSegmentCurrent.updateVariation = segmentUpdateVariation;

        // Pruefen ob die Uebertragung per Stream moeglich ist.
        if ( (res = writeStream( deviceAddress, Stream_type::Update_background, nullptr, 0 )) == IV_ERROR_SUCCESS )
        {
            setStreamTransfer( true );
        }
    }

    // Listendaten Tabellenweise uebertragen
    for ( tableNum = controllerSegmentCurrent.number; tableNum < MAX_RECORD_DESCRIPTION; ++tableNum, ++controllerSegmentCurrent.number )
    {
        // Nicht importierte Listen uebergehen
        if ( m_listSize[tableNum] == 0 )
            continue;

        // Bei wiederaufnahme eines Blocktransfers ist dieses bereits durchgeführt worden.
        if ( controllerSegmentCurrent.resume == false )
        {
            // Bei Streaming werden die inneren Bloecke wegelassen.
            if ( isStreamTransfer() == false )
            {
                // Maximale Segmentanzahl nicht ueberschreiten
                if ( tableNum && !(tableNum % 16) )
                {
                    crc = 0;
                    // Systemstopp Listen uebertragen
                    if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
                    {
                        return res;
                    }

                    // Systemstopp Listen uebertragen
                    if ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
                    {
                        return res;
                    }
                }
            }

            memset( name, 0, sizeof( name ) );
            snprintf( name, sizeof( name ), "LIST.%i", tableNum );
            // Zur Liste Informationen ermitteln
            if ( (res = readTableStruct( deviceAddress, name, table )) != IV_ERROR_SUCCESS )
            {
                //TODO: Liste unbekannt Funktionsfehler 2 auswerten.
                return res;
            }
            strncpy( name, table.name, sizeof( name ) - 1 );
            std::string tmpName = name;
            int32_t count = snprintf( table.name, sizeof( table.name ), "LIST.%s", tmpName.c_str() );

            if ( count > static_cast<int32_t>(sizeof( table.name )) )
            {
                assert( false );
                return IV_ERROR_LIST_PROCESSING;
            }

            // Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
            if ( m_listSortCheck[tableNum] == false )
            {
                // Listen ohne Datensaetze nicht sortieren.
                // ACHTUNG: Ansonsten Division durch 0 bei Pruefung der Datensatzgroesse !!!
                if ( *reinterpret_cast<uint32_t *>(&m_pucListMem[tableNum][0]) )
                {
                    int32_t fieldOffs = 0;

                    if ( table.flags & TF_SORT && table.sort < table.count )
                    {
                        char name[64];
                        int32_t fieldNum;
                        StructTable2Field field;
                        memset( &field, 0, sizeof( StructTable2Field ) );
                        m_ucListRecordSize[tableNum] = 0;
                        // Alle Felder fuer Offset lesen
                        for ( fieldNum = 0; fieldNum < table.count; fieldNum++ )
                        {
                            if ( fieldNum < (table.sort + 1) )
                            {
                                // Datensatzoffset fuer Sortierfeld aufsummieren.
                                fieldOffs += field.size;
                            }

                            snprintf( name, sizeof( name ), "LIST.%i.%i", tableNum, fieldNum );

                            if ( (res = readTableColumnStruct( deviceAddress, name, field )) != IV_ERROR_SUCCESS )
                            {
                                if ( res < 0 )
                                {
                                    // Kommunikationsfehler
                                    return res;
                                }
                                else
                                {
                                    // Funktionsfehler
                                    return res;
                                }
                            }

                            // Datensatzgroesse fuer Swap-Routine berechnen
                            m_ucListRecordSize[tableNum] += field.size;
                        }

                        // Die Listensatzgröße muss mit der importierten übereinstimmen.
                        if ( ((m_listSize[tableNum] - sizeof( uint32_t )) / *(uint32_t *)&m_pucListMem[tableNum][0]) != m_ucListRecordSize[tableNum] )
                        {
                            trace( "%s(%i):ERROR(1): %s [ Listensatzgroesse (%i) stimmt nicht mit der im Setup definierten ueberein. ]\n", __FILE__, __LINE__, __func__, tableNum );
                            return -IV_ERROR_LIST_PROCESSING;
                        }

                        // Liste sortieren.
                        SortList( m_pucListMem[tableNum], m_ucListRecordSize[tableNum], fieldOffs );

                        // Bei allen Listen ist 0 die sortierte Spalte
                        // Flag dafuer dass die Liste sortiert wurde setzen.
                        table.flags |= TF_SORTED;
                    }
                }

                // Sortiert oder zumindest geprueft.
                m_listSortCheck[tableNum] = true;
            }

            // CRC berechnen
            crc = 0xFFFFFFFF;
            crc = crc32( (uint8_t *)&table, sizeof( StructTable2 ), crc, false );
            crc = crc32( m_pucListMem[tableNum], m_listSize[tableNum], crc );

            // Groesze berechnen
            size = sizeof( StructTable2 ) + m_listSize[tableNum];

            // Bei Streaming werden die inneren Bloecke wegelassen.
            if ( !isStreamTransfer() )
            {
                // Systemstopp Liste uebertragen
                if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_LIST, address, size, crc )) != IV_ERROR_SUCCESS )
                {
                    return res;
                }

                // Zusatzdaten
                controllerSegmentCurrent.address = address;
                controllerSegmentCurrent.crc = crc;
            }

            // Daten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, (uint8_t *)&table, sizeof( StructTable2 ) )) != IV_ERROR_SUCCESS )
            {
                // Speicherplatz reicht nicht aus.
                if ( res == 3 )
                {
                    res = -IV_ERROR_NOT_ENOUGH_FLASH;
                }

                return res;
            }

            // Daten uebertragen
            address += sizeof( StructTable2 );

            // Startadresse halten
            controllerSegmentCurrent.address = address;
            // Bereits uebertragene Datenmenge der Tabellenstruktur auf 0 setzen.
            controllerSegmentCurrent.size = 0;
            addressOffset = 0;
        }
        else
        {
            // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
            segmentUpdateVariation = controllerSegmentCurrent.updateVariation;
            address = controllerSegmentCurrent.address;
            crc = controllerSegmentCurrent.crc;
            addressOffset = controllerSegmentCurrent.size;
        }

        // Abbruch der Blockuebertragung.
        if ( controllerSegmentCurrent.discard == false )
        {
            if ( (res = writeToFlash( deviceAddress, address + addressOffset, m_pucListMem[tableNum] + addressOffset, m_listSize[tableNum] - addressOffset )) != IV_ERROR_SUCCESS )
            {
                // Speicherplatz reicht nicht aus.
                if ( res == 3 )
                {
                    res = -IV_ERROR_NOT_ENOUGH_FLASH;
                }

                if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
                {
                    controllerSegmentCurrent.resume = true;
                }

                return res;
            }
        }

        // Listendaten komplett uebertragen.
        controllerSegmentCurrent.resume = false;

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( isStreamTransfer() && m_streamBuffer.size() )
        {
            if ( (res = writeStream( deviceAddress, Stream_type::Setup_list, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
            {
// Gepufferte Daten loeschen.
                clearStreamTransfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
            // Systemstopp Liste uebertragen
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_LIST, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }

        // Abbruch der Blockuebertragung.
        if ( controllerSegmentCurrent.discard )
            break;
    }

    crc = 0;
    // Systemstopp Liste uebertragen
    if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // BlockTransfer Beendet.
    controllerSegmentCurrent.endTimePoint = std::chrono::system_clock::now();
    controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::inactive;

    return res;
}

int32_t Channel4::writeEntrance2ListWithTimeboySlot( uint8_t deviceAddress, uint8_t listNumber )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        res = writeEntrance2ListData4( deviceAddress, listNumber );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// ListenBuffer in AE laden
int32_t Channel4::writeEntrance2ListData4( uint8_t deviceAddress, uint8_t listNumber )
{
    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    uint16_t segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    int32_t res = IV_ERROR_SUCCESS;
    uint32_t address, size, crc;
    address = size = crc = 0;

    if ( controllerSegmentCurrent.transferState == ControllerSegment::BlockTransferState::inactive )
    {
        // Alle Listen laden?
        if ( listNumber == 0xFF )
        {
            // Mindestens eine Liste geladen?
            int32_t i;
            for ( i = 0; i < 8; i++ )
            {
                if ( m_ulEntrance2ListSize[i] != 0 )
                {
                    break;
                }
            }
            if ( i == 8 )
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
        else
        {
            if ( m_ulEntrance2ListSize[listNumber] == 0 )
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }

        // Ggf. gesetztes Streaming deaktivieren.
        setStreamTransfer( false );
        clearStreamTransfer();

        // Systemstopp Setup uebertragen
        while ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
            if ( ((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG) )
            {
                segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
                continue;
            }

            return res;
        }

        // Beginn eintragen.
        controllerSegmentCurrent.reset();
        controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::active;
        controllerSegmentCurrent.transferType = ControllerSegment::btt_access_control_list;
        if ( controllerSegmentCurrent.allowedTransferTypes == ControllerSegment::btt_none )
        {
            // Settings ggf. von Global-Object holen.
            ControllerSegment &controllerSegmentGlobal = controllerSegment( 255 );
            if ( controllerSegmentGlobal.allowedTransferTypes != ControllerSegment::btt_none )
            {
                controllerSegmentCurrent.allowedTransferTypes = controllerSegmentGlobal.allowedTransferTypes;
                controllerSegmentCurrent.duration = controllerSegmentGlobal.duration;
                controllerSegmentCurrent.transferFlags.settingsFromGlobal = true;
            }
        }
        controllerSegmentCurrent.beginTimePoint = std::chrono::system_clock::now();
        // UpdateVariation halten.
        controllerSegmentCurrent.updateVariation = segmentUpdateVariation;

        if ( listNumber == 0xFF )
        {
            controllerSegmentCurrent.number = 0;
            controllerSegmentCurrent.maxNumber = 8;
        }
        else
        {
            controllerSegmentCurrent.number = listNumber;
            controllerSegmentCurrent.maxNumber = listNumber + 1;
        }

        // Pruefen ob die Uebertragung per Stream moeglich ist.
        if ( (res = writeStream( deviceAddress, Stream_type::Update_background, nullptr, 0 )) == IV_ERROR_SUCCESS )
        {
            setStreamTransfer( true );
        }
    }
    else
    {
        // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
        segmentUpdateVariation = controllerSegmentCurrent.updateVariation;
    }

    for ( ; controllerSegmentCurrent.number < controllerSegmentCurrent.maxNumber; ++controllerSegmentCurrent.number )
    {
        // nur geladene Liste bei 0xFF uebertragen
        if ( m_ulEntrance2ListSize[controllerSegmentCurrent.number] != 0 )
        {
            // Liste uebertragen
            if ( (res = writeEntrance2List4( deviceAddress, controllerSegmentCurrent.number )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }

        // Abbruch der Blockuebertragung.
        if ( controllerSegmentCurrent.discard )
            break;
    }

    crc = 0;
    // Systemstopp Setup uebertragen
    if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // BlockTransfer Beendet.
    controllerSegmentCurrent.endTimePoint = std::chrono::system_clock::now();
    controllerSegmentCurrent.transferState = ControllerSegment::BlockTransferState::inactive;

    if ( (res > 0) && (((res >> 8) & 0xFF) == 3) )
    {
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.
    }

    return res;
}

// ListenBuffer in AE laden
int32_t Channel4::writeEntrance2List4( uint8_t deviceAddress, uint8_t listNumber )
{
    // BlockTransfer durchfuehren.
    ControllerSegment &controllerSegmentCurrent = controllerSegment( deviceAddress );

    int32_t res = IV_ERROR_SUCCESS;
    StructTable2 table;
    char name[32];

    //
    uint16_t art = IV_VALUE_SEGMENT_LIST_ACCESS;
    uint32_t address, size, crc;
    address = size = crc = 0;
    uint32_t addressOffset = 0;

    // Bei wiederaufnahme eines Blocktransfers ist dieses bereits durchgeführt worden.
    if ( controllerSegmentCurrent.resume == false )
    {
        memset( name, 0, sizeof( name ) );
        snprintf( name, sizeof( name ), "ACCESS.%i", listNumber );
        // Zur Liste Informationen ermitteln
        if ( (res = readTableStruct( deviceAddress, name, table )) != IV_ERROR_SUCCESS )
        {
            //TODO: Action2 unbekannt Funktionsfehler 2 auswerten.
            return res;
        }
        strncpy( name, table.name, sizeof( name ) - 1 );
        std::string tmpName = name;
        int32_t count = snprintf( table.name, sizeof( table.name ), "ACCESS.%s", tmpName.c_str() );

        if ( count > static_cast<int32_t>(sizeof( table.name )) )
        {
            assert( false );
            return IV_ERROR_LIST_PROCESSING;
        }

        // Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
        if ( m_entrance2SortCheck[listNumber] == false )
        {
            // Listen ohne Datensaetze nicht sortieren.
            // ACHTUNG: Ansonsten Division durch 0 bei Pruefung der Datensatzgroesse !!!
            if ( m_ulEntrance2ListSize[listNumber] )
            {
                int32_t fieldOffs = 0;

                if ( table.flags & TF_SORT && table.sort < table.count )
                {
                    //char name[64];
                    int32_t fieldNum;
                    StructTable2Field field;
                    memset( &field, 0, sizeof( StructTable2Field ) );
                    // Alle Felder fuer Offset lesen
                    for ( fieldNum = 0; fieldNum < table.count; fieldNum++ )
                    {
                        if ( fieldNum < (table.sort + 1) )
                        {
                            // Datensatzoffset fuer Sortierfeld aufsummieren.
                            fieldOffs += field.size;
                        }
                        else // Da die Datensatzgroesse bekannt ist, brauchen die restlichen Felder nicht ermittelt zu werden. 
                        {
                            break;
                        }

                        field = m_entrance2TableField[listNumber][fieldNum];

                        // Datensatzgroesse wird fuer Swap-Routine benoetigt!
                    }

                    // Liste sortieren.
                    SortList( m_pucEntrance2ListMem[listNumber], m_ucEntrance2RecordSize[listNumber], fieldOffs );

                    // Bei allen Listen ist 0 die sortierte Spalte
                    // Flag dafuer dass die Liste sortiert wurde setzen.
                    table.flags |= TF_SORTED;
                }
            }

            // Sortiert oder zumindes geprueft.
            m_entrance2SortCheck[listNumber] = true;
        }

        // CRC berechnen
        crc = 0xFFFFFFFF;
        crc = crc32( (uint8_t *)&table, sizeof( StructTable2 ), crc, false );
        crc = crc32( m_pucEntrance2ListMem[listNumber], m_ulEntrance2ListSize[listNumber], crc );

        // Groesze berechnen
        size = sizeof( StructTable2 ) + m_ulEntrance2ListSize[listNumber];

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {

            // Systemstopp List uebertragen
            while ( (res = enterWriteDataSegment( deviceAddress, art, address, size, crc )) != IV_ERROR_SUCCESS )
            {
                // Typ unbekannt?, dann als "normale" Liste versuchen..
                if ( ((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS) )
                {
                    art = IV_VALUE_SEGMENT_LIST;
                    continue;
                }

                return res;
            }

            // Zusatzdaten
            controllerSegmentCurrent.address = address;
            controllerSegmentCurrent.crc = crc;
        }

        // Daten uebertragen
        if ( (res = writeToFlash( deviceAddress, address, (uint8_t *)&table, sizeof( StructTable2 ) )) != IV_ERROR_SUCCESS )
        {
            // Speicherplatz reicht nicht aus.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }
            return res;
        }

        // Daten uebertragen
        address += sizeof( StructTable2 );

        // Startadresse halten
        controllerSegmentCurrent.address = address;
        // Bereits uebertragene Datenmenge der Tabellenstruktur auf 0 setzen.
        controllerSegmentCurrent.size = 0;
        addressOffset = 0;
    }
    else
    {
        // Zwischengespeicherte Daten fuer Wiederaufnahme setzen.
        address = controllerSegmentCurrent.address;
        crc = controllerSegmentCurrent.crc;
        addressOffset = controllerSegmentCurrent.size;
    }

    // Abbruch der Blockuebertragung.
    if ( controllerSegmentCurrent.discard == false )
    {
        if ( (res = writeToFlash( deviceAddress, address + addressOffset, m_pucEntrance2ListMem[listNumber] + addressOffset, m_ulEntrance2ListSize[listNumber] - addressOffset )) != IV_ERROR_SUCCESS )
        {
            // Speicherplatz reicht nicht aus.
            if ( res == 3 )
            {
                res = -IV_ERROR_NOT_ENOUGH_FLASH;
            }

            if ( res == -IV_ERROR_BLOCK_TRANSFER_BREAK )
            {
                controllerSegmentCurrent.resume = true;
            }

            return res;
        }
    }

    // Listendaten komplett uebertragen.
    controllerSegmentCurrent.resume = false;

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( isStreamTransfer() && m_streamBuffer.size() )
    {
        if ( (res = writeStream( deviceAddress, Stream_type::List_access, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
        {
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clearStreamTransfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if ( !isStreamTransfer() )
    {

        // Systemstopp Setup uebertragen
        if ( (res = exitWriteDataSegment( deviceAddress, art, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    return res;
}

// ListenBuffer in AE laden
int32_t Channel4::writeTimeboyListData4( uint8_t deviceAddress, uint8_t groupId, uint8_t listNumber, int32_t reserve )
{
    // Alle Listen laden?
    if ( groupId == 0xFF )
    {
        int32_t groupID, listNum, found = 0;
        for ( groupID = 0; groupID < MAX_RECORD_DESCRIPTION; groupID++ )
        {
            if ( listNumber == 0xFF )
            {
                // Mindestens eine Liste geladen?
                for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
                {
                    if ( m_ulTimeboyListSize[groupID][listNum] != 0 )
                    {
                        found = 1;
                        break;
                    }
                }
                if ( found )
                {
                    break;
                }
            }
            else
            {
                // Mindestens eine Liste geladen?
                if ( m_ulTimeboyListSize[groupID][listNumber] == 0 )
                {
                    found = 1;
                    break;
                }
            }
        }
        if ( found == 0 )
        {
            return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
        }
    }
    else
    {
        if ( listNumber == 0xFF )
        {
            int32_t listNum, found = 0;
            // Mindestens eine Liste geladen?
            for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
            {
                if ( m_ulTimeboyListSize[groupId][listNum] != 0 )
                {
                    found = 1;
                    break;
                }
            }
            if ( found == 0 )
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
        else
        {
            if ( m_ulTimeboyListSize[groupId][listNumber] == 0 )
            {
                return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
            }
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    uint16_t segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
    int32_t res = IV_ERROR_SUCCESS;
    uint32_t address, size, crc;
    address = size = crc = 0;

    // Systemstopp Listen uebertragen
    while ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
    {
        // Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
        if ( ((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG) )
        {
            segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
            continue;
        }

        return res;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = writeStream( deviceAddress, Stream_type::Update_background, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        setStreamTransfer( true );
    }

    // Alle ids abarbeiten?
    if ( groupId == 0xFF )
    {
        int32_t tableNum, tableCount = 0;
        int32_t groupID;
        for ( groupID = 0; groupID < MAX_RECORD_DESCRIPTION; groupID++ )
        {
            if ( listNumber == 0xFF )
            {
                for ( tableNum = 0; tableNum < MAX_RECORD_DESCRIPTION; tableNum++ )
                {
                    // nur geladene Liste bei 0xFF uebertragen
                    if ( m_ulTimeboyListSize[groupID][tableNum] != 0 )
                    {
                        // Maximale Segmentanzahl nicht ueberschreiten
                        if ( tableCount && !(tableCount % 16) )
                        {
                            crc = 0;
                            // Systemstopp Listen uebertragen
                            if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
                            {
                                return res;
                            }

                            // Systemstopp Listen uebertragen
                            if ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
                            {
                                return res;
                            }
                        }

                        // Liste uebertragen
                        if ( (res = writeTimeboyList4( deviceAddress, groupID, tableNum, reserve )) != IV_ERROR_SUCCESS )
                        {
                            return res;
                        }

                        tableCount++;
                    }
                }
            }
            else
            {
                // nur geladene Liste bei 0xFF uebertragen
                if ( m_ulTimeboyListSize[groupID][listNumber] != 0 )
                {
                    // Maximale Segmentanzahl nicht ueberschreiten
                    if ( tableCount && !(tableCount % 16) )
                    {
                        crc = 0;
                        // Systemstopp Listen uebertragen
                        if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
                        {
                            return res;
                        }

                        // Systemstopp Listen uebertragen
                        if ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
                        {
                            return res;
                        }
                    }

                    // Liste uebertragen
                    if ( (res = writeTimeboyList4( deviceAddress, groupID, listNumber, reserve )) != IV_ERROR_SUCCESS )
                    {
                        return res;
                    }

                    tableCount++;
                }
            }
        }
    }
    else
    {
        if ( listNumber == 0xFF )
        {
            int32_t tableNum, tableCount = 0;
            for ( tableNum = 0; tableNum < MAX_RECORD_DESCRIPTION; tableNum++ )
            {
                // nur geladene Liste bei 0xFF uebertragen
                if ( m_ulTimeboyListSize[groupId][tableNum] != 0 )
                {
                    // Maximale Segmentanzahl nicht ueberschreiten
                    if ( tableCount && !(tableCount % 16) )
                    {
                        crc = 0;
                        // Systemstopp Listen uebertragen
                        if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
                        {
                            return res;
                        }

                        // Systemstopp Listen uebertragen
                        if ( (res = enterWriteDataSegment( deviceAddress, segmentUpdateVariation, address, size, crc )) != IV_ERROR_SUCCESS )
                        {
                            return res;
                        }
                    }

                    // Liste uebertragen
                    if ( (res = writeTimeboyList4( deviceAddress, groupId, tableNum, reserve )) != IV_ERROR_SUCCESS )
                    {
                        return res;
                    }

                    tableCount++;
                }
            }
        }
        else
        {
            // Liste uebertragen
            if ( (res = writeTimeboyList4( deviceAddress, groupId, listNumber, reserve )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
    }

    crc = 0;
    // Systemstopp Listen uebertragen
    if ( (res = exitWriteDataSegment( deviceAddress, segmentUpdateVariation, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    return res;
}

int32_t Channel4::writeTimeboyListWithTimeboySlot( uint8_t deviceAddress, uint8_t groupId, uint8_t listNumber, int32_t reserve )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Listendaten zum AEIV uebertragen
        res = writeTimeboyListData4( deviceAddress, groupId, listNumber, reserve );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// ListenBuffer in AE laden
int32_t Channel4::writeTimeboyList4( uint8_t deviceAddress, uint8_t groupId, uint8_t listNumber, int32_t reserve )
{
    (void)reserve;
    int32_t res = IV_ERROR_SUCCESS;
    StructTable2 table;
    memset( &table, 0, sizeof( StructTable2 ) );
    table.number = listNumber;
    snprintf( table.name, sizeof( table.name ), "TBLIST.%i.%i", groupId, listNumber );
    table.flags = m_timeboyFlags[groupId][listNumber];

    //
    uint16_t art = IV_VALUE_SEGMENT_LIST_TIMEBOY;
    uint32_t address, size, crc;
    address = size = crc = 0;
    /*
    // Systemstopp Setup uebertragen
    if ((res = SegmentBegin(deviceAddress, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
    {
    return res;
    }
    */
    // CRC berechnen
    crc = 0xFFFFFFFF;
    crc = crc32( (uint8_t *)&table, sizeof( StructTable2 ), crc, false );
    crc = crc32( m_pucTimeboyListMem[groupId][listNumber], m_ulTimeboyListSize[groupId][listNumber], crc );

    // Groesze berechnen
    size = sizeof( StructTable2 ) + m_ulTimeboyListSize[groupId][listNumber];

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if ( !isStreamTransfer() )
    {

// Systemstopp List uebertragen
        while ( (res = enterWriteDataSegment( deviceAddress, art, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            // Typ unbekannt?, dann als "normale" Liste versuchen..
            if ( ((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS) )
            {
                art = IV_VALUE_SEGMENT_LIST;
                continue;
            }

            return res;
        }
    }

    // Daten uebertragen
    if ( (res = writeToFlash( deviceAddress, address, (uint8_t *)&table, sizeof( StructTable2 ) )) != IV_ERROR_SUCCESS )
    {
        // Speicherplatz reicht nicht aus.
        if ( res == 3 )
        {
            res = -IV_ERROR_NOT_ENOUGH_FLASH;
        }
        return res;
    }
    // Daten uebertragen
    address += sizeof( StructTable2 );
    if ( (res = writeToFlash( deviceAddress, address, m_pucTimeboyListMem[groupId][listNumber], m_ulTimeboyListSize[groupId][listNumber] )) != IV_ERROR_SUCCESS )
    {
        // Speicherplatz reicht nicht aus.
        if ( res == 3 )
        {
            res = -IV_ERROR_NOT_ENOUGH_FLASH;
        }
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( isStreamTransfer() && m_streamBuffer.size() )
    {
        if ( (res = writeStream( deviceAddress, Stream_type::List_timeboy, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
        {
// Gepufferte Daten loeschen.
            clearStreamTransfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clearStreamTransfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if ( !isStreamTransfer() )
    {

// Systemstopp Setup uebertragen
        if ( (res = exitWriteDataSegment( deviceAddress, art, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    if ( (res > 0) && (((res >> 8) & 0xFF) == 3) )
        return -IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.

    /*
    crc = 0;
    // Systemstopp Setup uebertragen
    if ((res = SegmentEnd(deviceAddress, SS_UPDATE, crc)) != IV_ERROR_SUCCESS)
    {
    return res;
    }
    */
    return res;
}

int32_t Channel4::writeDeviceWithTimeboySlot( uint8_t deviceAddress, std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeDevice( deviceAddress, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeDevice( uint8_t deviceAddress, std::string &filePath )
{
    if ( !m_deviceUpdate )
    {
        return -IV_ERROR_INTERN;
    }

    // Hardware- und Softwareinfo der ausgewaehlten Datei.
    StructHardwareInfo hwInfo;

    int32_t iRes = IV_ERROR_SUCCESS;

    // Puffer fuer die Hardwareinfo in Textform
    m_hardwareInfoAsText.clear();
    m_minimalUpdateVersion.reset();
    m_blacklistVersion.clear();

    // Strukturen loeschen.
    memset( &hwInfo, 0, sizeof( StructHardwareInfo ) );

    //######################################################################
    // 1. Ermittlung ob auf dem Geraet eine Applicationfirmware laeuft.
    // 2. Wenn Punkt 1 == TRUE, dann wird geprueft ob Datensaetze verfuegbar sind.
    // 2. Wenn Punkt 2 == TRUE, wird die Aktion abgebrochen.
    std::string version;
    int32_t modus = 0;
    // Geraetemodus (Application/Bootloader) ermitteln
    if ( (iRes = ModuleTable::readFirmwareVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
    {
        if ( IV_RESULT_MODULE( iRes ) == 2 )
        {
            // Bootloadermodus
            modus = 1;
        }
        else
        {
            return iRes;
        }
    }

    iRes = 0;
    // Abfrage nur wenn Geraet im Applicationmodus
    // Abfrage ob Daten vorhanden
    if ( !modus && (iRes = checkData( deviceAddress )) == -1 )
    {
        return -IV_ERROR_UNREACHABLE;
    }
    else if ( iRes )
    {
        return -IV_ERROR_DATA_AVAILABLE;
    }

    //#########################################################################
    // Pruefen ob gewaehlte Datei in das Geraet eingespielt werden kann.
    // Dateityp ermitteln
    struct stat st;
    if ( stat( filePath.c_str(), &st ) == -1 )
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Angabe ist eine Datei
    if ( S_ISREG( st.st_mode ) )
    {
        // Nur ".hex" gueltig.
        size_t pos = filePath.size() - 4;
#if defined(_MSC_VER) || defined(__MINGW32__)
        if ( filePath[pos] != '.' ||
             ((!(filePath[pos + 1] == 'h' || filePath[pos + 1] == 'H') ||
                !(filePath[pos + 2] == 'e' || filePath[pos + 2] == 'E') ||
                !(filePath[pos + 3] == 'x' || filePath[pos + 3] == 'X')) &&
               (!(filePath[pos + 1] == 'd' || filePath[pos + 1] == 'D') ||
                 !(filePath[pos + 2] == 'f' || filePath[pos + 2] == 'F') ||
                 !(filePath[pos + 3] == 'z' || filePath[pos + 3] == 'Z')) &&
               (!(filePath[pos + 1] == 'z' || filePath[pos + 1] == 'Z') ||
                 !(filePath[pos + 2] == 'i' || filePath[pos + 2] == 'I') ||
                 !(filePath[pos + 3] == 'p' || filePath[pos + 3] == 'P'))) )
        {
            return -IV_ERROR_FILE_EXTENSION;
        }
#else 
        if ( !S_ISREG( st.st_mode ) || filePath[pos] != '.' ||
             !(filePath[pos + 1] == 'h' || filePath[pos + 1] == 'H') ||
             !(filePath[pos + 2] == 'e' || filePath[pos + 2] == 'E') ||
             !(filePath[pos + 3] == 'x' || filePath[pos + 3] == 'X') )
        {
            return -IV_ERROR_FILE_EXTENSION;
        }
#endif	

        // Interne Daten des letzten Aufrufs zurücksetzen (fwList, minVersion und dbBlacklist)
        m_deviceUpdate->resetInternalData();

        // Handelt es sich um eine Hexdatei
        if ( filePath[pos + 1] == 'h' || filePath[pos + 1] == 'H' )
        {
            // Dateidaten laden
            m_fileDevice.resetData();

            trace( "%s(%i):INFO(1): %s [ Laden der Firmwaredatei. (%s) ]\n", __FILE__, __LINE__, __func__, filePath.c_str() );

            if ( !m_fileDevice.load( filePath ) )
            {
                int32_t err = IV_ERROR_SUCCESS;
                if ( m_fileDevice.errorState() == FileIntelHex::Error_state::version
                     || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown )
                {
                    err = -IV_ERROR_VERSION_DEVICEFILE;
                }
                else if ( m_fileDevice.errorState() == FileIntelHex::Error_state::file_open )
                {
                    err = -IV_ERROR_FILE_OPEN;
                }
                else
                {
                    err = -IV_ERROR_DEVICEFILE_LOAD;
                }

                m_fileDevice.resetData();
                return err;
            }

            trace( "%s(%i):INFO(1): %s [ Laden der Firmwaredatei erfolgreich. ]\n", __FILE__, __LINE__, __func__ );

            // Ist die Softwareinformation in der Geraetedatei verfuegbar?
            if ( m_fileDevice.GetHeaderSoftwareInfo().size )
            {
                HardwareInfo hardwareInfo;
                // Auslesen der Hardwareinfo
                if ( (iRes = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
                {
                    return iRes;
                }

                // Uebergangsweise auf die Vorgaengerversion wechseln.
                hwInfo = hardwareInfo.exportStructHardwareInfo();

                // Liegt keine Hardwareinfo vor?
                // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
                if ( hwInfo.version == 0 )
                {
                    int32_t handle;
                    std::string reserved;
                    if ( openTable( deviceAddress, "ACCESS.READER", 0, handle, reserved ) == IV_ERROR_SUCCESS &&
                         closeTable( deviceAddress, handle ) == IV_ERROR_SUCCESS )
                    {
                        // ZK-Option in Verwendung.
                        hwInfo.administrationFlags |= HIAF_HAS_ZK;
                    }
                }

                trace( "%s(%i):INFO(1): %s [ Firmwaredatei pruefen. ]\n", __FILE__, __LINE__, __func__ );

                if ( m_fileDevice.CompareInfos( hwInfo, m_fileDevice.GetHeaderSoftwareInfo() ) == 0 )
                {
                    // Hardwareinfo in Text aufloesen
                    hardwareInfoToString( hwInfo, m_hardwareInfoAsText );
                    m_fileDevice.setErrorTextExtended( m_fileDevice.errorText() );
                    return -IV_ERROR_FW_HEX_FAILED_INFO;
                }

                trace( "%s(%i):INFO(1): %s [ Firmwaredatei Pruefung abgeschlossen. ]\n", __FILE__, __LINE__, __func__ );
            }
            else
            {
                uint8_t deviceType = 255, boardType = 255;
                // Geraetetyp ermitteln
                if ( (iRes = readDeviceType( deviceAddress, deviceType )) != IV_ERROR_SUCCESS || (iRes = readBoardType( deviceAddress, boardType )) != IV_ERROR_SUCCESS )
                {
                    if ( iRes == -IV_ERROR_FUNCTION_UNAVAILABLE )
                    {
                        // Der Geraetetyp wird manuell gesetzt.
                        deviceType = 0;
                    }
                    else
                    {
                        return iRes;
                    }
                }

                //#######################################################################
                // Wenn die SoftwareInfo fehlt kann die Datei keine Mulitple sein!!!

                switch ( deviceType )
                {
                    case IV_VALUE_TYPE_DEVICE_PZE: // PZE-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_ZK: // ZK-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das ZK-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_BDE: // BDE-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_BDE )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das BDE-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_MOBIL: // Mobil-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_MDE: // MDE-BoxIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MDE )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das MDE-BoxIV
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_AE: // AE-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_AE )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das AE-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_TIMEBOY: // TimeboyIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das TimeboyIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_EXKLUSIV: // Exklusiv-LineIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_EXKLUSIV )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Exklusiv-LineIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_FLEX: // Flex-MasterIV
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_FLEX )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Flex-MasterIV.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE: // Timeboy-Mobil-PZE
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY_MOBIL_PZE )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Timeboy-Mobil-PZE.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_EVO_2_8_PURE: // Terminal 2.8
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TERMINAL_2_8 )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Terminal 2.8.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_EVO_4_3: // Terminal 4.3
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TERMINAL_4_3 )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Terminal 4.3.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7: // IPC-Extension
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_5_7 )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10: // IPC-Extension
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_VARIO_10 )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7: // IPC-Extension
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IPC_EXTENSION_Q7 )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das IPC-Extension.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_IO_BOX: // IO-Box
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_IO_BOX )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das IO-Box.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_KYO_INLOC: // KYO Inloc
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_KYO_INLOC )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer den KYO Inloc.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    case IV_VALUE_TYPE_DEVICE_MOBIL_BOX: // Mobil-Box
                        if ( m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL_BOX )
                        {
                            // Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-Box.
                            return -IV_ERROR_DEVICEFILE_WRONG;
                        }
                        break;

                    default:
                        if ( m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO
                             && (m_fileDevice.GetHeaderSoftwareInfo().deviceExtended & (1 << deviceType)) != 0 )
                        {
                            // Die Daten sind für den vorliegenden Gerätetyp geeignet.
                        }
                        else
                        {
                            if ( m_fileDevice.GetHeaderVersionFile() == IV_DFFILEDEVICE_VERSION_SOFTWARE_INFO )
                            {
                                // Die gewaehlte Geraetedatei ist NICHT fuer diese Gerätevariante.
                                return -IV_ERROR_DEVICEFILE_WRONG;
                            }
                            else
                            {
                                // Die Setupanwendung ist nicht aktuell genug fuer diesen Geraetetyp.
                                return -IV_ERROR_DEVICE_UNKNOWN;
                            }
                        }
                }

                std::string version;
                if ( boardType < IV_VALUE_TYPE_BOARD_3_0 )
                {
                    // Handelt es sich um einen Boardtyp < 3 muss die Firmwareversion 04.00.xx oder 04.01.xx sein.
                    if ( m_fileDevice.GetHeaderVersionFirmware( version ) )
                    {
                        if ( version.compare( 0, 6, "04.00." ) != 0 && version.compare( 0, 6, "04.01." ) != 0 )
                        {
                            return -IV_ERROR_DEVICEFILE_MATCH;
                        }
                    }
                }
                else
                {
                    // Handelt es sich um einen Boardtyp >= 3 darf die Firmwareversion nicht 04.00.xx oder 04.01.xx sein.
                    if ( m_fileDevice.GetHeaderVersionFirmware( version ) )
                    {
                        if ( version.compare( 0, 6, "04.00." ) == 0 || version.compare( 0, 6, "04.01." ) == 0 )
                        {
                            return -IV_ERROR_DEVICEFILE_MATCH;
                        }
                    }
                }

            }
        }
        else
        { // Es handelt sich um eine Zipdatei

            HardwareInfo hardwareInfo;
            // Auslesen der Hardwareinfo
            iRes = detectHWInfo( deviceAddress, hwInfo, hardwareInfo );
            if ( iRes != IV_ERROR_SUCCESS )
            {
                return iRes;
            }

            // Modulinformationen lesen und die erlaubte Mindestversion ermitteln
            std::list<std::pair<std::string, std::string>> processedModules{};  // Liste aus Paaren von Modulkennung und Mindestversion
            iRes = determineMinFWVersion( hwInfo, hardwareInfo, filePath, processedModules );
            switch ( iRes )
            {
                case -IV_ERROR_DEVICEFILE_LOAD:
                case -IV_ERROR_DEVICEFILE_MATCH:
                case -IV_ERROR_FILE_PROCESSING:
                {
                    hardwareInfoToString( hwInfo, m_hardwareInfoAsText, &processedModules );
                    cleanUpUpdateDB();
                    return iRes;
                }
                case IV_ERROR_SUCCESS:
                {
                    m_minimalUpdateVersion = m_deviceUpdate->minFwVersion();
                    m_blacklistVersion = m_deviceUpdate->blacklistVersions();
                    break;
                }
                // IV_ERROR_USER_CANCEL is returned if it is an old device and the check against the database is not performed.
            }

            if ( determineFWFile( filePath, hwInfo ) )
            {
                trace( "%s(%i):INFO(1): %s [ Extrahieren der passenden Firmwaredatei. (%s) ]\n", __FILE__, __LINE__, __func__, filePath.c_str() );

                if ( (0 == m_fileDevice.loadFirmwareFileZip( filePath, m_deviceUpdate->latestFwFile() )) )
                {
                    int32_t err = IV_ERROR_SUCCESS;
                    // Hardwareinfo in Text aufloesen
                    hardwareInfoToString( hwInfo, m_hardwareInfoAsText, &processedModules );
                    if ( m_fileDevice.errorState() == FileIntelHex::Error_state::version
                         || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown )
                    {
                        err = -IV_ERROR_VERSION_DEVICEFILE;
                    }
                    else if ( m_fileDevice.errorState() == FileIntelHex::Error_state::no_file_match )
                    {
                        err = -IV_ERROR_DEVICEFILE_MATCH;
                    }
                    else
                    {
                        err = -IV_ERROR_DEVICEFILE_LOAD;
                    }
                    cleanUpUpdateDB();
                    return err;
                }

                trace( "%s(%i):INFO(1): %s [ Extrahierte Firmwaredatei. (%s) ]\n", __FILE__, __LINE__, __func__, filePath.c_str() );
            }
            else
            {
                cleanUpUpdateDB();
                return -IV_ERROR_DEVICEFILE_MATCH;
            }
        }
    }
    else if ( S_ISDIR( st.st_mode ) )
    {
#if defined(_MSC_VER) || defined(__MINGW32__)
        // Auswahl aus Verzeichnis nur unter Linux gueltig!
        return -IV_ERROR_FUNCTION_PARAMETER;
#else
        // Interne Daten des letzten Aufrufs zurücksetzen (fwList, minVersion und dbBlacklist)
        m_deviceUpdate->resetInternalData();

        HardwareInfo hardwareInfo;
        // Auslesen der Hardwareinfo
        if ( (iRes = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
        {
            return iRes;
        }

        // Uebergangsweise auf die Vorgaengerversion wechseln.
        hwInfo = hardwareInfo.exportStructHardwareInfo();

        // Liegt keine Hardwareinfo vor?
        // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
        if ( hwInfo.version == 0 )
        {
            int32_t handle;
            std::string tmp;
            if ( openTable( deviceAddress, (const char *)"ACCESS.READER", 0, handle, tmp ) == IV_ERROR_SUCCESS &&
                 closeTable( deviceAddress, handle ) == IV_ERROR_SUCCESS )
            {
                // ZK-Option in Verwendung.
                hwInfo.administrationFlags |= HIAF_HAS_ZK;
            }
        }

        bool useUpdateDB = false;
        std::string tmpDBPath = filePath + "/firmwareUpdate.db";
        std::string errorMsg;
        if ( Utils::File::exists( tmpDBPath ) )
        {
            useUpdateDB = m_deviceUpdate->updatePathToDB( tmpDBPath, DeviceUpdate::DBPathType::Local, errorMsg );
        }

        // Modulinformationen lesen und die erlaubte Mindestversion ermitteln
        std::list<std::pair<std::string, std::string>> processedModules{};  // Liste aus Paaren von Modulkennung und Mindestversion
        if ( useUpdateDB && !hardwareInfo.isMasterIV( hwInfo ) )
        {
            trace( "%s(%i):INFO(1): %s [ Ermittlung der Mindestfirmwareversion. ]\n", __FILE__, __LINE__, __func__ );
            for ( auto &item : hardwareInfo.fileHardwareInfo().rootItemDeviceHardwareInfo().childs() )
            {
                if ( !m_deviceUpdate->detectMinFWVersion( item, processedModules ) )
                {
                    break;
                }
            }

            trace( "%s(%i):INFO(1): %s Moduluebersicht\n", __FILE__, __LINE__, __func__ );
            trace( "%s(%i):INFO(1): %s ---------------\n", __FILE__, __LINE__, __func__ );
            for ( const auto &item : processedModules )
            {
                trace( "%s(%i):INFO(1): %s %s\t--> Minimalversion: %s\n", __FILE__, __LINE__, __func__, item.first.c_str(), item.second.c_str() );
            }

            trace( "%s(%i):INFO(1): %s [ Ermittlung der Blacklistversionen. ]\n", __FILE__, __LINE__, __func__ );
            for ( auto &item : hardwareInfo.fileHardwareInfo().rootItemDeviceHardwareInfo().childs() )
            {
                if ( !m_deviceUpdate->collectBlacklistDefinition( item ) )
                {
                    break;
                }
            }

            m_minimalUpdateVersion = m_deviceUpdate->minFwVersion();
            m_blacklistVersion = m_deviceUpdate->blacklistVersions();
        }

        // Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
        if ( (0 == m_fileDevice.SearchInfosDir( filePath, hwInfo, m_deviceUpdate.get() )) &&
             (0 == m_fileDevice.loadFirmwareFile( filePath, m_deviceUpdate->latestFwFile() )) )
        {
            int32_t err = IV_ERROR_SUCCESS;
            // Hardwareinfo in Text aufloesen
            hardwareInfoToString( hwInfo, m_hardwareInfoAsText, &processedModules );
            if ( m_fileDevice.errorState() == FileIntelHex::Error_state::version
                 || m_fileDevice.errorState() == FileIntelHex::Error_state::device_unknown )
            {
                err = -IV_ERROR_VERSION_DEVICEFILE;
            }
            else if ( m_fileDevice.errorState() == FileIntelHex::Error_state::no_file_match )
            {
                err = -IV_ERROR_DEVICEFILE_MATCH;
            }
            else
            {
                err = -IV_ERROR_DEVICEFILE_LOAD;
            }
            cleanUpUpdateDB();
            return err;
        }
#endif
    }
    else
    {
        return -IV_ERROR_FILE_OPEN;
    }

    cleanUpUpdateDB();
    trace( "%s(%i):INFO(1): %s [ Weiterverarbeitung der zu uebertragenden Firmwaredatei '%s'. ]\n", __FILE__, __LINE__, __func__, filePath.c_str() );

    // 2013_04_17:
    // Es muss sichergestellt werden, dass eine Firmwareversion < 04.02.04.20 nur auf Timeboy Geraete mit Ladeschaltung aufgespielt werden kann.
    // Wenn dieses geschieht, dann geht der Timeboy aus und rebooted ständig.
    if ( m_fileDevice.GetHeaderVersionFirmware() >= "04.02.01.00" && m_fileDevice.GetHeaderVersionFirmware() < "04.02.04.20" )
    {
        // Im Bootloadermodus koennen wir das Flag der Ladeschaltung nicht auswerten, dadurch lassen wir das aufspielen zu.
        if ( modus == 0 && SoftwarebuildVersion( version ) >= "04.02.04.20" )
        {
            if ( (hwInfo.typeDevice == IV_VALUE_TYPE_DEVICE_TIMEBOY || hwInfo.typeDevice == IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE)
                 && !hwInfo.componentCharge ) // Bei Timeboy's ohne Ladeschaltung ist das aufspielen unzulaessig.
            {
                return -IV_ERROR_DEVICEFILE_MATCH;
            }
        }
    }

    // Befehlsausfuehrung starten.
    iRes = writeDevice4( deviceAddress );

    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Ggf. unbekannte Dockingstation
        protocol().resetTimeboyStation();
    }

    return iRes;
}

int32_t Channel4::writeDevice4( uint8_t deviceAddress )
{
    // Zeichensatz- und Grafikdaten
    FileDisplayFonts *pFileDisplayFonts = m_fileDevice.GetFileDisplayFonts();
    FileDisplayGrafics *pFileDisplayGrafics = m_fileDevice.GetFileDisplayGrafics();
    FileDeviceText *pFileDeviceText = m_fileDevice.GetFileDeviceText();
    FileIntelHex *pFileRawFirmware = m_fileDevice.GetFileRawFirmware();
    int32_t res = IV_ERROR_SUCCESS;
    uint32_t address = 0;
    uint8_t *buffer;
    uint32_t size;
    uint32_t crc = 0;
    std::string version;

    // Bootversion lesen 
    if ( (res = readBootloaderVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, 0, 0 )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Pruefen ob die Uebertragung per Stream moeglich ist.
    if ( (res = writeStream( deviceAddress, Stream_type::Update, nullptr, 0 )) == IV_ERROR_SUCCESS )
    {
        setStreamTransfer( true );
    }

    // Sind Textdaten vorhanden?
    if ( pFileDeviceText->sizeData() )
    {
        // Texte als ersten Block uebertragen
        // TEXTE

        address = 0;
        buffer = pFileDeviceText->data();
        size = pFileDeviceText->sizeData();
        crc = pFileDeviceText->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
// Systemstopp setzen
            res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TEXT, address, size, crc );
            if ( res != IV_ERROR_SUCCESS && (res & 0xFF) != 1 )
            {
                return res;
            }
        }

        // Funktionsfehler nicht Blocktyp unbekannt?
        if ( (res & 0xFF) != 1 )
        {
            // Textdaten uebertragen
            if ( (res = writeToFlash( deviceAddress, address, buffer, size )) != IV_ERROR_SUCCESS )
            {
                return res;
            }

            // Im Modus Streaming ?, dann gepufferte Daten senden.
            if ( isStreamTransfer() && m_streamBuffer.size() )
            {
                if ( (res = writeStream( deviceAddress, Stream_type::Text, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
                {
// Gepufferte Daten loeschen.
                    clearStreamTransfer();
                    return res;
                }
                // Gepufferte Daten loeschen.
                clearStreamTransfer();
            }

            // Bei Streaming werden die inneren Bloecke wegelassen.
            if ( !isStreamTransfer() )
            {
// Systemstopp aufheben
                if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_TEXT, crc )) != IV_ERROR_SUCCESS )
                {
                    return res;
                }
            }
        }
        else
        {
            // Texte als Zeichensatz mit uebertragen.
            int32_t err;
            if ( !(err = pFileDisplayFonts->AppendSegmentText( buffer, size )) )
            {
                if ( FileIntelHex::Error_state::alloc_mem == pFileDisplayFonts->errorState() )
                {
                    return -IV_ERROR_ALLOC_MEM;
                }

                return -IV_ERROR_FILE_PROCESSING;
            }

            // Ruecksetzen des Flags fuer "gueltige Default-Textdaten vorhanden"
            uint8_t v = 0;
            if ( (res = writeToEeprom( deviceAddress, 0x0ECB, &v, 1 )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
    }

    // ZEICHENSAETZE
    if ( pFileDisplayFonts->sizeData() )
    {
        address = 0;
        buffer = pFileDisplayFonts->data();
        size = pFileDisplayFonts->sizeData();
        crc = pFileDisplayFonts->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
// Systemstopp setzen
            if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FONTS, address, size, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }

        // Zeichensatzdaten uebertragen
        if ( (res = writeToFlash( deviceAddress, address, buffer, size )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( isStreamTransfer() && m_streamBuffer.size() )
        {
            if ( (res = writeStream( deviceAddress, Stream_type::Fonts, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
            {
// Gepufferte Daten loeschen.
                clearStreamTransfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
// Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FONTS, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }
    }

    // GRAFIKEN
    if ( pFileDisplayGrafics->sizeData() )
    {
        address = 0;
        buffer = pFileDisplayGrafics->data();
        size = pFileDisplayGrafics->sizeData();
        crc = pFileDisplayGrafics->checksum32bit();

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
// Systemstopp setzen
            if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_GRAFICS, address, size, crc )) != IV_ERROR_SUCCESS )
            {
                return res;
            }
        }

        // Grafikdaten uebertragen
        if ( (res = writeToFlash( deviceAddress, address, buffer, size )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Im Modus Streaming ?, dann gepufferte Daten senden.
        if ( isStreamTransfer() && m_streamBuffer.size() )
        {
            if ( (res = writeStream( deviceAddress, Stream_type::Graphics, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
            {
// Gepufferte Daten loeschen.
                clearStreamTransfer();
                return res;
            }
            // Gepufferte Daten loeschen.
            clearStreamTransfer();
        }

        // Bei Streaming werden die inneren Bloecke wegelassen.
        if ( !isStreamTransfer() )
        {
// Systemstopp aufheben
            if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_GRAFICS, crc )) != IV_ERROR_SUCCESS )
            {

                return res;
            }
        }
    }

    // FIRMWARE

    address = 0;
    buffer = pFileRawFirmware->data();
    size = pFileRawFirmware->sizeData();
    crc = pFileRawFirmware->checksum32bit();

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if ( !isStreamTransfer() )
    {
// Systemstopp setzen
        if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FIRMWARE, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    // Firmwaredaten uebertragen
    if ( (res = writeToFlash( deviceAddress, address, buffer, size )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Im Modus Streaming ?, dann gepufferte Daten senden.
    if ( isStreamTransfer() && m_streamBuffer.size() )
    {
        if ( (res = writeStream( deviceAddress, Stream_type::Firmware, m_streamBuffer.data(), static_cast<uint32_t>(m_streamBuffer.size()) )) != IV_ERROR_SUCCESS )
        {
// Gepufferte Daten loeschen.
            clearStreamTransfer();
            return res;
        }
        // Gepufferte Daten loeschen.
        clearStreamTransfer();
    }

    // Bei Streaming werden die inneren Bloecke wegelassen.
    if ( !isStreamTransfer() )
    {
// Systemstopp aufheben
        if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FIRMWARE, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
    }

    crc = 0;
    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, crc )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Neuaushandlung bei verschluesselter Kommunikation
    AesCryptSession().resetSessionKey();

    // Firmwareversion des Geraetes.
    m_versionFirmware.clear();

    return res;
}

int32_t Channel4::writeModuleWithTimeboySlot( uint8_t deviceAddress, int32_t type, const std::string &filePath )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = writeModule( deviceAddress, type, filePath );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeModule( uint8_t deviceAddress, int32_t type, const std::string &filePath )
{
    (void)type;
    int32_t res;
    uint32_t address = 0;
    uint8_t *buf = nullptr;
    uint32_t size = 0;
    uint32_t crc = 0;

    std::string version;
    HardwareInfo hardwareInfo;
    listItemBoard tagsBiokey;

    //	memset(version, 0, sizeof(version));
    // Bootversion lesen 
    //	if ((res = getVersionBootloader(deviceAddress, version)) != IV_ERROR_SUCCESS) {

    //		return res;
    //	}

    // Dateigroesse ermitteln
    // Datei oeffnen
    struct stat st;
    FILE *fp = nullptr;

    if ( stat( filePath.c_str(), &st ) != 0 || st.st_size == 0 || (fp = fopen( filePath.c_str(), "rb" )) == nullptr )
    {

        res = -IV_ERROR_FILE_OPEN;
        goto error_label;
    }

    // Dateigroesse muss groesser 4kByte sein.
    if ( st.st_size < 4096 )
    {
        res = -IV_ERROR_FILE_FORMAT;
        goto error_label;
    }

    // Speicher reservieren
    buf = new uint8_t[st.st_size + 1];
    if ( buf == nullptr )
    {

        res = -IV_ERROR_FILE_OPEN;
        goto error_label;
    }

    // Daten laden
    if ( (int32_t)fread( buf, sizeof( uint8_t ), st.st_size, fp ) != st.st_size )
    {

        res = -IV_ERROR_FILE_READ;
        goto error_label;
    }

    // Zugriff ueber Struktur initialisieren
    m_uploadModuleSignatur = (UploadModuleSignatur *)&buf[st.st_size - sizeof( UploadModuleSignatur )];
    // Gesamt-CRC pruefen
    crc = crc32( buf, st.st_size - sizeof( uint32_t ) );
    if ( crc != m_uploadModuleSignatur->crc32 )
        //if (crc = crc32(buf, st.st_size))
    {
        res = -IV_ERROR_FILE_CRC;
        goto error_label;
    }

    // Pruefung auf unterstutzte Version
    if ( m_uploadModuleSignatur->version != 1 )
    {
        res = -IV_ERROR_FILE_FORMAT;
        goto error_label;
    }

    // HardwareInfo abrufen.
    if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
    {
        if ( res > 0 )
        {

            if ( (res & 0xFF) == 28 )
            {

// Passwort fuer Kommunikation wird benoetigt.
                res = -IV_ERROR_NEED_PASSWORD;
            }
            else if ( (res & 0xFF) == 29 )
            {

// Nach 3 maliger Passwortfalscheingabe ist Geraet fuer 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            }
            else
            {

                     // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    tagsBiokey = hardwareInfo.moduleBiokey();
    if ( tagsBiokey.size() == 0 )
    {
        return -IV_ERROR_NO_MODULE;
    }
    else if ( tagsBiokey.size() > 1 )
    {
        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }
    else
    {
        // 1 = Idencom (BioKey 3000 - Modul)
        // 2 = Idencom (BioKey 4000 - Modul)
        // 3 = Idencom (BioKey 4020 - Modul)
        uint8_t val = IV_VALUE_TYPE_FINGERPRINT_NONE;
        switch ( tagsBiokey.front()->hc().id )
        {
            case EnumModuleBiokey::Biokey3000:
                val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_3000;
                break;

            case EnumModuleBiokey::Biokey4000:
                val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4000;
                break;

            case EnumModuleBiokey::Biokey4020:
                val = IV_VALUE_TYPE_FINGERPRINT_BIOKEY_4020;
                break;

            default:
                break;
        }

        // Pruefung auf zu aktualisierendes Modul
        if ( m_uploadModuleSignatur->type != val )
        {
            // In der Datei liegt ein Update fuer eine andere Modulversion vor
            res = -IV_ERROR_FILE_DATA_FORMAT;
            goto error_label;
        }
    }

    // Ggf. gesetztes Streaming deaktivieren.
    setStreamTransfer( false );
    clearStreamTransfer();

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, address, size, crc )) != IV_ERROR_SUCCESS )
    {

        if ( res > 0 )
        {

// Wenn der Modulfehler -Funktion unbekannt- ist,
// dann ist die Firmwareversion nicht kompatible zum Befehl.
            if ( ((res >> 8) & 0xFF) == 2 )
            {

// Hintergrundladen noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            }
            else if ( (res & 0xFF) == 28 )
            {

// Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            }
            else if ( (res & 0xFF) == 29 )
            {

// Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            }
            else
            {

                     // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }


    // FIRMWARE des Biokey laden

    address = 0;
    size = st.st_size - sizeof( UploadModuleSignatur );
    crc = crc32( buf, size );

    // Systemstopp setzen
    if ( (res = enterWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FINGERPRINT, address, size, crc )) != IV_ERROR_SUCCESS )
    {

        if ( res > 0 )
        {

// Wenn der Modulfehler -Funktion unbekannt- ist,
// dann ist die Firmwareversion nicht kompatible zum Befehl.
            if ( ((res >> 8) & 0xFF) == 2 )
            {

// Fingerprintmodul noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            }
            else if ( (res & 0xFF) == 28 )
            {

// Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            }
            else if ( (res & 0xFF) == 29 )
            {

// Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            }
            else
            {

                     // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Firmwaredaten uebertragen
    if ( (res = writeToFlash( deviceAddress, address, buf, size )) != IV_ERROR_SUCCESS )
    {

        if ( res > 0 )
        {

// Funktionsfehler
            res = -IV_ERROR_PROCESSING;
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_FINGERPRINT, crc )) != IV_ERROR_SUCCESS )
    {

        if ( res > 0 )
        {

// Wenn der Modulfehler -Funktion unbekannt- ist,
// dann ist die Firmwareversion nicht kompatible zum Befehl.
            if ( ((res >> 8) & 0xFF) == 2 )
            {

// Hintergrundladen noch nicht implementiert
                res = -IV_ERROR_VERSION_FIRMWARE;
            }
            else if ( (res & 0xFF) == 28 )
            {

// Passwort f?r Kommunikation wird ben?tigt.
                res = -IV_ERROR_NEED_PASSWORD;
            }
            else if ( (res & 0xFF) == 29 )
            {

// Nach 3 maliger Passwortfalscheingabe ist Ger?t f?r 60 Sekunden gesperrt.
                res = -IV_ERROR_PASSWORD_LOCK;
            }
            else
            {

                     // Funktionsfehler
                res = -IV_ERROR_PROCESSING;
            }
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }


    crc = 0;
    // Systemstopp aufheben
    if ( (res = exitWriteDataSegment( deviceAddress, IV_VALUE_SEGMENT_UPDATE, crc )) != IV_ERROR_SUCCESS )
    {

        if ( res > 0 )
        {

// Funktionsfehler
            res = -IV_ERROR_PROCESSING;
        }

        // Fehlerhafte durchfuehrung zurueckmelden.
        goto error_label;
    }

    // Ausfuehrung erfolgreich abgeschlossen

error_label:

    if ( buf ) delete[] buf;
    if ( fp ) fclose( fp );

    return res;
}

int32_t Channel4::writeMemToFile( uint8_t deviceAddress, int32_t type, int32_t address, int32_t size, const std::string &filePath )
{
    StructMemFileHeader header;
    int32_t res = IV_ERROR_SUCCESS;
    char buf[0x400];
    int32_t baseAddress = 0;

    FILE *fp = fopen( filePath.c_str(), "wb" );
    if ( fp == nullptr )
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Dateiheader aufsetzen
    memset( &header, 0, sizeof( header ) );
    header.structSize = sizeof( header );
    header.type = type;
    header.baseAddress = baseAddress;
    header.memAddress = address;
    header.memSize = size;
    // Dateiheader schreiben
    if ( fwrite( &header, sizeof( header ), 1, fp ) != sizeof( header ) )
    {
        return -IV_ERROR_FAIL;
    }

    switch ( type )
    {
        case 0:
            break;

        case 1:
            if ( (res = readFromEeprom( deviceAddress, address, (uint8_t *)buf, size )) != IV_ERROR_SUCCESS )
            {
                goto label_out;
            }
            break;
        case 2:
            if ( (res = readFromFlash( deviceAddress, address, (uint8_t *)buf, size )) != IV_ERROR_SUCCESS )
            {
                goto label_out;
            }
            break;

        case 3:
            break;
    }

label_out:
    if ( fp ) fclose( fp );
    return res;
}

int32_t Channel4::writeFileToMem( uint8_t deviceAddress, const std::string &filePath )
{
    StructMemFileHeader header;
    int32_t res = IV_ERROR_SUCCESS, chunk = 0;
    char buf[0x400];

    FILE *fp = fopen( filePath.c_str(), "rb" );
    if ( fp == nullptr )
    {
        return -IV_ERROR_FILE_OPEN;
    }

    // Dateiheader lesen
    if ( fread( &header, sizeof( header ), 1, fp ) != sizeof( header ) )
    {
        res = -IV_ERROR_FAIL;
        goto label_out;
    }
    switch ( header.type )
    {
        case 0://MB_ACCESS_SRAM:
            break;

        case 1://MB_ACCESS_EEPROM:
            if ( (res = writeToEeprom( deviceAddress, header.memAddress, (uint8_t *)buf, chunk )) != IV_ERROR_SUCCESS )
            {
                goto label_out;
            }
            break;

        case 2://MB_ACCESS_FLASH:
            if ( (res = writeToFlash( deviceAddress, header.memAddress, (uint8_t *)buf, chunk )) != IV_ERROR_SUCCESS )
            {
                goto label_out;
            }
            break;

        case 3://MB_ACCESS_PROGRAM:
            break;

        case 5://MB_ACCESS_FLASH_PAGE:
            break;

        case 6://MB_ACCESS_MMC:
            break;
    }

label_out:
    if ( fp ) fclose( fp );
    return res;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int32_t Channel4::checkData( uint8_t deviceAddress )
{
    int32_t iRes = IV_ERROR_SUCCESS;
    uint8_t buf[0x255];
    uint16_t count = sizeof( buf );
    uint8_t status = 0;

    // Befehlsausfuehrung starten.
    iRes = readRecord( deviceAddress, buf, count, status );

    if ( iRes == IV_ERROR_SUCCESS )
    {
        // OK
        iRes = status;
    }
    else
    {
        iRes = -1;
    }

    return iRes;
}

// ########################################################
// Überlagerungen aus Channel4ModuleApplication
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int32_t Channel4::readRecord( uint8_t deviceAddress, uint8_t *buffer, uint16_t &count, uint8_t &status )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readRecord( deviceAddress, buffer, count, status );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::quitRecord( uint8_t deviceAddress, uint8_t &status, std::string &param )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        // Datensatz quittieren
        res = ModuleTable::quitRecord( deviceAddress, status, param );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::clearRecords( uint8_t deviceAddress )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::clearRecords( deviceAddress );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::restoreRecords( uint8_t deviceAddress )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::restoreRecords( deviceAddress );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeMemoToDisplay( uint8_t deviceAddress, uint8_t maschine, uint8_t time, uint8_t font, uint8_t audio, uint16_t flags, const std::string &text )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeMemoToDisplay( deviceAddress, maschine, time, font, audio, flags, text );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeMessageToDisplay( uint8_t deviceAddress, uint8_t time, uint8_t action, uint8_t audio, const std::string &text )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeMessageToDisplay( deviceAddress, time, action, audio, text );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeSetupVariable( uint8_t deviceAddress, const std::string &name, const std::string &val )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeSetupVariable( deviceAddress, name, val );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readSetupVariable( uint8_t deviceAddress, const std::string &name, std::string &val )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readSetupVariable( deviceAddress, name, val );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::getVersionFirmwareFromFile( const std::string &filePath, std::string &version )
{
    // Lokale Membervariable reseten
    m_fileDevice.resetData();

    if ( !m_fileDevice.GetVersionFirmware( filePath, version ) )
    {
        switch ( m_fileDevice.errorState() )
        {
            case FileDeviceZip::Error_state::none:
                return -IV_ERROR_FAIL;

            case FileDeviceZip::Error_state::file_open:
                return -IV_ERROR_FILE_OPEN;

            case FileDeviceZip::Error_state::head_crc:
                return -IV_ERROR_CRC_PACKET;

            case FileDeviceZip::Error_state::version:
            case FileDeviceZip::Error_state::device_unknown:
            case FileDeviceZip::Error_state::firmware_version_length:
            case FileDeviceZip::Error_state::function_unavailable:
                return -IV_ERROR_FUNCTION_UNAVAILABLE;

            case FileDeviceZip::Error_state::record_empty:
            case FileDeviceZip::Error_state::record_crc:
            case FileDeviceZip::Error_state::record_format:
            case FileDeviceZip::Error_state::record_buffer_overflow:
            case FileDeviceZip::Error_state::file_write:
            case FileDeviceZip::Error_state::file_read:
            case FileDeviceZip::Error_state::file_empty:
            case FileDeviceZip::Error_state::crc_fail:
            case FileDeviceZip::Error_state::format:
            case FileDeviceZip::Error_state::no_data:
            case FileDeviceZip::Error_state::load_display_data:
            case FileDeviceZip::Error_state::load_device_text:
            case FileDeviceZip::Error_state::program_overflow:
            case FileDeviceZip::Error_state::record_make:
            case FileDeviceZip::Error_state::crypt_fix_fail:
            case FileDeviceZip::Error_state::crypt_none:
            case FileDeviceZip::Error_state::crypt_is:
            case FileDeviceZip::Error_state::ini_key:
            case FileDeviceZip::Error_state::ini_key_value:
            case FileDeviceZip::Error_state::size:
            case FileDeviceZip::Error_state::record_end_make:
            case FileDeviceZip::Error_state::program_crc:
            case FileDeviceZip::Error_state::load_raw_firmware:
            case FileDeviceZip::Error_state::version_file:
            case FileDeviceZip::Error_state::as_text:
            case FileDeviceZip::Error_state::file_format:
            case FileDeviceZip::Error_state::line_format:
            case FileDeviceZip::Error_state::alloc_mem:
            case FileDeviceZip::Error_state::no_file_match:
            case FileDeviceZip::Error_state::serialize_file_data:
                return -IV_ERROR_FUNCTION_UNAVAILABLE;
        }

        return -IV_ERROR_FUNCTION_UNAVAILABLE;
    }

    return IV_ERROR_SUCCESS;
}

int32_t Channel4::writeFontType( uint8_t deviceAddress, uint8_t type )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeFontType( deviceAddress, type );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// ########################################################
// Überlagerungen aus Channel4ModuleSystem
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int32_t Channel4::readSystemLogRecord( uint8_t deviceAddress, uint8_t *buffer, uint16_t &count, uint8_t &status )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readSystemLogRecord( deviceAddress, buffer, count, status );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::quitSystemLogRecord( uint8_t deviceAddress, uint8_t &status )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::quitSystemLogRecord( deviceAddress, status );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::restoreSystemLogRecords( uint8_t deviceAddress )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::restoreSystemLogRecords( deviceAddress );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readRtcDateTime( uint8_t deviceAddress, char *time )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readRtcDateTime( deviceAddress, time );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeRtcDateTime( uint8_t deviceAddress, char *time )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeRtcDateTime( deviceAddress, time );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readFirmwareVersion( uint8_t deviceAddress, std::string &version )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readFirmwareVersion( deviceAddress, version );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::systemReset( uint8_t deviceAddress, uint16_t modus )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::systemReset( deviceAddress, modus );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readSerialNumber( uint8_t deviceAddress, uint32_t &serial )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readSerialNumber( deviceAddress, serial );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeRelayState( uint8_t deviceAddress, const std::string &name, uint8_t busNumber, uint8_t relayNumber, uint8_t status, uint16_t time )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeRelayState( deviceAddress, name, busNumber, relayNumber, status, time );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readRelayState( uint8_t deviceAddress, const std::string &name, uint8_t busNumber, uint8_t relayNumber, uint8_t &status, uint16_t &time )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readRelayState( deviceAddress, name, busNumber, relayNumber, status, time );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeSystemVariable( uint8_t deviceAddress, const std::string &name, const std::string &val )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeSystemVariable( deviceAddress, name, val );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readSystemVariable( uint8_t deviceAddress, const std::string &name, std::string &val )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::readSystemVariable( deviceAddress, name, val );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeVirtualKey( uint8_t deviceAddress, uint8_t keyCode, uint8_t flags )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = ModuleTable::writeVirtualKey( deviceAddress, keyCode, flags );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::writeLedState( uint8_t deviceAddress, uint8_t slot, uint8_t id, uint8_t state )
{
    return ModuleSystem::writeLedState( deviceAddress, slot, id, state );
}

int32_t Channel4::readLedState( uint8_t deviceAddress, uint8_t slot, uint8_t id, uint8_t &state )
{
    return ModuleSystem::readLedState( deviceAddress, slot, id, state );
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int32_t Channel4::readDeviceInfo( uint8_t deviceAddress, DeviceInfo &info )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        uint32_t serial;

        std::string version;
        std::string info_text;
        std::array<char, IV_PACKET_LENGTH_DATA> buffer;

        // Initalisieren der DeviceInfo.
        info.clear();

        // Ggf. bei verschlüsselter Kommunikation, Aushandlung durchführen.
        if ( (res = ping( deviceAddress, nullptr, nullptr )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Abfrage der Bootloaderversion
        if ( (res = readBootloaderVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        // Der Infoklasse übergeben.
        info.setVersionBootloader( version );

        // HIP-Datei lesen (importiert ggf. "alte" Hardwareinfo in die Datei).
        HardwareInfo hardwareInfo;
        if ( (res = readHardwareInfo( deviceAddress, hardwareInfo )) != IV_ERROR_SUCCESS )
        {
            switch ( res )
            {
                case 28: // Funktionsfehler: Setup-Passwort wird gefordert.
                    break;

                case 29: // Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.
                    break;

                default:
                    return res;
            }
        }

        // der Infoklasse übergeben.
        info.setHardwareInfo( hardwareInfo );
        hardwareInfo.reset();

        version.clear();
        // Abfrage der Firmwareversion
        if ( (res = readFirmwareVersion( deviceAddress, version )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) == 2 )
            {
                res = IV_ERROR_SUCCESS;
                // Firmwareversion nicht verfuegbar.
                // Geraet befindet sich im Bootloadermodus.
                info.setBootloaderMode( true );
            }
            else
            {
                return res;
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setVersionFirmware( version );
        }

        SoftwareInfo softwareInfo;
        if ( (res = readSoftwareInfo( deviceAddress, softwareInfo )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
            {
                return res;
            }

            res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setSoftwareInfo( softwareInfo );
            softwareInfo.clear();
        }

        FirmwareOptions firmwareOptions;
        if ( (res = readFirmwareOptions( deviceAddress, firmwareOptions )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
            {
                return res;
            }

            res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setFirmwareOptions( firmwareOptions );
            firmwareOptions.clear();
        }

        // Abfrage der Seriennummer.
        if ( (res = readSerialNumber( deviceAddress, serial )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) != 2 )
            {
                return res;
            }

            res = IV_ERROR_SUCCESS;
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setSerialNumber( serial );
        }
        buffer.fill( '\0' );
        // Abfrage des Setuppasswortkeys
        if ( (res = readSetupPasswordKey( deviceAddress, buffer.data() )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion unbekannt?
            if ( IV_RESULT_MODULE( res ) == 2 )
            {
                res = IV_ERROR_SUCCESS;
                // Passwordkey nicht verfuegbar.
                buffer.fill( '\0' );
            }
            else
            {
                return res;
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            info.setPasswordKey( buffer.data() );
        }

        // Eingestellte Hauptkommunikation für die Übernahme der IP/MAC ermitteln.
        uint8_t interfaceType = 0;
        readMainInterfaceType( deviceAddress, interfaceType );
        info.setMainInterfaceType( interfaceType );

        // Zusätzliche Informationen zu verbauten Modulen verfügbar?
        std::vector<std::string> names;

        if ( info.hardwareInfo().moduleBiokey().size() )
        {
            names.push_back( "FINGERPRINT.MODULE" );
        }

        // Die Info liegt lediglich bei folgenden USB-Hosts vor.
        if ( info.hardwareInfo().moduleUSBHostCommunication( { EnumModuleUSBHostCommunication::Vinculum } ).size() )
        {
            names.push_back( "USBHOST.MODULE" );
        }

        if ( info.hardwareInfo().moduleGPS().size() )
        {
            names.push_back( "GPS.MODULE" );
        }

        if ( info.hardwareInfo().moduleLANCommunication().size()
             || info.hardwareInfo().onBoardLANCommunication().size()
             || info.hardwareInfo().moduleWLANCommunication().size() )
        {
            names.push_back( "COM.TCPIP" );
        }

        for ( const std::string &name : names )
        {
            info_text.clear();
            if ( (res = readModuleInfo( deviceAddress, name.c_str(), "", info_text )) != IV_ERROR_SUCCESS )
            {
                // War die Funktion im Modul unbekannt?
                if ( res > 0 && IV_RESULT_MODULE( res ) == 2 )
                {
                    res = IV_ERROR_SUCCESS;
                    // Modulinfo nicht verfuegbar.
                    info_text = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 1 || IV_RESULT_FUNCTION( res ) == 2) )
                {
                    res = IV_ERROR_SUCCESS;
                    // Systemvariable ist nicht verfügbar.
                    info_text = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 6 || IV_RESULT_FUNCTION( res ) == 7 || IV_RESULT_FUNCTION( res ) == 8) )
                {
                    // 6 = Die Information zur Komponente steht nicht zur Verfügung.
                    // 7 = Das Modul kann zur Zeit nicht angesprochen werden.
                    // 8 = Das anzusprechende Modul ist nicht vorhanden.
                    bool na = true;

                    if ( name == "COM.TCPIP" )
                    {
                        std::string mac;
                        if ( (res = readSystemVariable( deviceAddress, "com.mac", mac )) == IV_ERROR_SUCCESS && mac.size() == 12 )
                        {
                            // Bindestriche einfügen
                            int32_t i = static_cast<uint32_t>(mac.size()) - 2;
                            for ( ; i >= 2; i -= 2 )
                            {
                                mac.insert( i, "-" );
                            }

                            std::strncpy( buffer.data(), mac.c_str(), buffer.size() - 1 );

                            // Not available, abstellen.
                            na = false;
                        }
                    }

                    if ( na )
                    {
                        res = IV_ERROR_SUCCESS;
                        // Abfrage steht nicht zur Verfügung.
                        buffer.fill( '\0' );
                        std::strncpy( buffer.data(), "NA", buffer.size() - 1 );
                    }
                }
                else if ( res == 28 || res == 29 )
                {
                    // 28 == Funktionsfehler: Setup-Passwort wird gefordert.
                    // 29 == Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.

                    res = IV_ERROR_SUCCESS;
                    // Abfrage steht nicht zur Verfügung.
                    buffer.fill( '\0' );
                    std::strncpy( buffer.data(), "NA", buffer.size() - 1 );

                    break;
                }
                else
                {
                    return res;
                }
            }
            else
            {
                std::strncpy( buffer.data(), info_text.c_str(), buffer.size() - 1 );
            }

            if ( name == "FINGERPRINT.MODULE" )
            {
                info.setVersionFingerprint( buffer.data() );
            }
            else if ( name == "USBHOST.MODULE" )
            {
                info.setVersionUSBHost( buffer.data() );
            }
            else if ( name == "GPS.MODULE" )
            {
                info.setVersionGPS( buffer.data() );
            }
            else if ( name == "COM.TCPIP" )
            {
                info.setVersionTCPIP( buffer.data() );
            }
        }

        // Zusätzlich folgende Systemvariablen abrufen.
        names.clear();

        // Gesetzte Mac-Adresse
        if ( info.hardwareInfo().moduleLANCommunication().size()
             || info.hardwareInfo().onBoardLANCommunication().size() )
        {
            names.push_back( "COM.MAC" );
        }

        // Zutrittskontrolle nur bei verbautem RS485 aktiv.
        if ( info.hardwareInfo().moduleRS485Communication().size()
             || info.hardwareInfo().onBoardRS485Communication().size()
             || info.hardwareInfo().moduleAccessControl().size()
             || info.hardwareInfo().onBoardAccessControl().size() )
        {
            names.push_back( "ACCESS.MASTERID" );
        }

        if ( info.hardwareInfo().deviceType() == HardwareInfo::TimeboyIV || info.hardwareInfo().deviceType() == HardwareInfo::TimeboyMobilPZE )
        {
            names.push_back( "COM.GROUPID" );
        }

        if ( info.hardwareInfo().deviceType() == HardwareInfo::KYO_Oneloc
             || info.hardwareInfo().deviceType() == HardwareInfo::EVO_3_5_Universal )
        {
            if ( info.hardwareInfo().deviceType() == HardwareInfo::KYO_Oneloc )
            {
                names.push_back( "board.application" );
            }
            names.push_back( "board.place[0].output[1].mode" );
            names.push_back( "board.place[0].serial[1].use" );
            names.push_back( "board.place[0].serial[2].use" );
        }

        if ( info.hardwareInfo().deviceType() == HardwareInfo::IPCExtensionC )
        {
            // Bei Embedded Q7
            names.push_back( "system.powerswitch.timestartos" );
            names.push_back( "system.powerswitch.offtimelong" );
            names.push_back( "system.powerswitch.offtimeshort" );
            names.push_back( "system.powerswitch.ontime" );
            names.push_back( "system.powerswitch.restoreacpowerloss" );
        }

        if ( info.hardwareInfo().deviceType() == HardwareInfo::KYO_Cenloc )
        {
            names.push_back( "ACCESS.MODULES.STATE" );
        }

        for ( const std::string &name : names )
        {
            std::string val;
            if ( (res = readSystemVariable( deviceAddress, name.c_str(), val )) != IV_ERROR_SUCCESS )
            {
                // War die Funktion im Modul unbekannt?
                if ( res > 0 && IV_RESULT_MODULE( res ) == 2 )
                {
                    res = IV_ERROR_SUCCESS;
                    // Modulinfo nicht verfuegbar.
                    val = "NA";
                }
                else if ( res > 0 && (IV_RESULT_FUNCTION( res ) == 1 || IV_RESULT_FUNCTION( res ) == 2) )
                {
                    res = IV_ERROR_SUCCESS;
                    // Major- oder Minorname nicht verfügbar.
                    val = "NA";
                }
                else if ( res == -28 || res == -29 )
                {
                    // 28 == Funktionsfehler: Setup-Passwort wird gefordert.
                    // 29 == Funktionsfehler: Setup-Passwort wird gefordert, Gerät wegen zu vielen Fehlanfragen gesperrt.

                    res = IV_ERROR_SUCCESS;
                    // Abfrage steht nicht zur Verfügung.
                    val = "NA";
                    break;
                }
                else
                {
                    return res;
                }
            }

            if ( name == "COM.MAC" )
            {
                if ( val.size() == 12 )
                {
                    // Bindestriche einfügen
                    for ( auto i = val.size() - 2; i >= 2; i -= 2 )
                    {
                        val.insert( i, "-" );
                    }
                }

                info.setMacAddress( val );
            }
            else if ( name == "ACCESS.MASTERID" )
            {
                info.setAccessControlMasterId( val );
            }
            else if ( name == "COM.GROUPID" )
            {
                info.setTimeboyGroupId( val );
            }
            else if ( name == "board.application" )
            {
                info.setBoardApplication( val );
            }
            else if ( name == "board.place[0].output[1].mode" )
            {
                info.setBoardPlace_0_Output_1_Mode( val );
            }
            else if ( name == "board.place[0].serial[1].use" )
            {
                info.setBoardPlace_0_Serial_1_Use( val );
            }
            else if ( name == "board.place[0].serial[2].use" )
            {
                info.setBoardPlace_0_Serial_2_Use( val );
            }
            else if ( name == "system.powerswitch.timestartos" )
            {
                info.setTimeStartOS( val );
            }
            else if ( name == "system.powerswitch.offtimelong" )
            {
                info.setOffTimeLong( val );
            }
            else if ( name == "system.powerswitch.offtimeshort" )
            {
                info.setOffTimeShort( val );
            }
            else if ( name == "system.powerswitch.ontime" )
            {
                info.setOnTime( val );
            }
            else if ( name == "system.powerswitch.restoreacpowerloss" )
            {
                info.set_restore_ac_power_loss( val );
            }
            else if ( name == "ACCESS.MODULES.STATE" )
            {
                info.setAccessModulesState( val );
            }
        }

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readHardwareInfo( uint8_t deviceAddress, HardwareInfo &info )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        uint8_t val = 0;

        // Initalisieren.
        info.reset();

        // Lesen der Hardwareconfig-Nr
        if ( (res = readHardwareInfoVersion( deviceAddress, val )) != IV_ERROR_SUCCESS )
            return res;

        // Gelesene Version zur Information halten.
        info.setVersion( val );

        // Ab Version 4 kommt eine geänderte HWI zum Einsatz.
        if ( val < 4 )
        {
            uint8_t typeDevice;
            uint8_t typeBoard;
            // Lesen des Gerätetypen
            if ( (res = readDeviceType( deviceAddress, typeDevice )) != IV_ERROR_SUCCESS )
                return res;

            // Lesen der Boardversion
            if ( (res = readBoardType( deviceAddress, typeBoard )) != IV_ERROR_SUCCESS )
                return res;

            std::array<uint8_t, sizeof( StructHardwareInfo )> dshwi;
            dshwi.fill( '\0' );
            StructHardwareInfo &shw = reinterpret_cast<StructHardwareInfo &>(dshwi);
            shw.version = val;
            shw.typeDevice = typeDevice;
            shw.typeBoard = typeBoard;

            // Prüfen ob es sich um ein MasterIV handelt.
            if ( info.isMasterIV( shw ) )
            {
                // Lesen der Vorgaenger-Hardwareinfo
                if ( (res = ModuleTable::readHardwareInfo( deviceAddress, shw )) != IV_ERROR_SUCCESS )
                    return res;

                // Vorgängerversion importieren
                if ( !info.importStructHardwareInfo( shw ) )
                    return -IV_ERROR_FUNCTION_UNAVAILABLE;

                return res;
            }
            else
            {
                // Es handelt sich um eine Zwischenvariante bei der Gerätetyp und Boardrevision noch von der
                // Vorgängerversion genutzt wurden.
            }
        }

        //
        uint32_t address, size, crc;
        address = size = crc = 0;

        // Segmentbegin: Bedienung stoppen
        if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            if ( (res & 0xFF) == 1 )
            {
                // Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
                // Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
                res = -IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                return res;
            }
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc );

        // Daten aus dem Gerät in die Datei der HardwareInfo lesen.
        info.reset();
        FileHardwareInfo &fileHardwareInfo = info.fileHardwareInfo();

        //////////////////////////////
        // Begin

        // Segmentbegin: Config uebertragen
        if ( (res = enterReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_HARDWARECONFIGS, address, size, crc )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ beginGetSegment(%i): %u, %u, 0x%08X ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_HARDWARECONFIGS, address, size, crc );

        if ( size == 0 )
        {
            // Es liegen keine Daten vor.
            res = -IV_ERROR_NO_DATA_AVAILABLE;
            goto label_no_data;
        }
        //////////////////////////////
        // Daten
        info.fileHardwareInfo().setSizeData( size );

        // HardwareInfo lesen
        if ( (res = readFromFlash( deviceAddress, address, fileHardwareInfo.data(), fileHardwareInfo.sizeData() )) != IV_ERROR_SUCCESS )
        {
            return res;
        }

        //////////////////////////////
        // Ende
    label_no_data:
            // Segmentend: Setup uebertragen
        if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_HARDWARECONFIGS )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_HARDWARECONFIGS );

        // Segmentend: Bedienung stoppen
        if ( (res = exitReadDataSegment( deviceAddress, IV_VALUE_SEGMENT_GET_UPDATE_BG )) != IV_ERROR_SUCCESS )
        {
            return res;
        }
        trace( "%s(%i):INFO(1): %s [ endGetSegment(%i) ]\n", __FILE__, __LINE__, __func__, IV_VALUE_SEGMENT_GET_UPDATE_BG );

        if ( size )
        {
            // Datenanzahl in Datei uebernehmen.
            uint32_t crcTest = fileHardwareInfo.checksum32bit();
            if ( crc != crcTest )
            {
                trace( "%s(%i):ERROR(1): %s [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, __func__, crc, crcTest );
                return -IV_ERROR_CRC_SEGMENT;
            }

            trace( "%s(%i):INFO(1): %s [ CRC von Segment OK ]\n", __FILE__, __LINE__, __func__ );

            // Baum aus den Daten erzeugen.
            if ( !fileHardwareInfo.readSerialize() )
            {
                trace( "%s(%i):ERROR(1): %s [ Lesen der Datei gescheitert. ]\n", __FILE__, __LINE__, __func__ );
                return -IV_ERROR_FILE_WRITE;
            }

            trace( "%s(%i):INFO(1): %s [ Interpretation der HIP-Daten erfolgreich. ]\n", __FILE__, __LINE__, __func__ );
        }

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int32_t Channel4::readSoftwareInfo( uint8_t deviceAddress, SoftwareInfo &info )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        StructSoftwareInfo swInfo;
        int32_t size = sizeof( StructSoftwareInfo );
        memset( &swInfo, 0, sizeof( StructSoftwareInfo ) );

        // Initalisieren
        info.clear();

        // Abfrage der Softwareinfo
        if ( (res = ModuleTable::readSoftwareInfo( deviceAddress, (uint8_t *)&swInfo, size )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion im Modul unbekannt?
            if ( (res > 0) && (((res >> 8) & 0xFF) == 2) )
            {
                // Softwareinformation nicht verfuegbar.
            }
            else
            {
                return res;
            }
        }

        // Der Infoklasse übergeben.
        info.setStructSoftwareInfo( swInfo );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int32_t Channel4::readFirmwareOptions( uint8_t deviceAddress, FirmwareOptions &options )
{
    int32_t res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        bool haveFunction;
        uint32_t flags = 0;
        uint8_t type;
        uint8_t available;

        // Initalisieren
        options.clear();

        // Zusaetzlich das Flag fuer "Setup und Listen beibehalten" auswerten.
        if ( (res = readSystemFlags( deviceAddress, flags )) != IV_ERROR_SUCCESS )
        {
            // War die Funktion im Modul unbekannt?
            if ( (res > 0) && (((res >> 8) & 0xFF) == 2) )
            {
                // Versuchen direkt an der Adresse zu lesen und zu setzen
                if ( readFromEeprom( deviceAddress, 0x0ECC, (uint8_t *)&flags, sizeof( uint32_t ) ) == IV_ERROR_SUCCESS )
                {
                    // Der Infoklasse übergeben.
                    options.setSystemFlags( flags );
                }
            }
        }
        else
        {
            // Der Infoklasse übergeben.
            options.setSystemFlags( flags );
        }

#undef CHECK_COULD_HAVE_FUNCTION
#define CHECK_COULD_HAVE_FUNCTION( x ) \
    haveFunction = true; \
    if ( (res = x) != IV_ERROR_SUCCESS ) \
        { \
        if ( (res > 0)  && (((res >> 8) & 0xFF) == 2) ) \
        { \
        res = IV_ERROR_SUCCESS; \
        /* Funktion kann ggf. aufgrund eines älteren Firmwarestandes undefiniert sein. */ \
        haveFunction = false; \
        } \
    else \
        { \
        /* Davon ausgehen, dass das Geraet unerreichbar ist.*/ \
        return res; \
        } \
        }

        // Typ ermitteln
        CHECK_COULD_HAVE_FUNCTION( readExternalTransponderReaderType( deviceAddress, type ) );
        if ( haveFunction )
            options.setTypeExternalReader( type );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readIoModuleOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionIOModule( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readTimeboyAdaptionOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionTBAdaption( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readComServerOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionComServer( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readTransponderWriteOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionTPWrite( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readHttpOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionHTTP( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readLocalizationOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionLanguage( available );

        // Option ermitteln
        CHECK_COULD_HAVE_FUNCTION( readFingerprintScannerUpdateOption( deviceAddress, available ) );
        if ( haveFunction )
            options.setOptionBiokeyUpdate( available );

        uint16_t so_mask, so_options;
        // SetupOption: Signalverarbeitung
        CHECK_COULD_HAVE_FUNCTION( readSetupOptions( deviceAddress, so_mask, so_options ) );
        if ( haveFunction )
            options.setSetupOptions( so_mask, so_options );

        uint8_t po_mask, po_options;
        // ErweiterteOption: Data on Card
        CHECK_COULD_HAVE_FUNCTION( readBuyOptions( deviceAddress, po_mask, po_options ) );
        if ( haveFunction )
            options.setPayableOptions( po_mask, po_options );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

#undef CHECK_COULD_HAVE_FUNCTION

    return res;
}

// ##################################################################
/**
* @brief Ermittelt eine Sammlung von Informationen zum Gerät.
*
*/
int32_t Channel4::writeFirmwareOptions( uint8_t deviceAddress, const FirmwareOptions &options )
{
    (void)options;
    int32_t res = IV_ERROR_SUCCESS;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {



        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::enterTransparencyMode( uint8_t deviceAddress, uint16_t id, uint16_t flags, uint32_t readBufferSize, uint32_t writeBufferSize, uint32_t timeout )
{
    return ModuleSystem::enterTransparencyMode( deviceAddress, id, flags, readBufferSize, writeBufferSize, timeout );
}

int32_t Channel4::exitTransparencyMode( uint8_t deviceAddress )
{
    return ModuleSystem::exitTransparencyMode( deviceAddress );
}

bool Channel4::isTransparencyMode()
{
    return ModuleSystem::isTransparencyMode();
}

int32_t Channel4::readDataBlockWithTimeboySlot( uint8_t deviceAddress, const std::string &name, char *data, uint32_t *length, uint32_t flags )
{
    int32_t res;

    // Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
    if ( (res = setTimeboySlotState( deviceAddress, 1 )) == IV_ERROR_SUCCESS )
    {
        // Befehlsausfuehrung starten.
        res = readDataBlock( deviceAddress, name, data, length, flags );

        // Versuchen die rote LED abzuschalten.
        setTimeboySlotState( deviceAddress, 0, res );
    }

    return res;
}

int32_t Channel4::readDataBlock( uint8_t deviceAddress, const std::string &name, char *data, uint32_t *length, uint32_t flags )
{
    int32_t res = IV_ERROR_SUCCESS;
    std::vector<uint8_t> buffer;

    if ( (res = ModuleApplication::readDataBlock( deviceAddress, name, buffer )) != IV_ERROR_SUCCESS )
    {
        // War die Funktion im Modul unbekannt?
        if ( (res > 0) && (((res >> 8) & 0xFF) == 2) )
        {

        }
        else
        {
            return res;
        }

    }
    else
    {
        if ( *length >= buffer.size() )
        {
            std::copy( buffer.cbegin(), buffer.cend(), data );
            *length = static_cast<uint32_t>(buffer.size());
        }
        else if ( *length < buffer.size() )
        {
            std::copy_n( buffer.cbegin(), *length, data );
            // Es konnten nicht alle Daten kopiert werden.
        }
        else
        {
            if ( (flags & 1) == 1 )
            {
                // Datei mit dem Dateninhalt erstellen.
            }
        }
    }

    return res;
}

int32_t Channel4::dfzList( std::list<std::string> *list )
{
    //list = &m_dfzList;
    return IV_ERROR_SUCCESS;
}

std::tuple<std::string, const std::list<std::string>> Channel4::minFwVersion(const std::string& dbPath, HardwareInfo & info)
{
    int nRes;
    std::list<std::pair<std::string, std::string>> processedModules;
    std::string errorMsg;
    const std::list<std::string> tempList{};
    bool worked = m_deviceUpdate->updatePathToDB(dbPath, DeviceUpdate::DBPathType::Local, errorMsg);
    if (!worked) return std::make_tuple(errorMsg, tempList);
    if(!info.isMasterIV())
    {
        trace("%s(%i):INFO(1): %s [ Ermittlung der Mindestfirmwareversion. ]\n", __FILE__, __LINE__, __func__);
        for (auto& item : info.fileHardwareInfo().rootItemDeviceHardwareInfo().childs())
        {

            if (!m_deviceUpdate->detectMinFWVersion(item, processedModules))
            {
                nRes = -IV_ERROR_DEVICEFILE_MATCH;
            }
        }
        trace("%s(%i):INFO(1): %s Moduluebersicht\n", __FILE__, __LINE__, __func__);
        trace("%s(%i):INFO(1): %s ---------------\n", __FILE__, __LINE__, __func__);
        for (const auto& item : processedModules)
        {
            trace("%s(%i):INFO(1): %s %s\t--> Minimalversion: %s\n", __FILE__, __LINE__, __func__, item.first.c_str(), item.second.c_str());
        }

        trace("%s(%i):INFO(1): %s [ Ermittlung der Blacklistversionen. ]\n", __FILE__, __LINE__, __func__);
        for (auto& item : info.fileHardwareInfo().rootItemDeviceHardwareInfo().childs())
        {
            if (!m_deviceUpdate->collectBlacklistDefinition(item))
            {
                nRes = -IV_ERROR_FILE_PROCESSING;
            }
        }
        nRes =  IV_ERROR_SUCCESS;
    }
    auto b = m_deviceUpdate->minFwVersion();
    std::list<std::string> blackList{};
    for (auto const& blackMember : m_deviceUpdate->blacklistVersions()) {
        blackList.push_back(blackMember);
    }
    SoftwareVersion minV = static_cast<SoftwareVersion>(m_deviceUpdate->minFwVersion());
    unsigned int v = minV;
    const char* vers = minV;
    return std::make_tuple(vers, blackList);
}

void Channel4::cleanUpUpdateDB()
{
    // löschen der temporären Update-Datenbank
    if ( m_deviceUpdate )
    {
        const std::string dbPath = m_deviceUpdate->dbPath();
        if ( Utils::File::exists( dbPath ) )
        {
            std::string errorMsg;
            m_deviceUpdate->close( errorMsg );

            // Wenn die Datei erfolgreich gelöscht wurde, wird ein Nullwert zurückgegeben.
            // Bei einem Fehlschlag wird ein Wert ungleich Null zurückgegeben.
            if ( ::remove( dbPath.c_str() ) )
            {
                trace( boost::str( boost::format( "%1%(%2%):INFO(1): %3% [ InternalError: Error deleting temporary database file - %4% ]\n" ) % __FILE__ % __LINE__ % __func__ % strerror( errno ) ).c_str() );
            }
        }
    }
}

int32_t Channel4::detectHWInfo( const uint8_t &deviceAddress, StructHardwareInfo &oldHWInfo, HardwareInfo &hwInfo )
{
    int32_t res{};

    // Auslesen der Hardwareinfo
    if ( (res = readHardwareInfo( deviceAddress, hwInfo )) != IV_ERROR_SUCCESS )
    {
        return res;
    }

    // Uebergangsweise auf die Vorgaengerversion wechseln.
    oldHWInfo = hwInfo.exportStructHardwareInfo();

    // Liegt keine Hardwareinfo vor?
    // Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
    if ( oldHWInfo.version == 0 )
    {
        int32_t handle;
        std::string reserved;
        if ( openTable( deviceAddress, "ACCESS.READER", 0, handle, reserved ) == IV_ERROR_SUCCESS
             && closeTable( deviceAddress, handle ) == IV_ERROR_SUCCESS )
        {
            // ZK-Option in Verwendung.
            oldHWInfo.administrationFlags |= HIAF_HAS_ZK;
        }
    }

    return res;
}

int32_t Channel4::determineMinFWVersion( const StructHardwareInfo &oldHWInfo, HardwareInfo &hwInfo, const std::string &filePath,
                                     std::list<std::pair<std::string, std::string>> &processedModules )
{
    trace( "%s(%i):INFO(1): %s [ Entpacken der Updatedatenbank aus dem DFZ-Archiv. ]\n", __FILE__, __LINE__, __func__ );
    std::string errorMsg;
    DeviceUpdate::DBUnzipResult dbUnzipResult = m_deviceUpdate->unzipDB( m_fileDevice, filePath, errorMsg );
    if ( dbUnzipResult == DeviceUpdate::DBUnzipResult::Failed )
    {
        m_fileDevice.setErrorTextExtended( errorMsg );
        return -IV_ERROR_DEVICEFILE_LOAD;
    }
    trace( "%s(%i):INFO(1): %s [ Entpacken der Updatedatenbank abgeschlossen (Dateiname: '%s', Ergebnis: %i) ]\n", __FILE__, __LINE__, __func__, m_deviceUpdate->dbPath().c_str(), static_cast<uint8_t>(dbUnzipResult) );

    // Modulinformationen lesen und die erlaubte Mindestversion ermitteln
    if ( (dbUnzipResult == DeviceUpdate::DBUnzipResult::Ok) && !hwInfo.isMasterIV( oldHWInfo ) )
    {
        trace( "%s(%i):INFO(1): %s [ Ermittlung der Mindestfirmwareversion. ]\n", __FILE__, __LINE__, __func__ );
        for ( auto &item : hwInfo.fileHardwareInfo().rootItemDeviceHardwareInfo().childs() )
        {

            if ( !m_deviceUpdate->detectMinFWVersion( item, processedModules ) )
            {
                return -IV_ERROR_DEVICEFILE_MATCH;
            }
        }
        trace( "%s(%i):INFO(1): %s Moduluebersicht\n", __FILE__, __LINE__, __func__ );
        trace( "%s(%i):INFO(1): %s ---------------\n", __FILE__, __LINE__, __func__ );
        for ( const auto &item : processedModules )
        {
            trace( "%s(%i):INFO(1): %s %s\t--> Minimalversion: %s\n", __FILE__, __LINE__, __func__, item.first.c_str(), item.second.c_str() );
        }

        trace( "%s(%i):INFO(1): %s [ Ermittlung der Blacklistversionen. ]\n", __FILE__, __LINE__, __func__ );
        for ( auto &item : hwInfo.fileHardwareInfo().rootItemDeviceHardwareInfo().childs() )
        {
            if ( !m_deviceUpdate->collectBlacklistDefinition( item ) )
            {
                return -IV_ERROR_FILE_PROCESSING;
            }
        }
        return IV_ERROR_SUCCESS;
    }
    return IV_ERROR_USER_CANCEL;
}

int32_t Channel4::determineFWFile( std::string zipFilePath, const StructHardwareInfo &oldHWInfo )
{
    trace( "%s(%i):INFO(1): %s [ Suche der passenden Firmwaredatei. (%s) ]\n", __FILE__, __LINE__, __func__, zipFilePath.c_str() );
    // Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
    return (1 == m_fileDevice.SearchInfosZip( zipFilePath, oldHWInfo, m_deviceUpdate.get(), [this]( const char *msg )
    {
        trace( msg );
    } ));
}
